/*
 * DotConverter.cpp
 *
 *  Created on: Apr 1, 2013
 *      Author: Martin Zak
 */

#include "DotConverter.h"

#include <automaton/FSM/NFA.h>
#include <automaton/FSM/EpsilonNFA.h>
#include <automaton/FSM/MultiInitialStateNFA.h>
#include <automaton/FSM/DFA.h>
#include <automaton/FSM/ExtendedNFA.h>
#include <automaton/FSM/CompactNFA.h>
#include <automaton/TA/NFTA.h>
#include <automaton/TA/DFTA.h>
#include <automaton/PDA/NPDA.h>
#include <automaton/PDA/SinglePopNPDA.h>
#include <automaton/PDA/InputDrivenDPDA.h>
#include <automaton/PDA/InputDrivenNPDA.h>
#include <automaton/PDA/VisiblyPushdownDPDA.h>
#include <automaton/PDA/VisiblyPushdownNPDA.h>
#include <automaton/PDA/RealTimeHeightDeterministicDPDA.h>
#include <automaton/PDA/RealTimeHeightDeterministicNPDA.h>
#include <automaton/PDA/DPDA.h>
#include <automaton/PDA/SinglePopDPDA.h>
#include <automaton/TM/OneTapeDTM.h>

#include <factory/StringDataFactory.hpp>

#include <exception/CommonException.h>

#include <set>
#include <map>
#include <list>
#include <utility>
#include <vector>
#include <typeinfo>

auto replace = [](std::string& str, const std::string& what, const std::string& with) {
	size_t index = 0;
	while((index = str.find(what, index)) != std::string::npos) {
		str.replace(index, what.length(), with);
		index += with.length();
	}
};

void DotConverter::convert(std::ostream& out, const automaton::Automaton& a) {
	dispatch(out, a.getData());
}

void DotConverter::convert(std::ostream& out, const automaton::EpsilonNFA < > & a) {
	out << "digraph automaton {\n";
	out << "rankdir=LR;\n";
	int cnt = 1;

	//Map states to indices
	std::map<label::Label, int> states;
	for (const label::Label& state : a.getStates()) {
		states.insert(std::make_pair(state, cnt++));
	}

	//Print final states
	for (const label::Label& state : a.getFinalStates()) {
		out << "node [shape = doublecircle, label=\"" << alib::StringDataFactory::toString(state) << "\"]; " << states.find(state)->second << ";\n";
	}

	//Print nonfinal states
	for (const auto& state : states) {
		if (!a.getFinalStates().count(state.first)) {
			out << "node [shape = circle, label=\"" << alib::StringDataFactory::toString(state.first) << "\" ]; " << state.second << ";\n";
		}
	}

	//Mark initial states
	out << "node [shape = plaintext, label=\"start\"]; 0; \n";
	out << "0 -> " << states.find(a.getInitialState())->second << ";\n";

	transitions(a, states, out);
	out << "}";
}

auto DotConverterEpsilonNFA = DotConverter::RegistratorWrapper<void, automaton::EpsilonNFA < > >(DotConverter::convert);

void DotConverter::convert(std::ostream& out, const automaton::MultiInitialStateNFA < >& a) {
	out << "digraph automaton {\n";
	out << "rankdir=LR;\n";
	int cnt = 1;

	//Map states to indices
	std::map<label::Label, int> states;
	for (const label::Label& state : a.getStates()) {
		states.insert(std::make_pair(state, cnt++));
	}

	//Print final states
	for (const label::Label& state : a.getFinalStates()) {
		out << "node [shape = doublecircle, label=\"" << alib::StringDataFactory::toString(state) << "\"]; " << states.find(state)->second << ";\n";
	}

	//Print nonfinal states
	for (const auto& state : states) {
		if (!a.getFinalStates().count(state.first)) {
			out << "node [shape = circle, label=\"" << alib::StringDataFactory::toString(state.first) << "\" ]; " << state.second << ";\n";
		}
	}

	//Mark initial states
	out << "node [shape = plaintext, label=\"start\"]; 0; \n";
	for (const label::Label& state : a.getInitialStates()) {
		out << "0 -> " << states.find(state)->second << ";\n";
	}

	transitions(a, states, out);
	out << "}";
}

auto DotConverterMultiInitialStateNFA = DotConverter::RegistratorWrapper<void, automaton::MultiInitialStateNFA < >>(DotConverter::convert);

void DotConverter::convert(std::ostream& out, const automaton::NFA < > & a) {
	out << "digraph automaton {\n";
	out << "rankdir=LR;\n";
	int cnt = 1;

	//Map states to indices
	std::map<label::Label, int> states;
	for (const label::Label& state : a.getStates()) {
		states.insert(std::make_pair(state, cnt++));
	}

	//Print final states
	for (const label::Label& state : a.getFinalStates()) {
		out << "node [shape = doublecircle, label=\"" << alib::StringDataFactory::toString(state) << "\"]; " << states.find(state)->second << ";\n";
	}

	//Print nonfinal states
	for (const auto& state : states) {
		if (!a.getFinalStates().count(state.first)) {
			out << "node [shape = circle, label=\"" << alib::StringDataFactory::toString(state.first) << "\" ]; " << state.second << ";\n";
		}
	}

	//Mark initial states
	out << "node [shape = plaintext, label=\"start\"]; 0; \n";
	out << "0 -> " << states.find(a.getInitialState())->second << ";\n";

	transitions(a, states, out);
	out << "}";
}

auto DotConverterNFA = DotConverter::RegistratorWrapper<void, automaton::NFA < > >(DotConverter::convert);

void DotConverter::convert(std::ostream& out, const automaton::DFA<>& a) {
	out << "digraph automaton {\n";
	out << "rankdir=LR;\n";
	int cnt = 1;

	//Map states to indices
	std::map<label::Label, int> states;
	for (const label::Label& state : a.getStates()) {
		states.insert(std::make_pair(state, cnt++));
	}

	//Print final states
	for (const label::Label& state : a.getFinalStates()) {
		out << "node [shape = doublecircle, label=\"" << alib::StringDataFactory::toString(state) << "\"]; " << states.find(state)->second << ";\n";
	}

	//Print nonfinal states
	for (const auto& state : states) {
		if (!a.getFinalStates().count(state.first)) {
			out << "node [shape = circle, label=\"" << alib::StringDataFactory::toString(state.first) << "\" ]; " << state.second << ";\n";
		}
	}

	//Mark initial states
	out << "node [shape = plaintext, label=\"start\"]; 0; \n";
	out << "0 -> " << states.find(a.getInitialState())->second << ";\n";

	transitions(a, states, out);
	out << "}";
}

auto DotConverterDFA = DotConverter::RegistratorWrapper<void, automaton::DFA<>>(DotConverter::convert);

void DotConverter::convert(std::ostream& out, const automaton::ExtendedNFA& a) {
	out << "digraph automaton {\n";
	out << "rankdir=LR;\n";
	int cnt = 1;

	//Map states to indices
	std::map<label::Label, int> states;
	for (const label::Label& state : a.getStates()) {
		states.insert(std::make_pair(state, cnt++));
	}

	//Print final states
	for (const label::Label& state : a.getFinalStates()) {
		out << "node [shape = doublecircle, label=\"" << alib::StringDataFactory::toString(state) << "\"]; " << states.find(state)->second << ";\n";
	}

	//Print nonfinal states
	for (const auto& state : states) {
		if (!a.getFinalStates().count(state.first)) {
			out << "node [shape = circle, label=\"" << alib::StringDataFactory::toString(state.first) << "\" ]; " << state.second << ";\n";
		}
	}

	//Mark initial states
	out << "node [shape = plaintext, label=\"start\"]; 0; \n";
	out << "0 -> " << states.find(a.getInitialState())->second << ";\n";

	transitions(a, states, out);
	out << "}";
}

auto DotConverterExtendedNFA = DotConverter::RegistratorWrapper<void, automaton::ExtendedNFA>(DotConverter::convert);

void DotConverter::convert(std::ostream& out, const automaton::CompactNFA < > & a) {
	out << "digraph automaton {\n";
	out << "rankdir=LR;\n";
	int cnt = 1;

	//Map states to indices
	std::map<label::Label, int> states;
	for (const label::Label& state : a.getStates()) {
		states.insert(std::make_pair(state, cnt++));
	}

	//Print final states
	for (const label::Label& state : a.getFinalStates()) {
		out << "node [shape = doublecircle, label=\"" << alib::StringDataFactory::toString(state) << "\"]; " << states.find(state)->second << ";\n";
	}

	//Print nonfinal states
	for (const auto& state : states) {
		if (!a.getFinalStates().count(state.first)) {
			out << "node [shape = circle, label=\"" << alib::StringDataFactory::toString(state.first) << "\" ]; " << state.second << ";\n";
		}
	}

	//Mark initial states
	out << "node [shape = plaintext, label=\"start\"]; 0; \n";
	out << "0 -> " << states.find(a.getInitialState())->second << ";\n";

	transitions(a, states, out);
	out << "}";
}

auto DotConverterCompactNFA = DotConverter::RegistratorWrapper<void, automaton::CompactNFA < > >(DotConverter::convert);

void DotConverter::convert(std::ostream& out, const automaton::NFTA& a) {
	out << "digraph automaton {\n";
	out << "rankdir=LR;\n";
	int cnt = 1;

	//Map states to indices
	std::map<label::Label, int> states;
	for (const label::Label& state : a.getStates()) {
		states.insert(std::make_pair(state, cnt++));
	}

	//Print final states
	for (const label::Label& state : a.getFinalStates()) {
		out << "node [shape = doublecircle, label=\"" << alib::StringDataFactory::toString(state) << "\"]; " << states.find(state)->second << ";\n";
	}

	//Print nonfinal states
	for (const auto& state : states) {
		if (!a.getFinalStates().count(state.first)) {
			out << "node [shape = circle, label=\"" << alib::StringDataFactory::toString(state.first) << "\" ]; " << state.second << ";\n";
		}
	}

	transitions(a, states, out);
	out << "}";
}

auto DotConverterNFTA = DotConverter::RegistratorWrapper<void, automaton::NFTA>(DotConverter::convert);

void DotConverter::convert(std::ostream& out, const automaton::DFTA& a) {
	out << "digraph automaton {\n";
	out << "rankdir=LR;\n";
	int cnt = 1;

	//Map states to indices
	std::map<label::Label, int> states;
	for (const label::Label& state : a.getStates()) {
		states.insert(std::make_pair(state, cnt++));
	}

	//Print final states
	for (const label::Label& state : a.getFinalStates()) {
		out << "node [shape = doublecircle, label=\"" << alib::StringDataFactory::toString(state) << "\"]; " << states.find(state)->second << ";\n";
	}

	//Print nonfinal states
	for (const auto& state : states) {
		if (!a.getFinalStates().count(state.first)) {
			out << "node [shape = circle, label=\"" << alib::StringDataFactory::toString(state.first) << "\" ]; " << state.second << ";\n";
		}
	}

	transitions(a, states, out);
	out << "}";
}

auto DotConverterDFTA = DotConverter::RegistratorWrapper<void, automaton::DFTA>(DotConverter::convert);

void DotConverter::convert(std::ostream& out, const automaton::DPDA < > & a) {
	out << "digraph automaton {\n";
	out << "rankdir=LR;\n";
	int cnt = 1;

	//Map states to indices
	std::map<label::Label, int> states;
	for (const label::Label& state : a.getStates()) {
		states.insert(std::make_pair(state, cnt++));
	}

	//Print final states
	for (const label::Label& state : a.getFinalStates()) {
		out << "node [shape = doublecircle, label=\"" << alib::StringDataFactory::toString(state) << "\"]; " << states.find(state)->second << ";\n";
	}

	//Print nonfinal states
	for (const auto& state : states) {
		if (!a.getFinalStates().count(state.first)) {
			out << "node [shape = circle, label=\"" << alib::StringDataFactory::toString(state.first) << "\" ]; " << state.second << ";\n";
		}
	}

	//Mark initial states
	out << "node [shape = plaintext, label=\"start\"]; 0; \n";
	out << "0 -> " << states.find(a.getInitialState())->second << ";\n";

	transitions(a, states, out);
	out << "}";
}

auto DotConverterDPDA = DotConverter::RegistratorWrapper<void, automaton::DPDA < > >(DotConverter::convert);

void DotConverter::convert(std::ostream& out, const automaton::SinglePopDPDA < > & a) {
	out << "digraph automaton {\n";
	out << "rankdir=LR;\n";
	int cnt = 1;

	//Map states to indices
	std::map<label::Label, int> states;
	for (const label::Label& state : a.getStates()) {
		states.insert(std::make_pair(state, cnt++));
	}

	//Print final states
	for (const label::Label& state : a.getFinalStates()) {
		out << "node [shape = doublecircle, label=\"" << alib::StringDataFactory::toString(state) << "\"]; " << states.find(state)->second << ";\n";
	}

	//Print nonfinal states
	for (const auto& state : states) {
		if (!a.getFinalStates().count(state.first)) {
			out << "node [shape = circle, label=\"" << alib::StringDataFactory::toString(state.first) << "\" ]; " << state.second << ";\n";
		}
	}

	//Mark initial states
	out << "node [shape = plaintext, label=\"start\"]; 0; \n";
	out << "0 -> " << states.find(a.getInitialState())->second << ";\n";

	transitions(a, states, out);
	out << "}";
}

auto DotConverterSinglePopDPDA = DotConverter::RegistratorWrapper<void, automaton::SinglePopDPDA < > >(DotConverter::convert);

void DotConverter::convert(std::ostream& out, const automaton::InputDrivenDPDA < > & a) {
	out << "digraph automaton {\n";
	out << "rankdir=LR;\n";
	int cnt = 1;

	//Map states to indices
	std::map<label::Label, int> states;
	for (const label::Label& state : a.getStates()) {
		states.insert(std::make_pair(state, cnt++));
	}

	//Print final states
	for (const label::Label& state : a.getFinalStates()) {
		out << "node [shape = doublecircle, label=\"" << alib::StringDataFactory::toString(state) << "\"]; " << states.find(state)->second << ";\n";
	}

	//Print nonfinal states
	for (const auto& state : states) {
		if (!a.getFinalStates().count(state.first)) {
			out << "node [shape = circle, label=\"" << alib::StringDataFactory::toString(state.first) << "\" ]; " << state.second << ";\n";
		}
	}

	//Mark initial states
	out << "node [shape = plaintext, label=\"start\"]; 0; \n";
	out << "0 -> " << states.find(a.getInitialState())->second << ";\n";

	transitions(a, states, out);
	out << "}";
}

auto DotConverterInputDrivenDPDA = DotConverter::RegistratorWrapper<void, automaton::InputDrivenDPDA < > >(DotConverter::convert);

void DotConverter::convert(std::ostream& out, const automaton::InputDrivenNPDA < > & a) {
	out << "digraph automaton {\n";
	out << "rankdir=LR;\n";
	int cnt = 1;

	//Map states to indices
	std::map<label::Label, int> states;
	for (const label::Label& state : a.getStates()) {
		states.insert(std::make_pair(state, cnt++));
	}

	//Print final states
	for (const label::Label& state : a.getFinalStates()) {
		out << "node [shape = doublecircle, label=\"" << alib::StringDataFactory::toString(state) << "\"]; " << states.find(state)->second << ";\n";
	}

	//Print nonfinal states
	for (const auto& state : states) {
		if (!a.getFinalStates().count(state.first)) {
			out << "node [shape = circle, label=\"" << alib::StringDataFactory::toString(state.first) << "\" ]; " << state.second << ";\n";
		}
	}

	//Mark initial states
	out << "node [shape = plaintext, label=\"start\"]; 0; \n";
	out << "0 -> " << states.find(a.getInitialState())->second << ";\n";

	transitions(a, states, out);
	out << "}";
}

auto DotConverterInputDrivenNPDA = DotConverter::RegistratorWrapper<void, automaton::InputDrivenNPDA < > >(DotConverter::convert);

void DotConverter::convert(std::ostream& out, const automaton::VisiblyPushdownDPDA < > & a) {
	out << "digraph automaton {\n";
	out << "rankdir=LR;\n";
	int cnt = 1;

	//Map states to indices
	std::map<label::Label, int> states;
	for (const label::Label& state : a.getStates()) {
		states.insert(std::make_pair(state, cnt++));
	}

	//Print final states
	for (const label::Label& state : a.getFinalStates()) {
		out << "node [shape = doublecircle, label=\"" << alib::StringDataFactory::toString(state) << "\"]; " << states.find(state)->second << ";\n";
	}

	//Print nonfinal states
	for (const auto& state : states) {
		if (!a.getFinalStates().count(state.first)) {
			out << "node [shape = circle, label=\"" << alib::StringDataFactory::toString(state.first) << "\" ]; " << state.second << ";\n";
		}
	}

	//Mark initial states
	out << "node [shape = plaintext, label=\"start\"]; 0; \n";
	out << "0 -> " << states.find(a.getInitialState())->second << ";\n";

	transitions(a, states, out);
	out << "}";
}

auto DotConverterVisiblyPushdownDPDA = DotConverter::RegistratorWrapper<void, automaton::VisiblyPushdownDPDA < > >(DotConverter::convert);

void DotConverter::convert(std::ostream& out, const automaton::VisiblyPushdownNPDA < > & a) {
	out << "digraph automaton {\n";
	out << "rankdir=LR;\n";
	int cnt = 1;

	//Map states to indices
	std::map<label::Label, int> states;
	for (const label::Label& state : a.getStates()) {
		states.insert(std::make_pair(state, cnt++));
	}

	//Print final states
	for (const label::Label& state : a.getFinalStates()) {
		out << "node [shape = doublecircle, label=\"" << alib::StringDataFactory::toString(state) << "\"]; " << states.find(state)->second << ";\n";
	}

	//Print nonfinal states
	for (const auto& state : states) {
		if (!a.getFinalStates().count(state.first)) {
			out << "node [shape = circle, label=\"" << alib::StringDataFactory::toString(state.first) << "\" ]; " << state.second << ";\n";
		}
	}

	//Mark initial states
	out << "node [shape = plaintext, label=\"start\"]; 0; \n";
	for (const label::Label& state : a.getInitialStates()) {
		out << "0 -> " << states.find(state)->second << ";\n";
	}

	transitions(a, states, out);
	out << "}";
}

auto DotConverterVisiblyPushdownNPDA = DotConverter::RegistratorWrapper<void, automaton::VisiblyPushdownNPDA < > >(DotConverter::convert);

void DotConverter::convert(std::ostream& out, const automaton::RealTimeHeightDeterministicDPDA < > & a) {
	out << "digraph automaton {\n";
	out << "rankdir=LR;\n";
	int cnt = 1;

	//Map states to indices
	std::map<label::Label, int> states;
	for (const label::Label& state : a.getStates()) {
		states.insert(std::make_pair(state, cnt++));
	}

	//Print final states
	for (const label::Label& state : a.getFinalStates()) {
		out << "node [shape = doublecircle, label=\"" << alib::StringDataFactory::toString(state) << "\"]; " << states.find(state)->second << ";\n";
	}

	//Print nonfinal states
	for (const auto& state : states) {
		if (!a.getFinalStates().count(state.first)) {
			out << "node [shape = circle, label=\"" << alib::StringDataFactory::toString(state.first) << "\" ]; " << state.second << ";\n";
		}
	}

	//Mark initial states
	out << "node [shape = plaintext, label=\"start\"]; 0; \n";
	out << "0 -> " << states.find(a.getInitialState())->second << ";\n";

	transitions(a, states, out);
	out << "}";
}

auto DotConverterRealTimeHeightDeterministicDPDA = DotConverter::RegistratorWrapper<void, automaton::RealTimeHeightDeterministicDPDA < > >(DotConverter::convert);

void DotConverter::convert(std::ostream& out, const automaton::RealTimeHeightDeterministicNPDA < > & a) {
	out << "digraph automaton {\n";
	out << "rankdir=LR;\n";
	int cnt = 1;

	//Map states to indices
	std::map<label::Label, int> states;
	for (const label::Label& state : a.getStates()) {
		states.insert(std::make_pair(state, cnt++));
	}

	//Print final states
	for (const label::Label& state : a.getFinalStates()) {
		out << "node [shape = doublecircle, label=\"" << alib::StringDataFactory::toString(state) << "\"]; " << states.find(state)->second << ";\n";
	}

	//Print nonfinal states
	for (const auto& state : states) {
		if (!a.getFinalStates().count(state.first)) {
			out << "node [shape = circle, label=\"" << alib::StringDataFactory::toString(state.first) << "\" ]; " << state.second << ";\n";
		}
	}

	//Mark initial states
	out << "node [shape = plaintext, label=\"start\"]; 0; \n";
	for (const label::Label& state : a.getInitialStates()) {
		out << "0 -> " << states.find(state)->second << ";\n";
	}

	transitions(a, states, out);
	out << "}";
}

auto DotConverterRealTimeHeightDeterministicNPDA = DotConverter::RegistratorWrapper<void, automaton::RealTimeHeightDeterministicNPDA < > >(DotConverter::convert);

void DotConverter::convert(std::ostream& out, const automaton::NPDA < > & a) {
	out << "digraph automaton {\n";
	out << "rankdir=LR;\n";
	int cnt = 1;

	//Map states to indices
	std::map<label::Label, int> states;
	for (const label::Label& state : a.getStates()) {
		states.insert(std::make_pair(state, cnt++));
	}

	//Print final states
	for (const label::Label& state : a.getFinalStates()) {
		out << "node [shape = doublecircle, label=\"" << alib::StringDataFactory::toString(state) << "\"]; " << states.find(state)->second << ";\n";
	}

	//Print nonfinal states
	for (const auto& state : states) {
		if (!a.getFinalStates().count(state.first)) {
			out << "node [shape = circle, label=\"" << alib::StringDataFactory::toString(state.first) << "\" ]; " << state.second << ";\n";
		}
	}

	//Mark initial states
	out << "node [shape = plaintext, label=\"start\"]; 0; \n";
	out << "0 -> " << states.find(a.getInitialState())->second << ";\n";

	transitions(a, states, out);
	out << "}";
}

auto DotConverterNPDA = DotConverter::RegistratorWrapper<void, automaton::NPDA < > >(DotConverter::convert);

void DotConverter::convert(std::ostream& out, const automaton::SinglePopNPDA < > & a) {
	out << "digraph automaton {\n";
	out << "rankdir=LR;\n";
	int cnt = 1;

	//Map states to indices
	std::map<label::Label, int> states;
	for (const label::Label& state : a.getStates()) {
		states.insert(std::make_pair(state, cnt++));
	}

	//Print final states
	for (const label::Label& state : a.getFinalStates()) {
		out << "node [shape = doublecircle, label=\"" << alib::StringDataFactory::toString(state) << "\"]; " << states.find(state)->second << ";\n";
	}

	//Print nonfinal states
	for (const auto& state : states) {
		if (!a.getFinalStates().count(state.first)) {
			out << "node [shape = circle, label=\"" << alib::StringDataFactory::toString(state.first) << "\" ]; " << state.second << ";\n";
		}
	}

	//Mark initial states
	out << "node [shape = plaintext, label=\"start\"]; 0; \n";
	out << "0 -> " << states.find(a.getInitialState())->second << ";\n";

	transitions(a, states, out);
	out << "}";
}

auto DotConverterSinglePopNPDA = DotConverter::RegistratorWrapper<void, automaton::SinglePopNPDA < > >(DotConverter::convert);

void DotConverter::convert(std::ostream& out, const automaton::OneTapeDTM<>& a) {
	out << "digraph automaton {\n";
	out << "rankdir=LR;\n";
	int cnt = 1;

	//Map states to indices
	std::map<label::Label, int> states;
	for (const label::Label& state : a.getStates()) {
		states.insert(std::make_pair(state, cnt++));
	}

	//Print final states
	for (const label::Label& state : a.getFinalStates()) {
		out << "node [shape = doublecircle, label=\"" << alib::StringDataFactory::toString(state) << "\"]; " << states.find(state)->second << ";\n";
	}

	//Print nonfinal states
	for (const auto& state : states) {
		if (!a.getFinalStates().count(state.first)) {
			out << "node [shape = circle, label=\"" << alib::StringDataFactory::toString(state.first) << "\" ]; " << state.second << ";\n";
		}
	}

	//Mark initial states
	out << "node [shape = plaintext, label=\"start\"]; 0; \n";
	out << "0 -> " << states.find(a.getInitialState())->second << ";\n";

	transitions(a, states, out);
	out << "}";
}

auto DotConverterOneTapeDTM = DotConverter::RegistratorWrapper<void, automaton::OneTapeDTM<>>(DotConverter::convert);

void DotConverter::transitions(const automaton::EpsilonNFA < > & fsm, const std::map<label::Label, int>& states, std::ostream& out) {
	std::map<std::pair<int, int>, std::string> transitions;

	//put transitions from automaton to "transitions"
	for (const auto& transition : fsm.getTransitions()) {
		std::string symbol;
		if (transition.first.second.is<string::Epsilon < >>()) {
			symbol = "&epsilon;";
		} else {
			symbol = alib::StringDataFactory::toString(std::get<1>(transition.first).get<alphabet::Symbol>());
		}

		for(const label::Label& to : transition.second) {
			std::pair<int, int> key(states.find(transition.first.first)->second, states.find(to)->second);
			std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

			if (mapit == transitions.end()) {
				transitions.insert(std::make_pair(key, symbol));
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of("\n");
				if(pos == std::string::npos) pos = 0;
				if(mapit->second.size() - pos > 100) mapit->second += "\n";
				else mapit->second += " ";

				mapit->second += symbol;
			}
		}
	}

	//print the map
	for (std::pair<const std::pair<int, int>, std::string>& transition : transitions) {
		out << transition.first.first << " -> " << transition.first.second;
		replace(transition.second, "\n", "\\n");
		out << "[label=\"" << transition.second << "\"]\n";
	}
}

void DotConverter::transitions(const automaton::MultiInitialStateNFA < >& fsm, const std::map<label::Label, int>& states, std::ostream& out) {
	std::map<std::pair<int, int>, std::string> transitions;

	//put transitions from automaton to "transitions"
	for (const auto& transition : fsm.getTransitions()) {
		std::string symbol = alib::StringDataFactory::toString(transition.first.second);

		for(const label::Label& to : transition.second) {
			std::pair<int, int> key(states.find(transition.first.first)->second, states.find(to)->second);
			std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

			if (mapit == transitions.end()) {
				transitions.insert(std::make_pair(key, symbol));
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of("\n");
				if(pos == std::string::npos) pos = 0;
				if(mapit->second.size() - pos > 100) mapit->second += "\n";
				else mapit->second += " ";

				mapit->second += symbol;
			}
		}
	}

	//print the map
	for (std::pair< const std::pair<int, int>, std::string>& transition : transitions) {
		out << transition.first.first << " -> " << transition.first.second;
		replace(transition.second, "\n", "\\n");
		out << "[label=\"" << transition.second << "\"]\n";
	}
}

void DotConverter::transitions(const automaton::NFA < > & fsm, const std::map<label::Label, int>& states, std::ostream& out) {
	std::map<std::pair<int, int>, std::string> transitions;

	//put transitions from automaton to "transitions"
	for (const auto& transition : fsm.getTransitions()) {
		std::string symbol = alib::StringDataFactory::toString(transition.first.second);

		for(const label::Label& to : transition.second) {
			std::pair<int, int> key(states.find(transition.first.first)->second, states.find(to)->second);
			std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

			if (mapit == transitions.end()) {
				transitions.insert(std::make_pair(key, symbol));
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of("\n");
				if(pos == std::string::npos) pos = 0;
				if(mapit->second.size() - pos > 100) mapit->second += "\n";
				else mapit->second += " ";

				mapit->second += symbol;
			}
		}
	}

	//print the map
	for (std::pair<const std::pair<int, int>, std::string>& transition : transitions) {
		out << transition.first.first << " -> " << transition.first.second;
		replace(transition.second, "\n", "\\n");
		out << "[label=\"" << transition.second << "\"]\n";
	}
}

void DotConverter::transitions(const automaton::DFA<>& fsm, const std::map<label::Label, int>& states, std::ostream& out) {
	std::map<std::pair<int, int>, std::string> transitions;

	//put transitions from automaton to "transitions"
	for (const auto& transition : fsm.getTransitions()) {
		std::string symbol = alib::StringDataFactory::toString(transition.first.second);

		std::pair<int, int> key(states.find(transition.first.first)->second, states.find(transition.second)->second);
		std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

		if (mapit == transitions.end()) {
			transitions.insert(std::make_pair(key, symbol));
		} else {
			mapit->second += ",";

			size_t pos = mapit->second.find_last_of("\n");
			if(pos == std::string::npos) pos = 0;
			if(mapit->second.size() - pos > 100) mapit->second += "\n";
			else mapit->second += " ";

			mapit->second += symbol;
		}
	}

	//print the map
	for (std::pair<const std::pair<int, int>, std::string>& transition : transitions) {
		out << transition.first.first << " -> " << transition.first.second;
		replace(transition.second, "\n", "\\n");
		out << "[label=\"" << transition.second << "\"]\n";
	}
}

void DotConverter::transitions(const automaton::ExtendedNFA& fsm, const std::map<label::Label, int>& states, std::ostream& out) {
	std::map<std::pair<int, int>, std::string> transitions;

	//put transitions from automaton to "transitions"
	for (const auto& transition : fsm.getTransitions()) {
		std::string symbol = alib::StringDataFactory::toString(transition.first.second);

		for(const label::Label& to : transition.second) {
			std::pair<int, int> key(states.find(transition.first.first)->second, states.find(to)->second);
			std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

			if (mapit == transitions.end()) {
				transitions.insert(std::make_pair(key, symbol));
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of("\n");
				if(pos == std::string::npos) pos = 0;
				if(mapit->second.size() - pos > 100) mapit->second += "\n";
				else mapit->second += " ";

				mapit->second += symbol;
			}
		}
	}

	//print the map
	for (std::pair<const std::pair<int, int>, std::string>& transition : transitions) {
		out << transition.first.first << " -> " << transition.first.second;
		replace(transition.second, "\n", "\\n");
		out << "[label=\"" << transition.second << "\"]\n";
	}
}

void DotConverter::transitions(const automaton::CompactNFA < > & fsm, const std::map<label::Label, int>& states, std::ostream& out) {
	std::map<std::pair<int, int>, std::string> transitions;

	//put transitions from automaton to "transitions"
	for (const auto& transition : fsm.getTransitions()) {
		std::string symbol = alib::StringDataFactory::toString(string::stringFrom ( transition.first.second ) );

		for(const label::Label& to : transition.second) {
			std::pair<int, int> key(states.find(transition.first.first)->second, states.find(to)->second);
			std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

			if (mapit == transitions.end()) {
				transitions.insert(std::make_pair(key, symbol));
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of("\n");
				if(pos == std::string::npos) pos = 0;
				if(mapit->second.size() - pos > 100) mapit->second += "\n";
				else mapit->second += " ";

				mapit->second += symbol;
			}
		}
	}

	//print the map
	for (std::pair<const std::pair<int, int>, std::string>& transition : transitions) {
		out << transition.first.first << " -> " << transition.first.second;
		replace(transition.second, "\n", "\\n");
		out << "[label=\"" << transition.second << "\"]\n";
	}
}

void DotConverter::transitions(const automaton::NFTA& fta, const std::map<label::Label, int>& states, std::ostream& out) {
	std::map<std::pair<int, std::vector<int>>, std::string> transitions;

	//put transitions from automaton to "transitions"
	for (const auto& transition : fta.getTransitions()) {
		std::string symbol = alib::StringDataFactory::toString(transition.first.first.getSymbol());
		symbol += std::to_string(transition.first.first.getRank().getData());

		for(const label::Label& to : transition.second) {
			std::pair<int, std::vector<int>> key(states.find(to)->second, {});
			for(const label::Label& state : transition.first.second) {
				key.second.push_back(states.find(state)->second);
			}
			std::map<std::pair<int, std::vector<int>>, std::string>::iterator mapit = transitions.find(key);

			if (mapit == transitions.end()) {
				transitions.insert(std::make_pair(key, symbol));
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of("\n");
				if(pos == std::string::npos) pos = 0;
				if(mapit->second.size() - pos > 100) mapit->second += "\n";
				else mapit->second += " ";

				mapit->second += symbol;
			}
		}
	}

	//Print auxilary dots
	for (unsigned i = 1; i < transitions.size(); i++) {
		out << "node [shape = point, label=\"\"]; " << states.size() + i << ";\n";
	}

	//print the map
	unsigned i = states.size() + 1;
	for (std::pair<const std::pair<int, std::vector<int>>, std::string>& transition : transitions) {
		out << i << " -> " << transition.first.first;
		replace(transition.second, "\n", "\\n");
		out << "[label=\"" << transition.second << "\"]\n";
		unsigned j = 0;
		for(int from : transition.first.second) {
			out << from << " -> " << i;
			out << "[label=\"" << j << "\"]\n";
			j++;
		}
		i++;
	}
}

void DotConverter::transitions(const automaton::DFTA& fta, const std::map<label::Label, int>& states, std::ostream& out) {
	std::map<std::pair<int, std::vector<int>>, std::string> transitions;

	//put transitions from automaton to "transitions"
	for (const auto& transition : fta.getTransitions()) {
		std::string symbol = alib::StringDataFactory::toString(transition.first.first.getSymbol());
		symbol += std::to_string(transition.first.first.getRank().getData());

		std::pair<int, std::vector<int>> key(states.find(transition.second)->second, {});
		for(const label::Label& state : transition.first.second) {
			key.second.push_back(states.find(state)->second);
		}
		std::map<std::pair<int, std::vector<int>>, std::string>::iterator mapit = transitions.find(key);

		if (mapit == transitions.end()) {
			transitions.insert(std::make_pair(key, symbol));
		} else {
			mapit->second += ",";

			size_t pos = mapit->second.find_last_of("\n");
			if(pos == std::string::npos) pos = 0;
			if(mapit->second.size() - pos > 100) mapit->second += "\n";
			else mapit->second += " ";

			mapit->second += symbol;
		}
	}

	//Print auxilary dots
	for (unsigned i = 1; i < transitions.size(); i++) {
		out << "node [shape = point, label=\"\"]; " << states.size() + i << ";\n";
	}

	//print the map
	unsigned i = states.size() + 1;
	for (std::pair<const std::pair<int, std::vector<int>>, std::string>& transition : transitions) {
		out << i << " -> " << transition.first.first;
		replace(transition.second, "\n", "\\n");
		out << "[label=\"" << transition.second << "\"]\n";
		unsigned j = 0;
		for(int from : transition.first.second) {
			out << from << " -> " << i;
			out << "[label=\"" << j << "\"]\n";
			j++;
		}
		i++;
	}
}

void DotConverter::transitions(const automaton::DPDA < > & pda, const std::map<label::Label, int>& states, std::ostream& out) {
	std::map<std::pair<int, int>, std::string> transitions;

	for (const auto& transition : pda.getTransitions()) {
		std::string symbol;

		//input symbol
		if (std::get<1>(transition.first).is<string::Epsilon < >>()) {
			symbol = "&epsilon;";
		} else {
			symbol = alib::StringDataFactory::toString(std::get<1>(transition.first).get<alphabet::Symbol>());
		}

		symbol += " |";

		//Pop part
		if (std::get<2>(transition.first).size() == 0) {
			symbol += " &epsilon;";
		} else {
			for (alphabet::Symbol symb : std::get<2>(transition.first)) {
				symbol += " " + alib::StringDataFactory::toString(symb);
			}

		}

		symbol += " ->";

		//Push part
		if (transition.second.second.size() == 0) {
			symbol += " &epsilon;";
		} else {
			for (alphabet::Symbol symb : transition.second.second) {
				symbol += " " + alib::StringDataFactory::toString(symb);
			}

		}

		//Insert into map
		std::pair<int, int> key(states.find(std::get<0>(transition.first))->second, states.find(transition.second.first)->second);
		std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

		if (mapit == transitions.end()) {
			transitions.insert(std::make_pair(key, symbol));
		} else {
			mapit->second += ",";
			
			size_t pos = mapit->second.find_last_of("\n");
			if(pos == std::string::npos) pos = 0;
			if(mapit->second.size() - pos > 100) mapit->second += "\n";
			else mapit->second += " ";

			mapit->second += symbol;
		}
	}

	//print the map
	for (std::pair<const std::pair<int, int>, std::string>& transition : transitions) {
		out << transition.first.first << " -> " << transition.first.second;
		replace(transition.second, "\n", "\\n");
		out << "[label=\"" <<  transition.second << "\"]\n";
	}
}

void DotConverter::transitions(const automaton::SinglePopDPDA < > & pda, const std::map<label::Label, int>& states, std::ostream& out) {
	std::map<std::pair<int, int>, std::string> transitions;

	for (const auto& transition : pda.getTransitions()) {
		std::string symbol;

		//input symbol
		if (std::get<1>(transition.first).is<string::Epsilon < >>()) {
			symbol = "&epsilon;";
		} else {
			symbol = alib::StringDataFactory::toString(std::get<1>(transition.first).get<alphabet::Symbol>());
		}

		symbol += " |";

		//Pop part
		symbol += " " + alib::StringDataFactory::toString(std::get<2>(transition.first));

		symbol += " ->";

		//Push part
		if (transition.second.second.size() == 0) {
			symbol += " &epsilon;";
		} else {
			for (alphabet::Symbol symb : transition.second.second) {
				symbol += " " + alib::StringDataFactory::toString(symb);
			}

		}

		//Insert into map
		std::pair<int, int> key(states.find(std::get<0>(transition.first))->second, states.find(transition.second.first)->second);
		std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

		if (mapit == transitions.end()) {
			transitions.insert(std::make_pair(key, symbol));
		} else {
			mapit->second += ",";

			size_t pos = mapit->second.find_last_of("\n");
			if(pos == std::string::npos) pos = 0;
			if(mapit->second.size() - pos > 100) mapit->second += "\n";
			else mapit->second += " ";

			mapit->second += symbol;
		}
	}

	//print the map
	for (std::pair<const std::pair<int, int>, std::string>& transition : transitions) {
		out << transition.first.first << " -> " << transition.first.second;
		replace(transition.second, "\n", "\\n");
		out << "[label=\"" << transition.second << "\"]\n";
	}
}

void DotConverter::transitions(const automaton::InputDrivenDPDA < > & pda, const std::map<label::Label, int>& states, std::ostream& out) {
	std::map<std::pair<int, int>, std::string> transitions;

	const auto& symbolToPDSOperation = pda.getPushdownStoreOperations();
	for (const auto& transition : pda.getTransitions()) {
		const auto& pop = symbolToPDSOperation.find(transition.first.second)->second.first;
		const auto& push = symbolToPDSOperation.find(transition.first.second)->second.second;

		std::string symbol;

		//input symbol
		symbol = alib::StringDataFactory::toString(transition.first.second);

		symbol += " |";

		//Pop part
		if (pop.size() == 0) {
			symbol += " &epsilon;";
		} else {
			for (alphabet::Symbol symb : pop) {
				symbol += " " + alib::StringDataFactory::toString(symb);
			}

		}

		symbol += " ->";

		const auto& to = transition.second;
		//Push part
		if (push.size() == 0) {
			symbol += " &epsilon;";
		} else {
			for (alphabet::Symbol symb : push) {
				symbol += " " + alib::StringDataFactory::toString(symb);
			}

		}

		//Insert into map
		std::pair<int, int> key(states.find(transition.first.first)->second, states.find(to)->second);
		std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

		if (mapit == transitions.end()) {
			transitions.insert(std::make_pair(key, symbol));
		} else {
			mapit->second += ",";

			size_t pos = mapit->second.find_last_of("\n");
			if(pos == std::string::npos) pos = 0;
			if(mapit->second.size() - pos > 100) mapit->second += "\n";
			else mapit->second += " ";

			mapit->second += symbol;
		}
	}

	//print the map
	for (std::pair<const std::pair<int, int>, std::string>& transition : transitions) {
		out << transition.first.first << " -> " << transition.first.second;
		replace(transition.second, "\n", "\\n");
		out << "[label=\"" << transition.second << "\"]\n";
	}
}

void DotConverter::transitions(const automaton::InputDrivenNPDA < > & pda, const std::map<label::Label, int>& states, std::ostream& out) {
	std::map<std::pair<int, int>, std::string> transitions;

	const auto& symbolToPDSOperation = pda.getPushdownStoreOperations();
	for (const auto& transition : pda.getTransitions()) {
		const auto& pop = symbolToPDSOperation.find(transition.first.second)->second.first;
		const auto& push = symbolToPDSOperation.find(transition.first.second)->second.second;

		std::string symbol;

		//input symbol
		symbol = alib::StringDataFactory::toString(transition.first.second);

		symbol += " |";

		//Pop part
		if (pop.size() == 0) {
			symbol += " &epsilon;";
		} else {
			for (alphabet::Symbol symb : pop) {
				symbol += " " + alib::StringDataFactory::toString(symb);
			}

		}

		symbol += " ->";

		std::string symbol2;
		for(const auto& to : transition.second) {
			symbol2 = symbol;
			//Push part
			if (push.size() == 0) {
				symbol2 += " &epsilon;";
			} else {
				for (alphabet::Symbol symb : push) {
					symbol2 += " " + alib::StringDataFactory::toString(symb);
				}

			}

			//Insert into map
			std::pair<int, int> key(states.find(transition.first.first)->second, states.find(to)->second);
			std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

			if (mapit == transitions.end()) {
				transitions.insert(std::make_pair(key, symbol2));
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of("\n");
				if(pos == std::string::npos) pos = 0;
				if(mapit->second.size() - pos > 100) mapit->second += "\n";
				else mapit->second += " ";

				mapit->second += symbol2;
			}
		}
	}

	//print the map
	for (std::pair<const std::pair<int, int>, std::string>& transition : transitions) {
		out << transition.first.first << " -> " << transition.first.second;
		replace(transition.second, "\n", "\\n");
		out << "[label=\"" << transition.second << "\"]\n";
	}
}

void DotConverter::transitions(const automaton::VisiblyPushdownDPDA < > & pda, const std::map<label::Label, int>& states, std::ostream& out) {
	std::map<std::pair<int, int>, std::string> transitions;

	for (const auto& transition : pda.getCallTransitions()) {
		std::string symbol;

		//input symbol
		symbol = alib::StringDataFactory::toString(transition.first.second);

		symbol += " |";

		//Pop part
		symbol += " &epsilon;";
		symbol += " ->";

		symbol += " " + alib::StringDataFactory::toString(transition.second.second);

		//Insert into map
		std::pair<int, int> key(states.find(transition.first.first)->second, states.find(transition.second.first)->second);
		std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

		if (mapit == transitions.end()) {
			transitions.insert(std::make_pair(key, symbol));
		} else {
			mapit->second += ",";

			size_t pos = mapit->second.find_last_of("\n");
			if(pos == std::string::npos) pos = 0;
			if(mapit->second.size() - pos > 100) mapit->second += "\n";
			else mapit->second += " ";

			mapit->second += symbol;
		}
	}

	for (const auto& transition : pda.getReturnTransitions()) {
		std::string symbol;

		//input symbol
		symbol = alib::StringDataFactory::toString(std::get<1>(transition.first));

		symbol += " |";

		//Pop part
		symbol += " " + alib::StringDataFactory::toString(std::get<2>(transition.first));
		symbol += " ->";

		symbol += " &epsilon;";

		//Insert into map
		std::pair<int, int> key(states.find(std::get<0>(transition.first))->second, states.find(transition.second)->second);
		std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

		if (mapit == transitions.end()) {
			transitions.insert(std::make_pair(key, symbol));
		} else {
			mapit->second += ",";

			size_t pos = mapit->second.find_last_of("\n");
			if(pos == std::string::npos) pos = 0;
			if(mapit->second.size() - pos > 100) mapit->second += "\n";
			else mapit->second += " ";

			mapit->second += symbol;
		}
	}

	for (const auto& transition : pda.getLocalTransitions()) {
		std::string symbol;

		//input symbol
		symbol = alib::StringDataFactory::toString(transition.first.second);

		symbol += " |";

		//Pop part
		symbol += " &epsilon;";
		symbol += " ->";

		symbol += " &epsilon;";

		//Insert into map
		std::pair<int, int> key(states.find(transition.first.first)->second, states.find(transition.second)->second);
		std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

		if (mapit == transitions.end()) {
			transitions.insert(std::make_pair(key, symbol));
		} else {
			mapit->second += ",";

			size_t pos = mapit->second.find_last_of("\n");
			if(pos == std::string::npos) pos = 0;
			if(mapit->second.size() - pos > 100) mapit->second += "\n";
			else mapit->second += " ";

			mapit->second += symbol;
		}
	}

	//print the map
	for (std::pair<const std::pair<int, int>, std::string>& transition : transitions) {
		out << transition.first.first << " -> " << transition.first.second;
		replace(transition.second, "\n", "\\n");
		out << "[label=\"" << transition.second << "\"]\n";
	}
}

void DotConverter::transitions(const automaton::VisiblyPushdownNPDA < > & pda, const std::map<label::Label, int>& states, std::ostream& out) {
	std::map<std::pair<int, int>, std::string> transitions;

	for (const auto& transition : pda.getCallTransitions()) {
		std::string symbol;

		//input symbol
		symbol = alib::StringDataFactory::toString(transition.first.second);

		symbol += " |";

		//Pop part
		symbol += " &epsilon;";
		symbol += " ->";

		std::string symbol2;
		for(const auto& to : transition.second) {
			symbol2 = symbol;
			symbol2 += " " + alib::StringDataFactory::toString(to.second);

			//Insert into map
			std::pair<int, int> key(states.find(transition.first.first)->second, states.find(to.first)->second);
			std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

			if (mapit == transitions.end()) {
				transitions.insert(std::make_pair(key, symbol2));
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of("\n");
				if(pos == std::string::npos) pos = 0;
				if(mapit->second.size() - pos > 100) mapit->second += "\n";
				else mapit->second += " ";

				mapit->second += symbol2;
			}
		}
	}

	for (const auto& transition : pda.getReturnTransitions()) {
		std::string symbol;

		//input symbol
		symbol = alib::StringDataFactory::toString(std::get<1>(transition.first));

		symbol += " |";

		//Pop part
		symbol += " " + alib::StringDataFactory::toString(std::get<2>(transition.first));
		symbol += " ->";

		std::string symbol2;
		for(const auto& to : transition.second) {
			symbol2 = symbol;
			symbol2 += " &epsilon;";

			//Insert into map
			std::pair<int, int> key(states.find(std::get<0>(transition.first))->second, states.find(to)->second);
			std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

			if (mapit == transitions.end()) {
				transitions.insert(std::make_pair(key, symbol2));
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of("\n");
				if(pos == std::string::npos) pos = 0;
				if(mapit->second.size() - pos > 100) mapit->second += "\n";
				else mapit->second += " ";

				mapit->second += symbol2;
			}
		}
	}

	for (const auto& transition : pda.getLocalTransitions()) {
		std::string symbol;

		//input symbol
		symbol = alib::StringDataFactory::toString(transition.first.second);

		symbol += " |";

		//Pop part
		symbol += " &epsilon;";
		symbol += " ->";

		std::string symbol2;
		for(const auto& to : transition.second) {
			symbol2 = symbol;
			symbol2 += " &epsilon;";

			//Insert into map
			std::pair<int, int> key(states.find(transition.first.first)->second, states.find(to)->second);
			std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

			if (mapit == transitions.end()) {
				transitions.insert(std::make_pair(key, symbol2));
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of("\n");
				if(pos == std::string::npos) pos = 0;
				if(mapit->second.size() - pos > 100) mapit->second += "\n";
				else mapit->second += " ";

				mapit->second += symbol2;
			}
		}
	}

	//print the map
	for (std::pair<const std::pair<int, int>, std::string>& transition : transitions) {
		out << transition.first.first << " -> " << transition.first.second;
		replace(transition.second, "\n", "\\n");
		out << "[label=\"" << transition.second << "\"]\n";
	}
}

void DotConverter::transitions(const automaton::RealTimeHeightDeterministicDPDA < > & pda, const std::map<label::Label, int>& states, std::ostream& out) {
	std::map<std::pair<int, int>, std::string> transitions;

	for (const auto& transition : pda.getCallTransitions()) {
		std::string symbol;

		//input symbol
		if(transition.first.second.is<string::Epsilon < >>())
			symbol = "&epsilon";
		else
			symbol = alib::StringDataFactory::toString(transition.first.second.get<alphabet::Symbol>());

		symbol += " |";

		//Pop part
		symbol += " &epsilon;";
		symbol += " ->";

		symbol += " " + alib::StringDataFactory::toString(transition.second.second);

		//Insert into map
		std::pair<int, int> key(states.find(transition.first.first)->second, states.find(transition.second.first)->second);
		std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

		if (mapit == transitions.end()) {
			transitions.insert(std::make_pair(key, symbol));
		} else {
			mapit->second += ",";

			size_t pos = mapit->second.find_last_of("\n");
			if(pos == std::string::npos) pos = 0;
			if(mapit->second.size() - pos > 100) mapit->second += "\n";
			else mapit->second += " ";

			mapit->second += symbol;
		}
	}

	for (const auto& transition : pda.getReturnTransitions()) {
		std::string symbol;

		//input symbol
		if(std::get<1>(transition.first).is<string::Epsilon < >>())
			symbol = "&epsilon";
		else
			symbol = alib::StringDataFactory::toString(std::get<1>(transition.first).get<alphabet::Symbol>());

		symbol += " |";

		//Pop part
		symbol += " " + alib::StringDataFactory::toString(std::get<2>(transition.first));
		symbol += " ->";

		symbol += " &epsilon;";

		//Insert into map
		std::pair<int, int> key(states.find(std::get<0>(transition.first))->second, states.find(transition.second)->second);
		std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

		if (mapit == transitions.end()) {
			transitions.insert(std::make_pair(key, symbol));
		} else {
			mapit->second += ",";

			size_t pos = mapit->second.find_last_of("\n");
			if(pos == std::string::npos) pos = 0;
			if(mapit->second.size() - pos > 100) mapit->second += "\n";
			else mapit->second += " ";

			mapit->second += symbol;
		}
	}

	for (const auto& transition : pda.getLocalTransitions()) {
		std::string symbol;

		//input symbol
		if(transition.first.second.is<string::Epsilon < >>())
			symbol = "&epsilon";
		else
			symbol = alib::StringDataFactory::toString(transition.first.second.get<alphabet::Symbol>());

		symbol += " |";

		//Pop part
		symbol += " &epsilon;";
		symbol += " ->";

		symbol += " &epsilon;";

		//Insert into map
		std::pair<int, int> key(states.find(transition.first.first)->second, states.find(transition.second)->second);
		std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

		if (mapit == transitions.end()) {
			transitions.insert(std::make_pair(key, symbol));
		} else {
			mapit->second += ",";

			size_t pos = mapit->second.find_last_of("\n");
			if(pos == std::string::npos) pos = 0;
			if(mapit->second.size() - pos > 100) mapit->second += "\n";
			else mapit->second += " ";

			mapit->second += symbol;
		}
	}

	//print the map
	for (std::pair<const std::pair<int, int>, std::string>& transition : transitions) {
		out << transition.first.first << " -> " << transition.first.second;
		replace(transition.second, "\n", "\\n");
		out << "[label=\"" << transition.second << "\"]\n";
	}
}

void DotConverter::transitions(const automaton::RealTimeHeightDeterministicNPDA < > & pda, const std::map<label::Label, int>& states, std::ostream& out) {
	std::map<std::pair<int, int>, std::string> transitions;

	for (const auto& transition : pda.getCallTransitions()) {
		std::string symbol;

		//input symbol
		if(transition.first.second.is<string::Epsilon < >>())
			symbol = "&epsilon";
		else
			symbol = alib::StringDataFactory::toString(transition.first.second.get<alphabet::Symbol>());

		symbol += " |";

		//Pop part
		symbol += " &epsilon;";
		symbol += " ->";

		std::string symbol2;
		for(const auto& to : transition.second) {
			symbol2 = symbol;
			symbol2 += " " + alib::StringDataFactory::toString(to.second);

			//Insert into map
			std::pair<int, int> key(states.find(transition.first.first)->second, states.find(to.first)->second);
			std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

			if (mapit == transitions.end()) {
				transitions.insert(std::make_pair(key, symbol2));
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of("\n");
				if(pos == std::string::npos) pos = 0;
				if(mapit->second.size() - pos > 100) mapit->second += "\n";
				else mapit->second += " ";

				mapit->second += symbol2;
			}
		}
	}

	for (const auto& transition : pda.getReturnTransitions()) {
		std::string symbol;

		//input symbol
		if(std::get<1>(transition.first).is<string::Epsilon < >>())
			symbol = "&epsilon;";
		else
			symbol = alib::StringDataFactory::toString(std::get<1>(transition.first).get<alphabet::Symbol>());

		symbol += " |";

		//Pop part
		symbol += " " + alib::StringDataFactory::toString(std::get<2>(transition.first));
		symbol += " ->";

		std::string symbol2;
		for(const auto& to : transition.second) {
			symbol2 = symbol;
			symbol2 += " &epsilon;";

			//Insert into map
			std::pair<int, int> key(states.find(std::get<0>(transition.first))->second, states.find(to)->second);
			std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

			if (mapit == transitions.end()) {
				transitions.insert(std::make_pair(key, symbol2));
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of("\n");
				if(pos == std::string::npos) pos = 0;
				if(mapit->second.size() - pos > 100) mapit->second += "\n";
				else mapit->second += " ";

				mapit->second += symbol2;
			}
		}
	}

	for (const auto& transition : pda.getLocalTransitions()) {
		std::string symbol;

		//input symbol
		if(transition.first.second.is<string::Epsilon < >>())
			symbol = "&epsilon;";
		else
			symbol = alib::StringDataFactory::toString(transition.first.second.get<alphabet::Symbol>());

		symbol += " |";

		//Pop part
		symbol += " &epsilon;";
		symbol += " ->";

		std::string symbol2;
		for(const auto& to : transition.second) {
			symbol2 = symbol;
			symbol2 += " &epsilon;";

			//Insert into map
			std::pair<int, int> key(states.find(transition.first.first)->second, states.find(to)->second);
			std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

			if (mapit == transitions.end()) {
				transitions.insert(std::make_pair(key, symbol2));
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of("\n");
				if(pos == std::string::npos) pos = 0;
				if(mapit->second.size() - pos > 100) mapit->second += "\n";
				else mapit->second += " ";

				mapit->second += symbol2;
			}
		}
	}

	//print the map
	for (std::pair<const std::pair<int, int>, std::string>& transition : transitions) {
		out << transition.first.first << " -> " << transition.first.second;
		replace(transition.second, "\n", "\\n");
		out << "[label=\"" << transition.second << "\"]\n";
	}
}

void DotConverter::transitions(const automaton::NPDA < > & pda, const std::map<label::Label, int>& states, std::ostream& out) {
	std::map<std::pair<int, int>, std::string> transitions;

	for (const auto& transition : pda.getTransitions()) {
		std::string symbol;

		//input symbol
		if (std::get<1>(transition.first).is<string::Epsilon < >>()) {
			symbol = "&epsilon;";
		} else {
			symbol = alib::StringDataFactory::toString(std::get<1>(transition.first).get<alphabet::Symbol>());
		}

		symbol += " |";

		//Pop part
		if (std::get<2>(transition.first).size() == 0) {
			symbol += " &epsilon;";
		} else {
			for (alphabet::Symbol symb : std::get<2>(transition.first)) {
				symbol += " " + alib::StringDataFactory::toString(symb);
			}

		}

		symbol += " ->";

		std::string symbol2;
		for(const auto& to : transition.second) {
			symbol2 = symbol;
			//Push part
			if (to.second.size() == 0) {
				symbol2 += " &epsilon;";
			} else {
				for (alphabet::Symbol symb : to.second) {
					symbol2 += " " + alib::StringDataFactory::toString(symb);
				}

			}

			//Insert into map
			std::pair<int, int> key(states.find(std::get<0>(transition.first))->second, states.find(to.first)->second);
			std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

			if (mapit == transitions.end()) {
				transitions.insert(std::make_pair(key, symbol2));
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of("\n");
				if(pos == std::string::npos) pos = 0;
				if(mapit->second.size() - pos > 100) mapit->second += "\n";
				else mapit->second += " ";

				mapit->second += symbol2;
			}
		}
	}

	//print the map
	for (std::pair<const std::pair<int, int>, std::string>& transition : transitions) {
		out << transition.first.first << " -> " << transition.first.second;
		replace(transition.second, "\n", "\\n");
		out << "[label=\"" << transition.second << "\"]\n";
	}
}

void DotConverter::transitions(const automaton::SinglePopNPDA < > & pda, const std::map<label::Label, int>& states, std::ostream& out) {
	std::map<std::pair<int, int>, std::string> transitions;

	for (const auto& transition : pda.getTransitions()) {
		std::string symbol;

		//input symbol
		if (std::get<1>(transition.first).is<string::Epsilon < >>()) {
			symbol = "&epsilon;";
		} else {
			symbol = alib::StringDataFactory::toString(std::get<1>(transition.first).get<alphabet::Symbol>());
		}

		symbol += " |";

		//Pop part
		symbol += " " + alib::StringDataFactory::toString(std::get<2>(transition.first));

		symbol += " ->";

		std::string symbol2;
		for(const auto& to : transition.second) {
			symbol2 = symbol;
			//Push part
			if (to.second.size() == 0) {
				symbol2 += " &epsilon;";
			} else {
				for (alphabet::Symbol symb : to.second) {
					symbol2 += " " + alib::StringDataFactory::toString(symb);
				}

			}

			//Insert into map
			std::pair<int, int> key(states.find(std::get<0>(transition.first))->second, states.find(to.first)->second);
			std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

			if (mapit == transitions.end()) {
				transitions.insert(std::make_pair(key, symbol2));
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of("\n");
				if(pos == std::string::npos) pos = 0;
				if(mapit->second.size() - pos > 100) mapit->second += "\n";
				else mapit->second += " ";

				mapit->second += symbol2;
			}
		}
	}

	//print the map
	for (std::pair<const std::pair<int, int>, std::string>& transition : transitions) {
		out << transition.first.first << " -> " << transition.first.second;
		replace(transition.second, "\n", "\\n");
		out << "[label=\"" << transition.second << "\"]\n";
	}
}

void DotConverter::transitions(const automaton::OneTapeDTM<>& tm, const std::map<label::Label, int>& states, std::ostream& out) {
	std::map<std::pair<int, int>, std::string> transitions;
	for (const auto& transition : tm.getTransitions()) {
		std::string symbol;

		//input symbol
		symbol = "(";
		symbol += alib::StringDataFactory::toString(transition.first.second);
		symbol += ", ";
		symbol += alib::StringDataFactory::toString(std::get<1>(transition.second));
		symbol += " ";
		switch(std::get<2>(transition.second)) {
		case automaton::Shift::LEFT:
			symbol += "&larr;";
			break;
		case automaton::Shift::RIGHT:
			symbol += "&rarr;";
			break;
		case automaton::Shift::NONE:
			symbol += "&times;";
			break;
		default:
			throw exception::CommonException("Unexpected shift direction");
		}

		//Insert into map
		std::pair<int, int> key(states.find(transition.first.first)->second, states.find(std::get<0>(transition.second))->second);
		std::map<std::pair<int, int>, std::string>::iterator mapit = transitions.find(key);

		if (mapit == transitions.end()) {
			transitions.insert(std::make_pair(key, symbol));
		} else {
			mapit->second += ",";

			size_t pos = mapit->second.find_last_of("\n");
			if(pos == std::string::npos) pos = 0;
			if(mapit->second.size() - pos > 100) mapit->second += "\n";
			else mapit->second += " ";

			mapit->second += symbol;
		}
	}

	//print the map
	for (std::pair<const std::pair<int, int>, std::string>& transition : transitions) {
		out << transition.first.first << " -> " << transition.first.second;
		replace(transition.second, "\n", "\\n");
		out << "[label=\"" << transition.second << "\"]\n";
	}
}

