/*
 * GasTexConverter.cpp
 *
 *  Created on: Jan 1, 2014
 *      Author: Martin Zak
 */

#include "GasTexConverter.h"

#include <automaton/FSM/NFA.h>
#include <automaton/FSM/MultiInitialStateNFA.h>
#include <automaton/FSM/EpsilonNFA.h>
#include <automaton/FSM/DFA.h>
#include <automaton/FSM/ExtendedNFA.h>
#include <automaton/FSM/CompactNFA.h>
#include <automaton/TA/NFTA.h>
#include <automaton/TA/DFTA.h>
#include <automaton/PDA/DPDA.h>
#include <automaton/PDA/SinglePopDPDA.h>
#include <automaton/PDA/InputDrivenDPDA.h>
#include <automaton/PDA/InputDrivenNPDA.h>
#include <automaton/PDA/VisiblyPushdownDPDA.h>
#include <automaton/PDA/VisiblyPushdownNPDA.h>
#include <automaton/PDA/RealTimeHeightDeterministicDPDA.h>
#include <automaton/PDA/RealTimeHeightDeterministicNPDA.h>
#include <automaton/PDA/NPDA.h>
#include <automaton/PDA/SinglePopNPDA.h>
#include <automaton/TM/OneTapeDTM.h>

#include <factory/StringDataFactory.hpp>

#include <set>
#include <map>
#include <typeinfo>
#include "exception/CommonException.h"

void GasTexConverter::convert(std::ostream& out, const automaton::Automaton& a) {
	dispatch(out, a.getData());
}

void GasTexConverter::convert(std::ostream& out, const automaton::EpsilonNFA < > & a) {
	out << "\\begin{center}\n";
	out << "\\begin{picture}(,)(,)\n";

	for (auto& state : a.getStates()) {
		bool initial = false;
		bool final = false;

		if(a.getInitialState() == state){
			initial = true;
		}
		if(a.getFinalStates().count(state)){
			final = true;
		}

		if(initial || final) {
			out << "\\node[Nmarks=";
			if(initial){
				out << "i";
			}
			if(final){
				out << "r";
			}
			out<<"](";
		} else {
			out <<"\\node(";
		}

		out << state;
		out << ")(,){";
		out << state;
		out << "}\n";
	}

	transitions(a, out);
	out << "\\end{center}\n";
	out << "\\end{picture}\n";
}

auto GasTexConverterEpsilonNFA = GasTexConverter::RegistratorWrapper<void, automaton::EpsilonNFA < > >(GasTexConverter::convert);

void GasTexConverter::convert(std::ostream& out, const automaton::MultiInitialStateNFA < >& a) {
	out << "\\begin{center}\n";
	out << "\\begin{picture}(,)(,)\n";

	for (auto& state : a.getStates()) {
		bool initial = false;
		bool final = false;

		if(a.getInitialStates().count(state)){
			initial = true;
		}
		if(a.getFinalStates().count(state)){
			final = true;
		}

		if(initial || final) {
			out << "\\node[Nmarks=";
			if(initial){
				out << "i";
			}
			if(final){
				out << "r";
			}
			out<<"](";
		} else {
			out <<"\\node(";
		}

		out << state;
		out << ")(,){";
		out << state;
		out << "}\n";
	}

	transitions(a, out);
	out << "\\end{center}\n";
	out << "\\end{picture}\n";
}

auto GasTexConverterMultiInitialStateNFA = GasTexConverter::RegistratorWrapper<void, automaton::MultiInitialStateNFA < >>(GasTexConverter::convert);

void GasTexConverter::convert(std::ostream& out, const automaton::NFA < > & a) {
	out << "\\begin{center}\n";
	out << "\\begin{picture}(,)(,)\n";

	for (auto& state : a.getStates()) {
		bool initial = false;
		bool final = false;

		if(a.getInitialState() == state){
			initial = true;
		}
		if(a.getFinalStates().count(state)){
			final = true;
		}

		if(initial || final) {
			out << "\\node[Nmarks=";
			if(initial){
				out << "i";
			}
			if(final){
				out << "r";
			}
			out<<"](";
		} else {
			out <<"\\node(";
		}

		out << state;
		out << ")(,){";
		out << state;
		out << "}\n";
	}

	transitions(a, out);
	out << "\\end{center}\n";
	out << "\\end{picture}\n";
}

auto GasTexConverterNFA = GasTexConverter::RegistratorWrapper<void, automaton::NFA < > >(GasTexConverter::convert);

void GasTexConverter::convert(std::ostream& out, const automaton::DFA<>& a) {
	out << "\\begin{center}\n";
	out << "\\begin{picture}(,)(,)\n";

	for (auto& state : a.getStates()) {
		bool initial = false;
		bool final = false;

		if(a.getInitialState() == state){
			initial = true;
		}
		if(a.getFinalStates().count(state)){
			final = true;
		}

		if(initial || final) {
			out << "\\node[Nmarks=";
			if(initial){
				out << "i";
			}
			if(final){
				out << "r";
			}
			out<<"](";
		} else {
			out <<"\\node(";
		}

		out << state;
		out << ")(,){";
		out << state;
		out << "}\n";
	}

	transitions(a, out);
	out << "\\end{center}\n";
	out << "\\end{picture}\n";
}

auto GasTexConverterDFA = GasTexConverter::RegistratorWrapper<void, automaton::DFA<>>(GasTexConverter::convert);

void GasTexConverter::convert(std::ostream& out, const automaton::ExtendedNFA& a) {
	out << "\\begin{center}\n";
	out << "\\begin{picture}(,)(,)\n";

	for (auto& state : a.getStates()) {
		bool initial = false;
		bool final = false;

		if(a.getInitialState() == state){
			initial = true;
		}
		if(a.getFinalStates().count(state)){
			final = true;
		}

		if(initial || final) {
			out << "\\node[Nmarks=";
			if(initial){
				out << "i";
			}
			if(final){
				out << "r";
			}
			out<<"](";
		} else {
			out <<"\\node(";
		}

		out << state;
		out << ")(,){";
		out << state;
		out << "}\n";
	}

	transitions(a, out);
	out << "\\end{center}\n";
	out << "\\end{picture}\n";
}

auto GasTexConverterExtendedNFA = GasTexConverter::RegistratorWrapper<void, automaton::ExtendedNFA>(GasTexConverter::convert);

void GasTexConverter::convert(std::ostream& out, const automaton::CompactNFA < > & a) {
	out << "\\begin{center}\n";
	out << "\\begin{picture}(,)(,)\n";

	for (auto& state : a.getStates()) {
		bool initial = false;
		bool final = false;

		if(a.getInitialState() == state){
			initial = true;
		}
		if(a.getFinalStates().count(state)){
			final = true;
		}

		if(initial || final) {
			out << "\\node[Nmarks=";
			if(initial){
				out << "i";
			}
			if(final){
				out << "r";
			}
			out<<"](";
		} else {
			out <<"\\node(";
		}

		out << state;
		out << ")(,){";
		out << state;
		out << "}\n";
	}

	transitions(a, out);
	out << "\\end{center}\n";
	out << "\\end{picture}\n";
}

auto GasTexConverterCompactNFA = GasTexConverter::RegistratorWrapper<void, automaton::CompactNFA < > >(GasTexConverter::convert);

void GasTexConverter::convert(std::ostream&, const automaton::NFTA&) {
	//TODO
}

auto GasTexConverterNFTA = GasTexConverter::RegistratorWrapper<void, automaton::NFTA>(GasTexConverter::convert);

void GasTexConverter::convert(std::ostream&, const automaton::DFTA&) {
	//TODO
}

auto GasTexConverterDFTA = GasTexConverter::RegistratorWrapper<void, automaton::DFTA>(GasTexConverter::convert);

void GasTexConverter::convert(std::ostream& out, const automaton::DPDA < > & a) {
	out << "\\begin{center}\n";
	out << "\\begin{picture}(,)(,)\n";

	for (auto& state : a.getStates()) {
		bool initial = false;
		bool final = false;

		if(a.getInitialState() == state) {
			initial = true;
		}
		if(a.getFinalStates().count(state)) {
			final = true;
		}

		if(initial || final) {
			out << "\\node[Nmarks=";
			if(initial){
				out << "i";
			}
			if(final){
				out << "r";
			}
			out<<"](";
		} else {
			out <<"\\node(";
		}

		out << state;
		out << ")(,){";
		out << state;
		out << "}\n";
	}

	transitions(a, out);
	out << "\\end{center}\n";
	out << "\\end{picture}\n";
}

auto GasTexConverterDPDA = GasTexConverter::RegistratorWrapper<void, automaton::DPDA < > >(GasTexConverter::convert);

void GasTexConverter::convert(std::ostream& out, const automaton::SinglePopDPDA < > & a) {
	out << "\\begin{center}\n";
	out << "\\begin{picture}(,)(,)\n";

	for (auto& state : a.getStates()) {
		bool initial = false;
		bool final = false;

		if(a.getInitialState() == state) {
			initial = true;
		}
		if(a.getFinalStates().count(state)){
			final = true;
		}

		if(initial || final) {
			out << "\\node[Nmarks=";
			if(initial){
				out << "i";
			}
			if(final){
				out << "r";
			}
			out<<"](";
		} else {
			out <<"\\node(";
		}

		out << state;
		out << ")(,){";
		out << state;
		out << "}\n";
	}

	transitions(a, out);
	out << "\\end{center}\n";
	out << "\\end{picture}\n";
}

auto GasTexConverterSinglePopDPDA = GasTexConverter::RegistratorWrapper<void, automaton::SinglePopDPDA < > >(GasTexConverter::convert);

void GasTexConverter::convert(std::ostream& out, const automaton::InputDrivenDPDA < > & a) {
	out << "\\begin{center}\n";
	out << "\\begin{picture}(,)(,)\n";

	for (auto& state : a.getStates()) {
		bool initial = false;
		bool final = false;

		if(a.getInitialState() == state){
			initial = true;
		}
		if(a.getFinalStates().count(state)){
			final = true;
		}

		if(initial || final) {
			out << "\\node[Nmarks=";
			if(initial){
				out << "i";
			}
			if(final){
				out << "r";
			}
			out<<"](";
		} else {
			out <<"\\node(";
		}

		out << state;
		out << ")(,){";
		out << state;
		out << "}\n";
	}

	transitions(a, out);
	out << "\\end{center}\n";
	out << "\\end{picture}\n";
}

auto GasTexConverterInputDrivenDPDA = GasTexConverter::RegistratorWrapper<void, automaton::InputDrivenDPDA < > >(GasTexConverter::convert);

void GasTexConverter::convert(std::ostream& out, const automaton::InputDrivenNPDA < > & a) {
	out << "\\begin{center}\n";
	out << "\\begin{picture}(,)(,)\n";

	for (auto& state : a.getStates()) {
		bool initial = false;
		bool final = false;

		if(a.getInitialState() == state){
			initial = true;
		}
		if(a.getFinalStates().count(state)){
			final = true;
		}

		if(initial || final) {
			out << "\\node[Nmarks=";
			if(initial){
				out << "i";
			}
			if(final){
				out << "r";
			}
			out<<"](";
		} else {
			out <<"\\node(";
		}

		out << state;
		out << ")(,){";
		out << state;
		out << "}\n";
	}

	transitions(a, out);
	out << "\\end{center}\n";
	out << "\\end{picture}\n";
}

auto GasTexConverterInputDrivenNPDA = GasTexConverter::RegistratorWrapper<void, automaton::InputDrivenNPDA < > >(GasTexConverter::convert);

void GasTexConverter::convert(std::ostream& out, const automaton::VisiblyPushdownDPDA < > & a) {
	out << "\\begin{center}\n";
	out << "\\begin{picture}(,)(,)\n";

	for (auto& state : a.getStates()) {
		bool initial = false;
		bool final = false;

		if(a.getInitialState() == state){
			initial = true;
		}
		if(a.getFinalStates().count(state)){
			final = true;
		}

		if(initial || final) {
			out << "\\node[Nmarks=";
			if(initial){
				out << "i";
			}
			if(final){
				out << "r";
			}
			out<<"](";
		} else {
			out <<"\\node(";
		}

		out << state;
		out << ")(,){";
		out << state;
		out << "}\n";
	}

	transitions(a, out);
	out << "\\end{center}\n";
	out << "\\end{picture}\n";
}

auto GasTexConverterVisiblyPushdownDPDA = GasTexConverter::RegistratorWrapper<void, automaton::VisiblyPushdownDPDA < > >(GasTexConverter::convert);

void GasTexConverter::convert(std::ostream& out, const automaton::VisiblyPushdownNPDA < > & a) {
	out << "\\begin{center}\n";
	out << "\\begin{picture}(,)(,)\n";

	for (auto& state : a.getStates()) {
		bool initial = false;
		bool final = false;

		if(a.getInitialStates().count(state)){
			initial = true;
		}
		if(a.getFinalStates().count(state)){
			final = true;
		}

		if(initial || final) {
			out << "\\node[Nmarks=";
			if(initial){
				out << "i";
			}
			if(final){
				out << "r";
			}
			out<<"](";
		} else {
			out <<"\\node(";
		}

		out << state;
		out << ")(,){";
		out << state;
		out << "}\n";
	}

	transitions(a, out);
	out << "\\end{center}\n";
	out << "\\end{picture}\n";
}

auto GasTexConverterVisiblyPushdownNPDA = GasTexConverter::RegistratorWrapper<void, automaton::VisiblyPushdownNPDA < > >(GasTexConverter::convert);

void GasTexConverter::convert(std::ostream& out, const automaton::RealTimeHeightDeterministicDPDA < > & a) {
	out << "\\begin{center}\n";
	out << "\\begin{picture}(,)(,)\n";

	for (auto& state : a.getStates()) {
		bool initial = false;
		bool final = false;

		if(a.getInitialState() == state){
			initial = true;
		}
		if(a.getFinalStates().count(state)){
			final = true;
		}

		if(initial || final) {
			out << "\\node[Nmarks=";
			if(initial){
				out << "i";
			}
			if(final){
				out << "r";
			}
			out<<"](";
		} else {
			out <<"\\node(";
		}

		out << state;
		out << ")(,){";
		out << state;
		out << "}\n";
	}

	transitions(a, out);
	out << "\\end{center}\n";
	out << "\\end{picture}\n";
}

auto GasTexConverterRealTimeHeightDeterministicDPDA = GasTexConverter::RegistratorWrapper<void, automaton::RealTimeHeightDeterministicDPDA < > >(GasTexConverter::convert);

void GasTexConverter::convert(std::ostream& out, const automaton::RealTimeHeightDeterministicNPDA < > & a) {
	out << "\\begin{center}\n";
	out << "\\begin{picture}(,)(,)\n";

	for (auto& state : a.getStates()) {
		bool initial = false;
		bool final = false;

		if(a.getInitialStates().count(state)){
			initial = true;
		}
		if(a.getFinalStates().count(state)){
			final = true;
		}

		if(initial || final) {
			out << "\\node[Nmarks=";
			if(initial){
				out << "i";
			}
			if(final){
				out << "r";
			}
			out<<"](";
		} else {
			out <<"\\node(";
		}

		out << state;
		out << ")(,){";
		out << state;
		out << "}\n";
	}

	transitions(a, out);
	out << "\\end{center}\n";
	out << "\\end{picture}\n";
}

auto GasTexConverterRealTimeHeightDeterministicNPDA = GasTexConverter::RegistratorWrapper<void, automaton::RealTimeHeightDeterministicNPDA < > >(GasTexConverter::convert);

void GasTexConverter::convert(std::ostream& out, const automaton::NPDA < > & a) {
	out << "\\begin{center}\n";
	out << "\\begin{picture}(,)(,)\n";

	for (auto& state : a.getStates()) {
		bool initial = false;
		bool final = false;

		if(a.getInitialState() == state){
			initial = true;
		}
		if(a.getFinalStates().count(state)){
			final = true;
		}

		if(initial || final) {
			out << "\\node[Nmarks=";
			if(initial){
				out << "i";
			}
			if(final){
				out << "r";
			}
			out<<"](";
		} else {
			out <<"\\node(";
		}

		out << state;
		out << ")(,){";
		out << state;
		out << "}\n";
	}

	transitions(a, out);
	out << "\\end{center}\n";
	out << "\\end{picture}\n";
}

auto GasTexConverterNPDA = GasTexConverter::RegistratorWrapper<void, automaton::NPDA < > >(GasTexConverter::convert);

void GasTexConverter::convert(std::ostream& out, const automaton::SinglePopNPDA < > & a) {
	out << "\\begin{center}\n";
	out << "\\begin{picture}(,)(,)\n";

	for (auto& state : a.getStates()) {
		bool initial = false;
		bool final = false;

		if(a.getInitialState() == state){
			initial = true;
		}
		if(a.getFinalStates().count(state)){
			final = true;
		}

		if(initial || final) {
			out << "\\node[Nmarks=";
			if(initial){
				out << "i";
			}
			if(final){
				out << "r";
			}
			out<<"](";
		} else {
			out <<"\\node(";
		}

		out << state;
		out << ")(,){";
		out << state;
		out << "}\n";
	}

	transitions(a, out);
	out << "\\end{center}\n";
	out << "\\end{picture}\n";
}

auto GasTexConverterSinglePopNPDA = GasTexConverter::RegistratorWrapper<void, automaton::SinglePopNPDA < > >(GasTexConverter::convert);

void GasTexConverter::convert(std::ostream& out, const automaton::OneTapeDTM<>& a) {
	out << "\\begin{center}\n";
	out << "\\begin{picture}(,)(,)\n";

	for (auto& state : a.getStates()) {
		bool initial = false;
		bool final = false;

		if(a.getInitialState() == state){
			initial = true;
		}
		if(a.getFinalStates().count(state)){
			final = true;
		}

		if(initial || final) {
			out << "\\node[Nmarks=";
			if(initial){
				out << "i";
			}
			if(final){
				out << "r";
			}
			out<<"](";
		} else {
			out <<"\\node(";
		}

		out << state;
		out << ")(,){";
		out << state;
		out << "}\n";
	}

	transitions(a, out);
	out << "\\end{center}\n";
	out << "\\end{picture}\n";
}

auto GasTexConverterOneTapeDTM = GasTexConverter::RegistratorWrapper<void, automaton::OneTapeDTM<>>(GasTexConverter::convert);

std::string GasTexConverter::getStackSymbols(const std::vector<alphabet::Symbol>& stackSymbols) {
	if (stackSymbols.size() == 0) {
		return "$\\varepsilon$";
	}

	std::string symbols = "";
	int i=0;
	for (const auto& symbol : stackSymbols) {
		if(i++ !=0) {
			symbols +=" ";
		}
		symbols += (std::string) symbol.getData();
	}
	return symbols;
}

void GasTexConverter::printTransitionMap(const std::map<std::pair<std::string, std::string>, std::string> transitionMap, std::ostream& out) {
	for (const auto& transition : transitionMap) {
		std::string from = transition.first.first;
		std::string to = transition.first.second;
		std::string symbols = transition.second;

		if (from == to) {
			out << "\\drawloop(";
			out << from;
			out << "){";
			out << symbols;
			out << "}\n";

		} else {
			out << "\\drawedge(";
			out << from;
			out << ",";
			out << to;
			out << "){";
			out << symbols;
			out << "}\n";
		}
	}
}

void GasTexConverter::transitions(const automaton::EpsilonNFA < > & fsm, std::ostream& out) {
	std::map<std::pair<std::string, std::string>, std::string> transitionMap;
	for (const auto& transition : fsm.getTransitions()) {
		for(const auto& to : transition.second) {
			std::pair<std::string, std::string> key((std::string) transition.first.first, (std::string) to);

			std::string symbol;
			if (transition.first.second.is<string::Epsilon < >>()) {
				symbol = "$\\varepsilon$";
			} else {
				symbol = (std::string) transition.first.second.get<alphabet::Symbol>();
			}

			auto mapIterator = transitionMap.find(key);
			if (mapIterator == transitionMap.end()) {
				transitionMap.insert(make_pair(key, symbol));
			} else {
				mapIterator->second += ", " + symbol;
			}
		}
	}
	printTransitionMap(transitionMap, out);
}

void GasTexConverter::transitions(const automaton::MultiInitialStateNFA < >& fsm, std::ostream& out) {
	std::map<std::pair<std::string, std::string>, std::string> transitionMap;
	for (const auto& transition : fsm.getTransitions()) {
		for(const auto& to : transition.second) {
			std::pair<std::string, std::string> key((std::string) transition.first.first, (std::string) to);

			std::string symbol = (std::string) transition.first.second;

			auto mapIterator = transitionMap.find(key);
			if (mapIterator == transitionMap.end()) {
				transitionMap.insert(make_pair(key, symbol));
			} else {
				mapIterator->second += ", " + symbol;
			}
		}
	}
	printTransitionMap(transitionMap, out);
}

void GasTexConverter::transitions(const automaton::NFA < > & fsm, std::ostream& out) {
	std::map<std::pair<std::string, std::string>, std::string> transitionMap;
	for (const auto& transition : fsm.getTransitions()) {
		for(const auto& to : transition.second) {
			std::pair<std::string, std::string> key((std::string) transition.first.first, (std::string) to);

			std::string symbol = (std::string) transition.first.second;

			auto mapIterator = transitionMap.find(key);
			if (mapIterator == transitionMap.end()) {
				transitionMap.insert(make_pair(key, symbol));
			} else {
				mapIterator->second += ", " + symbol;
			}
		}
	}
	printTransitionMap(transitionMap, out);
}

void GasTexConverter::transitions(const automaton::DFA<>& fsm, std::ostream& out) {
	std::map<std::pair<std::string, std::string>, std::string> transitionMap;
	for (const auto& transition : fsm.getTransitions()) {
		std::pair<std::string, std::string> key((std::string) transition.first.first, (std::string) transition.second);

		std::string symbol = (std::string) transition.first.second;

		auto mapIterator = transitionMap.find(key);
		if (mapIterator == transitionMap.end()) {
			transitionMap.insert(make_pair(key, symbol));
		} else {
			mapIterator->second += ", " + symbol;
		}
	}
	printTransitionMap(transitionMap, out);
}

void GasTexConverter::transitions(const automaton::ExtendedNFA& fsm, std::ostream& out) {
	std::map<std::pair<std::string, std::string>, std::string> transitionMap;

	for (const auto& transition : fsm.getTransitions()) {
		for(const auto& to : transition.second) {
			std::pair<std::string, std::string> key((std::string) transition.first.first, (std::string) to);

			std::string symbol = alib::StringDataFactory::toString(transition.first.second);

			auto mapIterator = transitionMap.find(key);
			if (mapIterator == transitionMap.end()) {
				transitionMap.insert(make_pair(key, symbol));
			} else {
				mapIterator->second += ", " + symbol;
			}
		}
	}
	printTransitionMap(transitionMap, out);
}

void GasTexConverter::transitions(const automaton::CompactNFA < > & fsm, std::ostream& out) {
	std::map<std::pair<std::string, std::string>, std::string> transitionMap;

	for (const auto& transition : fsm.getTransitions()) {
		for(const auto& to : transition.second) {
			std::pair<std::string, std::string> key((std::string) transition.first.first, (std::string) to);

			std::string symbol = alib::StringDataFactory::toString(string::stringFrom(transition.first.second));

			auto mapIterator = transitionMap.find(key);
			if (mapIterator == transitionMap.end()) {
				transitionMap.insert(make_pair(key, symbol));
			} else {
				mapIterator->second += ", " + symbol;
			}
		}
	}
	printTransitionMap(transitionMap, out);
}

void GasTexConverter::transitions(const automaton::NFTA&, std::ostream&) {
	//TODO
}

void GasTexConverter::transitions(const automaton::DFTA&, std::ostream&) {
	//TODO
}

void GasTexConverter::transitions(const automaton::DPDA < > & pda, std::ostream& out) {
	std::map<std::pair<std::string, std::string>, std::string> transitionMap;

	for (const auto& transition : pda.getTransitions()) {
		std::pair<std::string, std::string> key((std::string) std::get<0>(transition.first), (std::string) transition.second.first);
		auto mapIterator = transitionMap.find(key);

		std::string symbol;
		if (std::get<1>(transition.first).is<string::Epsilon < >>()) {
			symbol = "$\\varepsilon;$";
		} else {
			symbol = alib::StringDataFactory::toString(std::get<1>(transition.first).get<alphabet::Symbol>());
		}

		symbol += "|";

		symbol += getStackSymbols(std::get<2>(transition.first));
		symbol += "\\rarrow";
		symbol += getStackSymbols(transition.second.second);

		if (mapIterator == transitionMap.end()) {
			transitionMap.insert(std::make_pair(key, symbol));
		} else {
			mapIterator->second += "; " + symbol;
		}
	}

	printTransitionMap(transitionMap, out);
}

void GasTexConverter::transitions(const automaton::SinglePopDPDA < > & pda, std::ostream& out) {
	std::map<std::pair<std::string, std::string>, std::string> transitionMap;

	for (const auto& transition : pda.getTransitions()) {
		std::pair<std::string, std::string> key((std::string) std::get<0>(transition.first), (std::string) transition.second.first);
		auto mapIterator = transitionMap.find(key);

		std::string symbol;
		if (std::get<1>(transition.first).is<string::Epsilon < >>()) {
			symbol = "$\\varepsilon;$";
		} else {
			symbol = alib::StringDataFactory::toString(std::get<1>(transition.first).get<alphabet::Symbol>());
		}

		symbol += "|";

		symbol += alib::StringDataFactory::toString(std::get<2>(transition.first));
		symbol += "\\rarrow";
		symbol += getStackSymbols(transition.second.second);

		if (mapIterator == transitionMap.end()) {
			transitionMap.insert(std::make_pair(key, symbol));
		} else {
			mapIterator->second += "; " + symbol;
		}
	}

	printTransitionMap(transitionMap, out);
}

void GasTexConverter::transitions(const automaton::InputDrivenDPDA < > & pda, std::ostream& out) {
	std::map<std::pair<std::string, std::string>, std::string> transitionMap;

	const auto& symbolToPDSOperation = pda.getPushdownStoreOperations();
	for (const auto& transition : pda.getTransitions()) {
		const auto& pop = symbolToPDSOperation.find(std::get<1>(transition.first))->second.first;
		const auto& push = symbolToPDSOperation.find(std::get<1>(transition.first))->second.second;

		const auto& to = transition.second;
		std::pair<std::string, std::string> key((std::string) transition.first.first, (std::string) to);
		auto mapIterator = transitionMap.find(key);

		std::string symbol = alib::StringDataFactory::toString(transition.first.second);

		symbol += "|";

		symbol += getStackSymbols(pop);
		symbol += "\\rarrow";
		symbol += getStackSymbols(push);

		if (mapIterator == transitionMap.end()) {
			transitionMap.insert(std::make_pair(key, symbol));
		} else {
			mapIterator->second += "; " + symbol;
		}
	}

	printTransitionMap(transitionMap, out);
}

void GasTexConverter::transitions(const automaton::InputDrivenNPDA < > & pda, std::ostream& out) {
	std::map<std::pair<std::string, std::string>, std::string> transitionMap;

	const auto& symbolToPDSOperation = pda.getPushdownStoreOperations();
	for (const auto& transition : pda.getTransitions()) {
		const auto& pop = symbolToPDSOperation.find(std::get<1>(transition.first))->second.first;
		const auto& push = symbolToPDSOperation.find(std::get<1>(transition.first))->second.second;

		for(const auto& to : transition.second) {
			std::pair<std::string, std::string> key((std::string) transition.first.first, (std::string) to);
			auto mapIterator = transitionMap.find(key);

			std::string symbol = alib::StringDataFactory::toString(transition.first.second);

			symbol += "|";

			symbol += getStackSymbols(pop);
			symbol += "\\rarrow";
			symbol += getStackSymbols(push);

			if (mapIterator == transitionMap.end()) {
				transitionMap.insert(std::make_pair(key, symbol));
			} else {
				mapIterator->second += "; " + symbol;
			}
		}
	}

	printTransitionMap(transitionMap, out);
}

void GasTexConverter::transitions(const automaton::VisiblyPushdownDPDA < > & pda, std::ostream& out) {
	std::map<std::pair<std::string, std::string>, std::string> transitionMap;

	for (const auto& transition : pda.getCallTransitions()) {
		std::pair<std::string, std::string> key((std::string) transition.first.first, (std::string) transition.second.first);
		auto mapIterator = transitionMap.find(key);

		std::string symbol;
		symbol = alib::StringDataFactory::toString(transition.first.second);

		symbol += "|";

		symbol += "$\\varepsilon;$";
		symbol += "\\rarrow";
		symbol += alib::StringDataFactory::toString(transition.second.second);

		if (mapIterator == transitionMap.end()) {
			transitionMap.insert(std::make_pair(key, symbol));
		} else {
			mapIterator->second += "; " + symbol;
		}
	}

	for (const auto& transition : pda.getReturnTransitions()) {
		std::pair<std::string, std::string> key((std::string) std::get<0>(transition.first), (std::string) transition.second);
		auto mapIterator = transitionMap.find(key);

		std::string symbol;
		symbol = alib::StringDataFactory::toString(std::get<1>(transition.first));

		symbol += "|";

		symbol += alib::StringDataFactory::toString(std::get<2>(transition.first));
		symbol += "\\rarrow";
		symbol += "$\\varepsilon;$";

		if (mapIterator == transitionMap.end()) {
			transitionMap.insert(std::make_pair(key, symbol));
		} else {
			mapIterator->second += "; " + symbol;
		}
	}

	for (const auto& transition : pda.getLocalTransitions()) {
		std::pair<std::string, std::string> key((std::string) transition.first.first, (std::string) transition.second);
		auto mapIterator = transitionMap.find(key);

		std::string symbol;
		symbol = alib::StringDataFactory::toString(transition.first.second);

		symbol += "|";

		symbol += "$\\varepsilon;$";
		symbol += "\\rarrow";
		symbol += "$\\varepsilon;$";

		if (mapIterator == transitionMap.end()) {
			transitionMap.insert(std::make_pair(key, symbol));
		} else {
			mapIterator->second += "; " + symbol;
		}
	}

	printTransitionMap(transitionMap, out);
}

void GasTexConverter::transitions(const automaton::VisiblyPushdownNPDA < > & pda, std::ostream& out) {
	std::map<std::pair<std::string, std::string>, std::string> transitionMap;

	for (const auto& transition : pda.getCallTransitions()) {
		for(const auto& to : transition.second) {
			std::pair<std::string, std::string> key((std::string) transition.first.first, (std::string) to.first);
			auto mapIterator = transitionMap.find(key);

			std::string symbol;
			symbol = alib::StringDataFactory::toString(transition.first.second);

			symbol += "|";

			symbol += "$\\varepsilon;$";
			symbol += "\\rarrow";
			symbol += alib::StringDataFactory::toString(to.second);

			if (mapIterator == transitionMap.end()) {
				transitionMap.insert(std::make_pair(key, symbol));
			} else {
				mapIterator->second += "; " + symbol;
			}
		}
	}

	for (const auto& transition : pda.getReturnTransitions()) {
		for(const auto& to : transition.second) {
			std::pair<std::string, std::string> key((std::string) std::get<0>(transition.first), (std::string) to);
			auto mapIterator = transitionMap.find(key);

			std::string symbol;
			symbol = alib::StringDataFactory::toString(std::get<1>(transition.first));

			symbol += "|";

			symbol += alib::StringDataFactory::toString(std::get<2>(transition.first));
			symbol += "\\rarrow";
			symbol += "$\\varepsilon;$";

			if (mapIterator == transitionMap.end()) {
				transitionMap.insert(std::make_pair(key, symbol));
			} else {
				mapIterator->second += "; " + symbol;
			}
		}
	}

	for (const auto& transition : pda.getLocalTransitions()) {
		for(const auto& to : transition.second) {
			std::pair<std::string, std::string> key((std::string) transition.first.first, (std::string) to);
			auto mapIterator = transitionMap.find(key);

			std::string symbol;
			symbol = alib::StringDataFactory::toString(transition.first.second);

			symbol += "|";

			symbol += "$\\varepsilon;$";
			symbol += "\\rarrow";
			symbol += "$\\varepsilon;$";

			if (mapIterator == transitionMap.end()) {
				transitionMap.insert(std::make_pair(key, symbol));
			} else {
				mapIterator->second += "; " + symbol;
			}
		}
	}

	printTransitionMap(transitionMap, out);
}

void GasTexConverter::transitions(const automaton::RealTimeHeightDeterministicDPDA < > & pda, std::ostream& out) {
	std::map<std::pair<std::string, std::string>, std::string> transitionMap;

	for (const auto& transition : pda.getCallTransitions()) {
		std::pair<std::string, std::string> key((std::string) transition.first.first, (std::string) transition.second.first);
		auto mapIterator = transitionMap.find(key);

		std::string symbol;
		if(transition.first.second.is<string::Epsilon < >>())
			symbol = "$\\varepsilon;$";
		else
			symbol = alib::StringDataFactory::toString(transition.first.second.get<alphabet::Symbol>());

		symbol += "|";

		symbol += "$\\varepsilon;$";
		symbol += "\\rarrow";
		symbol += alib::StringDataFactory::toString(transition.second.second);

		if (mapIterator == transitionMap.end()) {
			transitionMap.insert(std::make_pair(key, symbol));
		} else {
			mapIterator->second += "; " + symbol;
		}
	}

	for (const auto& transition : pda.getReturnTransitions()) {
		std::pair<std::string, std::string> key((std::string) std::get<0>(transition.first), (std::string) transition.second);
		auto mapIterator = transitionMap.find(key);

		std::string symbol;
		if(std::get<1>(transition.first).is<string::Epsilon < >>())
			symbol = "$\\varepsilon;$";
		else
			symbol = alib::StringDataFactory::toString(std::get<1>(transition.first).get<alphabet::Symbol>());

		symbol += "|";

		symbol += alib::StringDataFactory::toString(std::get<2>(transition.first));
		symbol += "\\rarrow";
		symbol += "$\\varepsilon;$";

		if (mapIterator == transitionMap.end()) {
			transitionMap.insert(std::make_pair(key, symbol));
		} else {
			mapIterator->second += "; " + symbol;
		}
	}

	for (const auto& transition : pda.getLocalTransitions()) {
		std::pair<std::string, std::string> key((std::string) transition.first.first, (std::string) transition.second);
		auto mapIterator = transitionMap.find(key);

		std::string symbol;
		if(transition.first.second.is<string::Epsilon < >>())
			symbol = "$\\varepsilon;$";
		else
			symbol = alib::StringDataFactory::toString(transition.first.second.get<alphabet::Symbol>());

		symbol += "|";

		symbol += "$\\varepsilon;$";
		symbol += "\\rarrow";
		symbol += "$\\varepsilon;$";

		if (mapIterator == transitionMap.end()) {
			transitionMap.insert(std::make_pair(key, symbol));
		} else {
			mapIterator->second += "; " + symbol;
		}
	}

	printTransitionMap(transitionMap, out);
}

void GasTexConverter::transitions(const automaton::RealTimeHeightDeterministicNPDA < > & pda, std::ostream& out) {
	std::map<std::pair<std::string, std::string>, std::string> transitionMap;

	for (const auto& transition : pda.getCallTransitions()) {
		for(const auto& to : transition.second) {
			std::pair<std::string, std::string> key((std::string) transition.first.first, (std::string) to.first);
			auto mapIterator = transitionMap.find(key);

			std::string symbol;
			if(transition.first.second.is<string::Epsilon < >>())
				symbol = "$\\varepsilon;$";
			else
				symbol = alib::StringDataFactory::toString(transition.first.second.get<alphabet::Symbol>());

			symbol += "|";

			symbol += "$\\varepsilon;$";
			symbol += "\\rarrow";
			symbol += alib::StringDataFactory::toString(to.second);

			if (mapIterator == transitionMap.end()) {
				transitionMap.insert(std::make_pair(key, symbol));
			} else {
				mapIterator->second += "; " + symbol;
			}
		}
	}

	for (const auto& transition : pda.getReturnTransitions()) {
		for(const auto& to : transition.second) {
			std::pair<std::string, std::string> key((std::string) std::get<0>(transition.first), (std::string) to);
			auto mapIterator = transitionMap.find(key);

			std::string symbol;
			if(std::get<1>(transition.first).is<string::Epsilon < >>())
				symbol = "$\\varepsilon;$";
			else
				symbol = alib::StringDataFactory::toString(std::get<1>(transition.first).get<alphabet::Symbol>());

			symbol += "|";

			symbol += alib::StringDataFactory::toString(std::get<2>(transition.first));
			symbol += "\\rarrow";
			symbol += "$\\varepsilon;$";

			if (mapIterator == transitionMap.end()) {
				transitionMap.insert(std::make_pair(key, symbol));
			} else {
				mapIterator->second += "; " + symbol;
			}
		}
	}

	for (const auto& transition : pda.getLocalTransitions()) {
		for(const auto& to : transition.second) {
			std::pair<std::string, std::string> key((std::string) transition.first.first, (std::string) to);
			auto mapIterator = transitionMap.find(key);

			std::string symbol;
			if(transition.first.second.is<string::Epsilon < >>())
				symbol = "$\\varepsilon;$";
			else
				symbol = alib::StringDataFactory::toString(transition.first.second.get<alphabet::Symbol>());

			symbol += "|";

			symbol += "$\\varepsilon;$";
			symbol += "\\rarrow";
			symbol += "$\\varepsilon;$";

			if (mapIterator == transitionMap.end()) {
				transitionMap.insert(std::make_pair(key, symbol));
			} else {
				mapIterator->second += "; " + symbol;
			}
		}
	}

	printTransitionMap(transitionMap, out);
}

void GasTexConverter::transitions(const automaton::NPDA < > & pda, std::ostream& out) {
	std::map<std::pair<std::string, std::string>, std::string> transitionMap;

	for (const auto& transition : pda.getTransitions()) {
		for(const auto& to : transition.second) {
			std::pair<std::string, std::string> key((std::string) std::get<0>(transition.first), (std::string) to.first);
			auto mapIterator = transitionMap.find(key);

			std::string symbol;
			if (std::get<1>(transition.first).is<string::Epsilon < >>()) {
				symbol = "$\\varepsilon;$";
			} else {
				symbol = alib::StringDataFactory::toString(std::get<1>(transition.first).get<alphabet::Symbol>());
			}

			symbol += "|";

			symbol += getStackSymbols(std::get<2>(transition.first));
			symbol += "\\rarrow";
			symbol += getStackSymbols(to.second);

			if (mapIterator == transitionMap.end()) {
				transitionMap.insert(std::make_pair(key, symbol));
			} else {
				mapIterator->second += "; " + symbol;
			}
		}
	}

	printTransitionMap(transitionMap, out);
}

void GasTexConverter::transitions(const automaton::SinglePopNPDA < > & pda, std::ostream& out) {
	std::map<std::pair<std::string, std::string>, std::string> transitionMap;

	for (const auto& transition : pda.getTransitions()) {
		for(const auto& to : transition.second) {
			std::pair<std::string, std::string> key((std::string) std::get<0>(transition.first), (std::string) to.first);
			auto mapIterator = transitionMap.find(key);

			std::string symbol;
			if (std::get<1>(transition.first).is<string::Epsilon < >>()) {
				symbol = "$\\varepsilon;$";
			} else {
				symbol = alib::StringDataFactory::toString(std::get<1>(transition.first).get<alphabet::Symbol>());
			}

			symbol += "|";

			symbol += alib::StringDataFactory::toString(std::get<2>(transition.first));
			symbol += "\\rarrow";
			symbol += getStackSymbols(to.second);

			if (mapIterator == transitionMap.end()) {
				transitionMap.insert(std::make_pair(key, symbol));
			} else {
				mapIterator->second += "; " + symbol;
			}
		}
	}

	printTransitionMap(transitionMap, out);
}

void GasTexConverter::transitions(const automaton::OneTapeDTM<>& tm, std::ostream& out) {
	std::map<std::pair<std::string, std::string>, std::string> transitionMap;

	for (auto& transition : tm.getTransitions()) {
		std::pair<std::string, std::string> key((std::string) transition.first.first, (std::string) std::get<0>(transition.second));
		auto mapIterator = transitionMap.find(key);

		std::string symbol = alib::StringDataFactory::toString(transition.first.second);
		symbol += "/";
		symbol += alib::StringDataFactory::toString(std::get<1>(transition.second));
		symbol += ",";
		switch(std::get<2>(transition.second)) {
		case automaton::Shift::LEFT:
			symbol += "$\\leftarrow$";
			break;
		case automaton::Shift::RIGHT:
			symbol += "$\\rightarrow$";
			break;
		case automaton::Shift::NONE:
			symbol += "$\\times$";
			break;
		default:
			throw exception::CommonException("Unexpected shift direction");
		}

		if (mapIterator == transitionMap.end()) {
			transitionMap.insert(std::make_pair(key, symbol));
		} else {
			mapIterator->second += "; " + symbol;
		}
	}

	printTransitionMap(transitionMap, out);
}

