/*
 * DFTA.cpp
 *
 *  Created on: Apr 14, 2015
 *      Author: Stepan Plachy
 */

#include "DFTA.h"
#include <ostream>
#include <sstream>

#include <sax/FromXMLParserHelper.h>
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../Automaton.h"
#include <object/Object.h>
#include <XmlApi.hpp>

namespace automaton {

DFTA::DFTA ( std::set < label::Label > states, std::set < alphabet::RankedSymbol > inputAlphabet, std::set < label::Label > finalStates ) : std::Components < DFTA, alphabet::RankedSymbol, std::tuple < InputAlphabet >, std::tuple < >, label::Label, std::tuple < States, FinalStates >, std::tuple < > > ( std::make_tuple ( std::move ( inputAlphabet ) ), std::tuple < > ( ), std::make_tuple ( std::move ( states ), std::move ( finalStates ) ), std::tuple < > ( ) ) {
}

DFTA::DFTA() : DFTA ( std::set < label::Label > { }, std::set < alphabet::RankedSymbol > { }, std::set < label::Label > { } ) {
}

AutomatonBase* DFTA::clone() const {
	return new DFTA(*this);
}

AutomatonBase* DFTA::plunder() && {
	return new DFTA(std::move(*this));
}

bool DFTA::addTransition(const alphabet::RankedSymbol & symbol, const std::vector<label::Label> & prevStates, const label::Label & next) {
	if ( prevStates.size() != symbol.getRank().getData())
		throw AutomatonException("Number of states doesn't match rank of the symbol");

	if (! getInputAlphabet().count(symbol))
		throw AutomatonException("Input symbol \"" + std::to_string ( symbol ) + "\" doesn't exist.");

	if (! getStates().count(next))
		throw AutomatonException("State \"" + std::to_string ( next ) + "\" doesn't exist.");

	for (const label::Label& it : prevStates) {
		if (! getStates().count(it))
			throw AutomatonException("State \"" + std::to_string ( it ) + "\" doesn't exist.");
	}

	std::pair<alphabet::RankedSymbol, std::vector<label::Label> > key = std::make_pair(symbol, prevStates);
	if ( transitions.find ( key ) != transitions.end ( ) ) {
		if ( transitions.find ( key )->second == next )
			return false;
		else
			throw AutomatonException("Transition already exists");
	}

	transitions.insert(std::make_pair(key, next));
	return true;
}

bool DFTA::removeTransition(const alphabet::RankedSymbol symbol, const std::vector<label::Label> & states, const label::Label & next) {
	std::pair<alphabet::RankedSymbol, std::vector<label::Label> > key = std::make_pair(symbol, states);

	if ( transitions.find ( key ) == transitions.end ( ) )
		return false;

	if ( transitions.find ( key )->second != next )
		throw AutomatonException("Transition does not exist");

	transitions.erase(key);
	return true;
}

int DFTA::compare(const DFTA& other) const {
	auto first = std::tie(getStates(), getInputAlphabet(), getFinalStates(), transitions);
	auto second = std::tie(other.getStates(), other.getInputAlphabet(), other.getFinalStates(), other.transitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void DFTA::operator>>(std::ostream& out) const {
	out << "(DFTA "
		<< " states = " << getStates()
		<< " inputAlphabet = " << getInputAlphabet()
		<< " finalStates = " << getFinalStates()
		<< " transitions = " << transitions
		<< ")";
}

DFTA::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

DFTA DFTA::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, DFTA::getXmlTagName());

	std::set<label::Label> states = AutomatonFromXMLParser::parseStates<label::Label>(input);
	std::set<alphabet::RankedSymbol> inputSymbols = AutomatonFromXMLParser::parseRankedInputAlphabet<alphabet::RankedSymbol>(input);
	std::set<label::Label> finalStates = AutomatonFromXMLParser::parseFinalStates<label::Label>(input);

	DFTA automaton;
	automaton.setStates(std::move(states));
	automaton.setInputAlphabet(std::move(inputSymbols));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<DFTA>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, DFTA::getXmlTagName());
	return automaton;
}

void DFTA::parseTransition(std::deque<sax::Token>::iterator& input, DFTA& automaton) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "transition");
	alphabet::RankedSymbol inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol<alphabet::RankedSymbol>(input);
	std::vector<label::Label> from = AutomatonFromXMLParser::parseTransitionFromMultiple<label::Label>(input);
	label::Label to = AutomatonFromXMLParser::parseTransitionTo<label::Label>(input);
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "transition");

	automaton.addTransition(std::move(inputSymbol), std::move(from), std::move(to));
}

void DFTA::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(DFTA::getXmlTagName(), sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeRankedInputAlphabet(out, this->getInputAlphabet());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	composeTransitions(out);

	out.emplace_back(DFTA::getXmlTagName(), sax::Token::TokenType::END_ELEMENT);
}

void DFTA::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back(sax::Token("transitions", sax::Token::TokenType::START_ELEMENT));
	for(const auto& transition : this->getTransitions()) {
		out.emplace_back(sax::Token("transition", sax::Token::TokenType::START_ELEMENT));

		AutomatonToXMLComposer::composeTransitionInputSymbol(out, transition.first.first);
		AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.second);
		AutomatonToXMLComposer::composeTransitionTo(out, transition.second);

		out.emplace_back(sax::Token("transition", sax::Token::TokenType::END_ELEMENT));
	}

	out.emplace_back(sax::Token("transitions", sax::Token::TokenType::END_ELEMENT));
}

} /* namespace automaton */

namespace alib {

auto DFTAParserRegister = xmlApi<automaton::Automaton>::ParserRegister<automaton::DFTA>();
auto DFTAParserRegister2 = xmlApi<alib::Object>::ParserRegister<automaton::DFTA>();

} /* namespace alib */
