/*
 * NFTA.cpp
 *
 *  Created on: Mar 21, 2015
 *      Author: Stepan Plachy
 */

#include "NFTA.h"
#include <ostream>
#include <sstream>

#include <sax/FromXMLParserHelper.h>
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../Automaton.h"
#include <object/Object.h>
#include <XmlApi.hpp>

namespace automaton {

NFTA::NFTA ( std::set < label::Label > states, std::set < alphabet::RankedSymbol > inputAlphabet, std::set < label::Label > finalStates ) : std::Components < NFTA, alphabet::RankedSymbol, std::tuple < InputAlphabet >, std::tuple < >, label::Label, std::tuple < States, FinalStates >, std::tuple < > > ( std::make_tuple ( std::move ( inputAlphabet ) ), std::tuple < > ( ), std::make_tuple ( std::move ( states ), std::move ( finalStates ) ), std::tuple < > ( ) ) {
}

NFTA::NFTA() : NFTA ( std::set < label::Label > { }, std::set < alphabet::RankedSymbol > { }, std::set < label::Label > { } ) {
}

NFTA::NFTA(const DFTA& other) : NFTA ( other.getStates(), other.getInputAlphabet(), other.getFinalStates() ) {
	for(const auto& transition : other.getTransitions()) {
		transitions[transition.first].insert(transition.second);
	}
}

AutomatonBase* NFTA::clone() const {
	return new NFTA(*this);
}

AutomatonBase* NFTA::plunder() && {
	return new NFTA(std::move(*this));
}

bool NFTA::addTransition(const alphabet::RankedSymbol & symbol, const std::vector<label::Label> & prevStates, const label::Label & next) {
	if (prevStates.size() != symbol.getRank().getData())
		throw AutomatonException("Number of states doesn't match rank of the symbol");

	if (! getInputAlphabet().count(symbol))
		throw AutomatonException("Input symbol \"" + std::to_string ( symbol ) + "\" doesn't exist.");

	if (! getStates().count(next))
		throw AutomatonException("State \"" + std::to_string ( next ) + "\" doesn't exist.");

	for (const label::Label& it : prevStates) {
		if (! getStates().count(it))
			throw AutomatonException("State \"" + std::to_string ( it ) + "\" doesn't exist.");
	}

	std::pair<alphabet::RankedSymbol, std::vector<label::Label> > key = std::make_pair(symbol, prevStates);
	return transitions[key].insert(next).second;
}

bool NFTA::removeTransition(const alphabet::RankedSymbol symbol, const std::vector<label::Label> & states, const label::Label & next) {
	std::pair<alphabet::RankedSymbol, std::vector<label::Label> > key = std::make_pair(symbol, states);
	return transitions[key].erase(next);
}

bool NFTA::isDeterministic() const {
	for (const std::pair<const std::pair<alphabet::RankedSymbol, std::vector<label::Label> >, std::set<label::Label> >& t : transitions)
		if (t . second.size() != 1 || t . second.size() != 0)
			return false;

	return true;
}

unsigned NFTA::transitionsSize() const {
	int res = 0;

	for(const auto& transition : transitions )
		res += transition.second.size();

	return res;
}

int NFTA::compare(const NFTA& other) const {
	auto first = std::tie(getStates(), getInputAlphabet(), getFinalStates(), transitions);
	auto second = std::tie(other.getStates(), other.getInputAlphabet(), other.getFinalStates(), other.transitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void NFTA::operator>>(std::ostream& out) const {
	out << "(NFTA "
		<< " states = " << getStates()
		<< " inputAlphabet = " << getInputAlphabet()
		<< " finalStates = " << getFinalStates()
		<< " transitions = " << transitions
		<< ")";
}

NFTA::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

NFTA NFTA::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, NFTA::getXmlTagName());

	std::set<label::Label> states = AutomatonFromXMLParser::parseStates<label::Label>(input);
	std::set<alphabet::RankedSymbol> inputSymbols = AutomatonFromXMLParser::parseRankedInputAlphabet<alphabet::RankedSymbol>(input);
	std::set<label::Label> finalStates = AutomatonFromXMLParser::parseFinalStates<label::Label>(input);

	NFTA automaton;
	automaton.setStates(std::move(states));
	automaton.setInputAlphabet(std::move(inputSymbols));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<NFTA>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, NFTA::getXmlTagName());
	return automaton;
}

void NFTA::parseTransition(std::deque<sax::Token>::iterator& input, NFTA& automaton) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "transition");
	alphabet::RankedSymbol inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol<alphabet::RankedSymbol>(input);
	std::vector<label::Label> from = AutomatonFromXMLParser::parseTransitionFromMultiple<label::Label>(input);
	label::Label to = AutomatonFromXMLParser::parseTransitionTo<label::Label>(input);
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "transition");

	automaton.addTransition(std::move(inputSymbol), std::move(from), std::move(to));
}

void NFTA::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(NFTA::getXmlTagName(), sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeRankedInputAlphabet(out, this->getInputAlphabet());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	composeTransitions(out);

	out.emplace_back(NFTA::getXmlTagName(), sax::Token::TokenType::END_ELEMENT);
}

void NFTA::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back(sax::Token("transitions", sax::Token::TokenType::START_ELEMENT));
	for(const auto& transition : this->getTransitions()) {
		for(const auto& targetState: transition.second) {
			out.emplace_back(sax::Token("transition", sax::Token::TokenType::START_ELEMENT));

			AutomatonToXMLComposer::composeTransitionInputSymbol(out, transition.first.first);
			AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.second);
			AutomatonToXMLComposer::composeTransitionTo(out, targetState);

			out.emplace_back(sax::Token("transition", sax::Token::TokenType::END_ELEMENT));
		}
	}

	out.emplace_back(sax::Token("transitions", sax::Token::TokenType::END_ELEMENT));
}

} /* namespace automaton */

namespace alib {

auto NFTAParserRegister = xmlApi<automaton::Automaton>::ParserRegister<automaton::NFTA>();
auto NFTAParserRegister2 = xmlApi<alib::Object>::ParserRegister<automaton::NFTA>();

} /* namespace alib */
