/*
 * AutomatonFromXMLParser.h
 *
 *  Created on: Oct 12, 2013
 *      Author: Martin Zak
 */

#ifndef AUTOMATON_FROM_XML_PARSER_H_
#define AUTOMATON_FROM_XML_PARSER_H_

#include <sax/FromXMLParserHelper.h>
#include <set>
#include <vector>
#include <variant>
#include <map>
#include "Shift.h"
#include <XmlApi.hpp>
#include "../../regexp/RegExpFeatures.h"

namespace automaton {

/**
 * Parser used to get general FSM or EpsilonNFA, NFA, DFA from XML parsed into deque of Tokens.
 */
class AutomatonFromXMLParser {
public:
	template<class T>
	static void parseTransitions(std::deque<sax::Token>::iterator& input, T& automaton);

	template < class StateType >
	static std::set<StateType> parseStates(std::deque<sax::Token>::iterator& input);
	template < class SymbolType >
	static std::set<SymbolType> parseInputAlphabet(std::deque<sax::Token>::iterator& input);
	template < class SymbolType >
	static std::set<SymbolType> parseCallInputAlphabet(std::deque<sax::Token>::iterator& input);
	template < class SymbolType >
	static std::set<SymbolType> parseReturnInputAlphabet(std::deque<sax::Token>::iterator& input);
	template < class SymbolType >
	static std::set<SymbolType> parseLocalInputAlphabet(std::deque<sax::Token>::iterator& input);
	template < class SymbolType >
	static std::set<SymbolType> parseRankedInputAlphabet(std::deque<sax::Token>::iterator& input);
	template < class SymbolType >
	static std::set<SymbolType> parsePushdownStoreAlphabet(std::deque<sax::Token>::iterator& input);
	template < class SymbolType >
	static std::set<SymbolType> parseInitialPushdownStoreSymbols(std::deque<sax::Token>::iterator& input);
	template < class SymbolType >
	static SymbolType parseInitialPushdownStoreSymbol(std::deque<sax::Token>::iterator& input);
	template < class SymbolType >
	static std::set<SymbolType> parseOutputAlphabet(std::deque<sax::Token>::iterator& input);
	template < class SymbolType >
	static std::set<SymbolType> parseTapeAlphabet(std::deque<sax::Token>::iterator& input);
	template < class StateType >
	static StateType parseInitialState(std::deque<sax::Token>::iterator& input);
	template < class StateType >
	static std::set<StateType> parseInitialStates(std::deque<sax::Token>::iterator& input);
	template < class StateType >
	static std::set<StateType> parseFinalStates(std::deque<sax::Token>::iterator& input);
	template < class SymbolType >
	static SymbolType parseBlankSymbol(std::deque<sax::Token>::iterator& input);
	template < class SymbolType >
	static SymbolType parseBottomOfTheStackSymbol(std::deque<sax::Token>::iterator& input);
	template < class SymbolType >
	static std::map<SymbolType, std::pair<std::vector<SymbolType>, std::vector<SymbolType> > > parseInputToPushdownStoreOperation(std::deque<sax::Token>::iterator& input);

	template < class StateType >
	static StateType parseTransitionFrom(std::deque<sax::Token>::iterator& input);
	template < class StateType >
	static std::vector<StateType> parseTransitionFromMultiple(std::deque<sax::Token>::iterator& input);
	template < class StateType >
	static StateType parseTransitionTo(std::deque<sax::Token>::iterator& input);
	template < class SymbolType >
	static std::vector<SymbolType> parseTransitionPop(std::deque<sax::Token>::iterator& input);
	template < class SymbolType >
	static SymbolType parseTransitionSinglePop(std::deque<sax::Token>::iterator& input);
	template < class SymbolType >
	static std::vector<SymbolType> parseTransitionPush(std::deque<sax::Token>::iterator& input);
	template < class SymbolType >
	static SymbolType parseTransitionSinglePush(std::deque<sax::Token>::iterator& input);

	template < class SymbolType >
	static SymbolType parseTransitionInputSymbol(std::deque<sax::Token>::iterator& input);
	template < class EpsilonType, class SymbolType >
	static std::variant<EpsilonType, SymbolType> parseTransitionInputEpsilonSymbol(std::deque<sax::Token>::iterator& input);
	template < class SymbolType >
	static std::vector<SymbolType> parseTransitionInputSymbolMultiple(std::deque<sax::Token>::iterator& input);

	template < class SymbolType >
	static SymbolType parseTransitionOutputSymbol(std::deque<sax::Token>::iterator& input);
	template < class EpsilonType, class SymbolType >
	static std::variant<EpsilonType, SymbolType> parseTransitionOutputEpsilonSymbol(std::deque<sax::Token>::iterator& input);
	template < class SymbolType >
	static std::vector<SymbolType> parseTransitionOutputSymbolMultiple(std::deque<sax::Token>::iterator& input);

	static Shift parseTransitionShift(std::deque<sax::Token>::iterator& input);
	static regexp::RegExp parseTransitionInputRegexp(std::deque<sax::Token>::iterator& input);
};

template<class T>
void AutomatonFromXMLParser::parseTransitions(std::deque<sax::Token>::iterator& input, T& automaton) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "transitions");
	while (sax::FromXMLParserHelper::isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		T::parseTransition(input, automaton);
	}
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "transitions");
}

template < class StateType >
std::set < StateType > AutomatonFromXMLParser::parseStates ( std::deque < sax::Token >::iterator & input ) {
	std::set < StateType > states;
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "states" );

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) )
		states.insert ( alib::xmlApi < StateType >::parse ( input ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "states" );
	return states;
}

template < class SymbolType >
std::set < SymbolType > AutomatonFromXMLParser::parseCallInputAlphabet ( std::deque < sax::Token >::iterator & input ) {
	std::set < SymbolType > inputSymbols;
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "callInputAlphabet" );

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) )
		inputSymbols.insert ( alib::xmlApi < SymbolType >::parse ( input ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "callInputAlphabet" );
	return inputSymbols;
}

template < class SymbolType >
std::set < SymbolType > AutomatonFromXMLParser::parseReturnInputAlphabet ( std::deque < sax::Token >::iterator & input ) {
	std::set < SymbolType > inputSymbols;
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "returnInputAlphabet" );

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) )
		inputSymbols.insert ( alib::xmlApi < SymbolType >::parse ( input ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "returnInputAlphabet" );
	return inputSymbols;
}

template < class SymbolType >
std::set < SymbolType > AutomatonFromXMLParser::parseLocalInputAlphabet ( std::deque < sax::Token >::iterator & input ) {
	std::set < SymbolType > inputSymbols;
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "localInputAlphabet" );

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) )
		inputSymbols.insert ( alib::xmlApi < SymbolType >::parse ( input ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "localInputAlphabet" );
	return inputSymbols;
}

template < class SymbolType >
std::set < SymbolType > AutomatonFromXMLParser::parseInputAlphabet ( std::deque < sax::Token >::iterator & input ) {
	std::set < SymbolType > inputSymbols;
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "inputAlphabet" );

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) )
		inputSymbols.insert ( alib::xmlApi < SymbolType >::parse ( input ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "inputAlphabet" );
	return inputSymbols;
}

template < class SymbolType >
std::set < SymbolType > AutomatonFromXMLParser::parseRankedInputAlphabet ( std::deque < sax::Token >::iterator & input ) {
	std::set < SymbolType > inputSymbols;
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "rankedInputAlphabet" );

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) )
		inputSymbols.insert ( alib::xmlApi < SymbolType >::parse ( input ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "rankedInputAlphabet" );
	return inputSymbols;
}

template < class StateType >
StateType AutomatonFromXMLParser::parseInitialState ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "initialState" );
	StateType state ( alib::xmlApi < StateType >::parse ( input ) );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "initialState" );

	return state;
}

template < class StateType >
std::set < StateType > AutomatonFromXMLParser::parseInitialStates ( std::deque < sax::Token >::iterator & input ) {
	std::set < StateType > initialStates;
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "initialStates" );

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) )
		initialStates.insert ( alib::xmlApi < StateType >::parse ( input ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "initialStates" );
	return initialStates;
}

template < class StateType >
std::set < StateType > AutomatonFromXMLParser::parseFinalStates ( std::deque < sax::Token >::iterator & input ) {
	std::set < StateType > finalStates;
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "finalStates" );

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) )
		finalStates.insert ( alib::xmlApi < StateType >::parse ( input ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "finalStates" );
	return finalStates;
}

template < class SymbolType >
std::set < SymbolType > AutomatonFromXMLParser::parsePushdownStoreAlphabet ( std::deque < sax::Token >::iterator & input ) {
	std::set < SymbolType > stackSymbols;
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "pushdownStoreAlphabet" );

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) )
		stackSymbols.insert ( alib::xmlApi < SymbolType >::parse ( input ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "pushdownStoreAlphabet" );
	return stackSymbols;
}

template < class SymbolType >
std::set < SymbolType > AutomatonFromXMLParser::parseInitialPushdownStoreSymbols ( std::deque < sax::Token >::iterator & input ) {
	std::set < SymbolType > initialSymbols;
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "initialPushdownStoreSymbols" );

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) )
		initialSymbols.insert ( alib::xmlApi < SymbolType >::parse ( input ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "initialPushdownStoreSymbols" );
	return initialSymbols;
}

template < class SymbolType >
std::set < SymbolType > AutomatonFromXMLParser::parseOutputAlphabet ( std::deque < sax::Token >::iterator & input ) {
	std::set < SymbolType > outputSymbols;
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "outputAlphabet" );

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) )
		outputSymbols.insert ( alib::xmlApi < SymbolType >::parse ( input ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "outputAlphabet" );
	return outputSymbols;
}

template < class SymbolType >
SymbolType AutomatonFromXMLParser::parseInitialPushdownStoreSymbol ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "initialPushdownStoreSymbol" );
	SymbolType initialSymbol ( alib::xmlApi < SymbolType >::parse ( input ) );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "initialPushdownStoreSymbol" );

	return initialSymbol;
}

template < class SymbolType >
std::set < SymbolType > AutomatonFromXMLParser::parseTapeAlphabet ( std::deque < sax::Token >::iterator & input ) {
	std::set < SymbolType > tapeSymbols;
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "tapeAlphabet" );

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) )
		tapeSymbols.insert ( alib::xmlApi < SymbolType >::parse ( input ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "tapeAlphabet" );
	return tapeSymbols;
}

template < class SymbolType >
SymbolType AutomatonFromXMLParser::parseBlankSymbol ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "blankSymbol" );
	SymbolType blank ( alib::xmlApi < SymbolType >::parse ( input ) );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "blankSymbol" );

	return blank;
}

template < class SymbolType >
SymbolType AutomatonFromXMLParser::parseBottomOfTheStackSymbol ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "bottomOfTheStackSymbol" );
	SymbolType blank ( alib::xmlApi < SymbolType >::parse ( input ) );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "bottomOfTheStackSymbol" );

	return blank;
}

template < class SymbolType >
std::map < SymbolType, std::pair < std::vector < SymbolType >, std::vector < SymbolType > > > AutomatonFromXMLParser::parseInputToPushdownStoreOperation ( std::deque < sax::Token >::iterator & input ) {
	std::map < SymbolType, std::pair < std::vector < SymbolType >, std::vector < SymbolType > > > operations;

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "inputToPushdownStoreOperations" );

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) ) {
		sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "operation" );

		SymbolType inputSymbol ( alib::xmlApi < SymbolType >::parse ( input ) );

		std::vector < SymbolType > pop  = parseTransitionPop < SymbolType > ( input );
		std::vector < SymbolType > push = parseTransitionPush < SymbolType > ( input );

		operations.insert ( std::make_pair ( std::move ( inputSymbol ), std::make_pair ( std::move ( pop ), std::move ( push ) ) ) );

		sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "operation" );
	}

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "inputToPushdownStoreOperations" );
	return operations;
}

template < class StateType >
StateType AutomatonFromXMLParser::parseTransitionTo ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "to" );
	StateType state ( alib::xmlApi < StateType >::parse ( input ) );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "to" );

	return state;
}

template < class StateType >
StateType AutomatonFromXMLParser::parseTransitionFrom ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "from" );
	StateType state ( alib::xmlApi < StateType >::parse ( input ) );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "from" );

	return state;
}

template < class StateType >
std::vector < StateType > AutomatonFromXMLParser::parseTransitionFromMultiple ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "from" );
	std::vector < StateType > states;

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) )
		states.push_back ( alib::xmlApi < StateType >::parse ( input ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "from" );
	return states;
}

template < class SymbolType >
std::vector < SymbolType > AutomatonFromXMLParser::parseTransitionPop ( std::deque < sax::Token >::iterator & input ) {
	std::vector < SymbolType > pops;
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "pop" );

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) )
		pops.push_back ( alib::xmlApi < SymbolType >::parse ( input ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "pop" );
	return pops;
}

template < class SymbolType >
SymbolType AutomatonFromXMLParser::parseTransitionSinglePop ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "pop" );
	SymbolType pop = alib::xmlApi < SymbolType >::parse ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "pop" );

	return pop;
}

template < class SymbolType >
std::vector < SymbolType > AutomatonFromXMLParser::parseTransitionPush ( std::deque < sax::Token >::iterator & input ) {
	std::vector < SymbolType > pushes;
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "push" );

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) )
		pushes.push_back ( alib::xmlApi < SymbolType >::parse ( input ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "push" );
	return pushes;
}

template < class SymbolType >
SymbolType AutomatonFromXMLParser::parseTransitionSinglePush ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "push" );
	SymbolType push = alib::xmlApi < SymbolType >::parse ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "push" );

	return push;
}

template < class SymbolType >
SymbolType AutomatonFromXMLParser::parseTransitionInputSymbol ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "input" );
	SymbolType result ( alib::xmlApi < SymbolType >::parse ( input ) );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "input" );

	return result;
}

template < class SymbolType >
SymbolType AutomatonFromXMLParser::parseTransitionOutputSymbol ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "output" );
	SymbolType result ( alib::xmlApi < SymbolType >::parse ( input ) );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "output" );

	return result;
}

template < class EpsilonType, class SymbolType >
std::variant < EpsilonType, SymbolType > AutomatonFromXMLParser::parseTransitionInputEpsilonSymbol ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "input" );

	if ( sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, "epsilon" ) ) {
		std::variant < EpsilonType, SymbolType > result = std::variant < EpsilonType, SymbolType >::template from < EpsilonType > ( );
		++input;
		sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "epsilon" );
		sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "input" );
		return result;
	} else {
		std::variant < EpsilonType, SymbolType > result ( alib::xmlApi < SymbolType >::parse ( input ) );
		sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "input" );
		return result;
	}
}

template < class EpsilonType, class SymbolType >
std::variant < EpsilonType, SymbolType > AutomatonFromXMLParser::parseTransitionOutputEpsilonSymbol ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "output" );

	if ( sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, "epsilon" ) ) {
		std::variant < EpsilonType, SymbolType > result = std::variant < EpsilonType, SymbolType >::template from < EpsilonType > ( );
		++input;
		sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "epsilon" );
		sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "output" );
		return result;
	} else {
		std::variant < EpsilonType, SymbolType > result ( alib::xmlApi < SymbolType >::parse ( input ) );
		sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "output" );
		return result;
	}
}

template < class SymbolType >
std::vector < SymbolType > AutomatonFromXMLParser::parseTransitionInputSymbolMultiple ( std::deque < sax::Token >::iterator & input ) {
	std::vector < SymbolType > inputSymbols;
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "input" );

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) )
		inputSymbols.push_back ( alib::xmlApi < SymbolType >::parse ( input ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "input" );
	return inputSymbols;
}

template < class SymbolType >
std::vector < SymbolType > AutomatonFromXMLParser::parseTransitionOutputSymbolMultiple ( std::deque < sax::Token >::iterator & input ) {
	std::vector < SymbolType > outputSymbols;
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "output" );

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) )
		outputSymbols.push_back ( alib::xmlApi < SymbolType >::parse ( input ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "output" );
	return outputSymbols;
}

} /* namespace automaton */

#endif /* AUTOMATON_FROM_XML_PARSER_H_ */

