/*
 * AutomatonToXMLComposer.h
 *
 *  Created on: Nov 11, 2013
 *      Author: Jan Travnicek
 */

#ifndef AUTOMATON_TO_XML_COMPOSER_H_
#define AUTOMATON_TO_XML_COMPOSER_H_

#include <deque>
#include <vector>
#include <set>
#include <variant>
#include <map>
#include "Shift.h"
#include <sax/Token.h>
#include <XmlApi.hpp>
#include "../../regexp/RegExpFeatures.h"

namespace automaton {

/**
 * This class contains methods to print XML representation of automata to the output stream.
 */
class AutomatonToXMLComposer {
public:
	template<class StateType>
	static void composeStates(std::deque<sax::Token>&, const std::set<StateType>& states);
	template<class SymbolType>
	static void composeInputAlphabet(std::deque<sax::Token>&, const std::set<SymbolType>& symbols);
	template<class SymbolType>
	static void composeCallInputAlphabet(std::deque<sax::Token>&, const std::set<SymbolType>& symbols);
	template<class SymbolType>
	static void composeReturnInputAlphabet(std::deque<sax::Token>&, const std::set<SymbolType>& symbols);
	template<class SymbolType>
	static void composeLocalInputAlphabet(std::deque<sax::Token>&, const std::set<SymbolType>& symbols);
	template<class SymbolType>
	static void composeRankedInputAlphabet(std::deque<sax::Token>&, const std::set<SymbolType>& symbols);
	template<class StateType>
	static void composeInitialStates(std::deque<sax::Token>&, const std::set<StateType>& states);
	template<class StateType>
	static void composeInitialState(std::deque<sax::Token>&, const StateType& state);
	template<class StateType>
	static void composeFinalStates(std::deque<sax::Token>&, const std::set<StateType>& states);
	template<class SymbolType>
	static void composePushdownStoreAlphabet(std::deque<sax::Token>&, const std::set<SymbolType>& symbols);
	template<class SymbolType>
	static void composeInitialPushdownStoreSymbols(std::deque<sax::Token>&, const std::set<SymbolType>& symbols);
	template<class SymbolType>
	static void composeInitialPushdownStoreSymbol(std::deque<sax::Token>&, const SymbolType& symbols);
	template<class SymbolType>
	static void composeOutputAlphabet(std::deque<sax::Token>&, const std::set<SymbolType>& symbols);
	template<class SymbolType>
	static void composeTapeAlphabet(std::deque<sax::Token>&, const std::set<SymbolType>& symbols);
	template<class SymbolType>
	static void composeBlankSymbol(std::deque<sax::Token>&, const SymbolType& symbol);
	template<class SymbolType>
	static void composeBottomOfTheStackSymbol(std::deque<sax::Token>&, const SymbolType& symbol);
	template<class SymbolType>
	static void composeInputToPushdownStoreOperation(std::deque<sax::Token>&, const std::map<SymbolType, std::pair<std::vector<SymbolType>, std::vector<SymbolType> > >& operations);

	template<class StateType>
	static void composeTransitionTo(std::deque<sax::Token>&, const StateType& state);
	template<class StateType>
	static void composeTransitionFrom(std::deque<sax::Token>&, const StateType& state);
	template<class StateType>
	static void composeTransitionFrom(std::deque<sax::Token>&, const std::vector<StateType> & states);
	template<class SymbolType>
	static void composeTransitionPop(std::deque<sax::Token>&, const std::vector<SymbolType>& symbols);
	template<class SymbolType>
	static void composeTransitionSinglePop(std::deque<sax::Token>&, const SymbolType& symbol);
	template<class SymbolType>
	static void composeTransitionPush(std::deque<sax::Token>&, const std::vector<SymbolType>& symbols);
	template<class SymbolType>
	static void composeTransitionSinglePush(std::deque<sax::Token>&, const SymbolType& symbol);
	template<class SymbolType>
	static void composeTransitionInputSymbol(std::deque<sax::Token>&, const SymbolType& symbol);
	template<class EpsilonType, class SymbolType>
	static void composeTransitionInputEpsilonSymbol(std::deque<sax::Token>&, const std::variant<EpsilonType, SymbolType>&);
	template<class SymbolType>
	static void composeTransitionInputSymbolMultiple(std::deque<sax::Token>&, const std::vector<SymbolType>& symbols);
	template<class SymbolType>
	static void composeTransitionOutputSymbol(std::deque<sax::Token>&, const SymbolType& symbol);
	template<class EpsilonType, class SymbolType>
	static void composeTransitionOutputEpsilonSymbol(std::deque<sax::Token>&, const std::variant<EpsilonType, SymbolType>&);
	template<class SymbolType>
	static void composeTransitionOutputSymbolMultiple(std::deque<sax::Token>&, const std::vector<SymbolType>& symbols);

	static void composeTransitionInputRegexp(std::deque<sax::Token>&, const regexp::RegExp& regexp);

	static void composeTransitionShift(std::deque<sax::Token>&, Shift shift);
};

template < class StateType >
void AutomatonToXMLComposer::composeStates(std::deque<sax::Token>& out, const std::set<StateType>& states) {
	out.emplace_back("states", sax::Token::TokenType::START_ELEMENT);
	for (const auto& state : states) {
		alib::xmlApi<StateType>::compose(out, state);
	}
	out.emplace_back("states", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void AutomatonToXMLComposer::composeInputAlphabet(std::deque<sax::Token>& out, const std::set<SymbolType>& symbols) {
	out.emplace_back("inputAlphabet", sax::Token::TokenType::START_ELEMENT);
	for (const auto& symbol : symbols) {
		alib::xmlApi<SymbolType>::compose(out, symbol);
	}
	out.emplace_back("inputAlphabet", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void AutomatonToXMLComposer::composeCallInputAlphabet(std::deque<sax::Token>& out, const std::set<SymbolType>& symbols) {
	out.emplace_back("callInputAlphabet", sax::Token::TokenType::START_ELEMENT);
	for (const auto& symbol : symbols) {
		alib::xmlApi<SymbolType>::compose(out, symbol);
	}
	out.emplace_back("callInputAlphabet", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void AutomatonToXMLComposer::composeReturnInputAlphabet(std::deque<sax::Token>& out, const std::set<SymbolType>& symbols) {
	out.emplace_back("returnInputAlphabet", sax::Token::TokenType::START_ELEMENT);
	for (const auto& symbol : symbols) {
		alib::xmlApi<SymbolType>::compose(out, symbol);
	}
	out.emplace_back("returnInputAlphabet", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void AutomatonToXMLComposer::composeLocalInputAlphabet(std::deque<sax::Token>& out, const std::set<SymbolType>& symbols) {
	out.emplace_back("localInputAlphabet", sax::Token::TokenType::START_ELEMENT);
	for (const auto& symbol : symbols) {
		alib::xmlApi<SymbolType>::compose(out, symbol);
	}
	out.emplace_back("localInputAlphabet", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void AutomatonToXMLComposer::composeRankedInputAlphabet(std::deque<sax::Token>& out, const std::set<SymbolType>& symbols) {
	out.emplace_back("rankedInputAlphabet", sax::Token::TokenType::START_ELEMENT);
	for (const auto& symbol : symbols) {
		alib::xmlApi<SymbolType>::compose(out, symbol);
	}
	out.emplace_back("rankedInputAlphabet", sax::Token::TokenType::END_ELEMENT);
}

template < class StateType >
void AutomatonToXMLComposer::composeInitialStates(std::deque<sax::Token>& out, const std::set<StateType>& states) {
	out.emplace_back("initialStates", sax::Token::TokenType::START_ELEMENT);
	for (const auto& state : states) {
		alib::xmlApi<StateType>::compose(out, state);
	}
	out.emplace_back("initialStates", sax::Token::TokenType::END_ELEMENT);
}

template < class StateType >
void AutomatonToXMLComposer::composeInitialState(std::deque<sax::Token>& out, const StateType& state) {
	out.emplace_back("initialState", sax::Token::TokenType::START_ELEMENT);
	alib::xmlApi<StateType>::compose(out, state);
	out.emplace_back("initialState", sax::Token::TokenType::END_ELEMENT);
}

template < class StateType >
void AutomatonToXMLComposer::composeFinalStates(std::deque<sax::Token>& out, const std::set<StateType>& states) {
	out.emplace_back("finalStates", sax::Token::TokenType::START_ELEMENT);
	for (const auto& state : states) {
		alib::xmlApi<StateType>::compose(out, state);
	}
	out.emplace_back("finalStates", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void AutomatonToXMLComposer::composePushdownStoreAlphabet(std::deque<sax::Token>& out, const std::set<SymbolType>& symbols) {
	out.emplace_back("pushdownStoreAlphabet", sax::Token::TokenType::START_ELEMENT);
	for (const auto& symbol : symbols) {
		alib::xmlApi<SymbolType>::compose(out, symbol);
	}
	out.emplace_back("pushdownStoreAlphabet", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void AutomatonToXMLComposer::composeInitialPushdownStoreSymbols(std::deque<sax::Token>& out, const std::set<SymbolType>& symbols) {
	out.emplace_back("initialPushdownStoreSymbols", sax::Token::TokenType::START_ELEMENT);
	for (const auto& symbol : symbols) {
		alib::xmlApi<SymbolType>::compose(out, symbol);
	}
	out.emplace_back("initialPushdownStoreSymbols", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void AutomatonToXMLComposer::composeInitialPushdownStoreSymbol(std::deque<sax::Token>& out, const SymbolType& symbol) {
	out.emplace_back("initialPushdownStoreSymbol", sax::Token::TokenType::START_ELEMENT);
	alib::xmlApi<SymbolType>::compose(out, symbol);
	out.emplace_back("initialPushdownStoreSymbol", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void AutomatonToXMLComposer::composeOutputAlphabet(std::deque<sax::Token>& out, const std::set<SymbolType>& symbols) {
	out.emplace_back("outputAlphabet", sax::Token::TokenType::START_ELEMENT);
	for (const auto& symbol : symbols) {
		alib::xmlApi<SymbolType>::compose(out, symbol);
	}
	out.emplace_back("outputAlphabet", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void AutomatonToXMLComposer::composeTapeAlphabet(std::deque<sax::Token>& out, const std::set<SymbolType>& symbols) {
	out.emplace_back("tapeAlphabet", sax::Token::TokenType::START_ELEMENT);
	for (const auto& symbol : symbols) {
		alib::xmlApi<SymbolType>::compose(out, symbol);
	}
	out.emplace_back("tapeAlphabet", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void AutomatonToXMLComposer::composeBlankSymbol(std::deque<sax::Token>& out, const SymbolType& symbol) {
	out.emplace_back("blankSymbol", sax::Token::TokenType::START_ELEMENT);
	alib::xmlApi<SymbolType>::compose(out, symbol);
	out.emplace_back("blankSymbol", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void AutomatonToXMLComposer::composeBottomOfTheStackSymbol(std::deque<sax::Token>& out, const SymbolType& symbol) {
	out.emplace_back("bottomOfTheStackSymbol", sax::Token::TokenType::START_ELEMENT);
	alib::xmlApi<SymbolType>::compose(out, symbol);
	out.emplace_back("bottomOfTheStackSymbol", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void AutomatonToXMLComposer::composeInputToPushdownStoreOperation(std::deque<sax::Token>& out, const std::map<SymbolType, std::pair<std::vector<SymbolType>, std::vector<SymbolType> > >& operations) {
	out.emplace_back("inputToPushdownStoreOperations", sax::Token::TokenType::START_ELEMENT);
	for(const auto& pushdownStoreOperation : operations) {
		out.emplace_back("operation", sax::Token::TokenType::START_ELEMENT);

		alib::xmlApi<SymbolType>::compose(out, pushdownStoreOperation.first);

		composeTransitionPop(out, pushdownStoreOperation.second.first);
		composeTransitionPush(out, pushdownStoreOperation.second.second);

		out.emplace_back("operation", sax::Token::TokenType::END_ELEMENT);
	}
	out.emplace_back("inputToPushdownStoreOperations", sax::Token::TokenType::END_ELEMENT);
}

template < class StateType >
void AutomatonToXMLComposer::composeTransitionTo(std::deque<sax::Token>& out, const StateType& state) {
	out.emplace_back("to", sax::Token::TokenType::START_ELEMENT);
	alib::xmlApi<StateType>::compose(out, state);
	out.emplace_back("to", sax::Token::TokenType::END_ELEMENT);
}

template < class StateType >
void AutomatonToXMLComposer::composeTransitionFrom(std::deque<sax::Token>& out, const StateType& state) {
	out.emplace_back("from", sax::Token::TokenType::START_ELEMENT);
	alib::xmlApi<StateType>::compose(out, state);
	out.emplace_back("from", sax::Token::TokenType::END_ELEMENT);
}

template < class StateType >
void AutomatonToXMLComposer::composeTransitionFrom(std::deque<sax::Token>& out, const std::vector<StateType> & states) {
	out.emplace_back("from", sax::Token::TokenType::START_ELEMENT);
	for (const auto& state : states) {
		alib::xmlApi<StateType>::compose(out, state);
	}
	out.emplace_back("from", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void AutomatonToXMLComposer::composeTransitionPop(std::deque<sax::Token>& out, const std::vector<SymbolType>& symbols) {
	out.emplace_back("pop", sax::Token::TokenType::START_ELEMENT);
	for (const auto& symbol : symbols) {
		alib::xmlApi<SymbolType>::compose(out, symbol);
	}
	out.emplace_back("pop", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void AutomatonToXMLComposer::composeTransitionSinglePop(std::deque<sax::Token>& out, const SymbolType& symbol) {
	out.emplace_back("pop", sax::Token::TokenType::START_ELEMENT);
	alib::xmlApi<SymbolType>::compose(out, symbol);
	out.emplace_back("pop", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void AutomatonToXMLComposer::composeTransitionPush(std::deque<sax::Token>& out, const std::vector<SymbolType>& symbols) {
	out.emplace_back("push", sax::Token::TokenType::START_ELEMENT);
	for (const auto& symbol : symbols) {
		alib::xmlApi<SymbolType>::compose(out, symbol);
	}
	out.emplace_back("push", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void AutomatonToXMLComposer::composeTransitionSinglePush(std::deque<sax::Token>& out, const SymbolType& symbol) {
	out.emplace_back("push", sax::Token::TokenType::START_ELEMENT);
	alib::xmlApi<SymbolType>::compose(out, symbol);
	out.emplace_back("push", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void AutomatonToXMLComposer::composeTransitionInputSymbol(std::deque<sax::Token>& out, const SymbolType& symbol) {
	out.emplace_back("input", sax::Token::TokenType::START_ELEMENT);
	alib::xmlApi<SymbolType>::compose(out, symbol);
	out.emplace_back("input", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void AutomatonToXMLComposer::composeTransitionOutputSymbol(std::deque<sax::Token>& out, const SymbolType& symbol) {
	out.emplace_back("output", sax::Token::TokenType::START_ELEMENT);
	alib::xmlApi<SymbolType>::compose(out, symbol);
	out.emplace_back("output", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void AutomatonToXMLComposer::composeTransitionOutputSymbolMultiple(std::deque<sax::Token>& out, const std::vector<SymbolType>& symbols) {
	out.emplace_back("output", sax::Token::TokenType::START_ELEMENT);
	for (const auto& symbol : symbols) {
		alib::xmlApi<SymbolType>::compose(out, symbol);
	}
	out.emplace_back("output", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void AutomatonToXMLComposer::composeTransitionInputSymbolMultiple(std::deque<sax::Token>& out, const std::vector<SymbolType>& symbols) {
	out.emplace_back("input", sax::Token::TokenType::START_ELEMENT);
	for (const auto& symbol : symbols) {
		alib::xmlApi<SymbolType>::compose(out, symbol);
	}
	out.emplace_back("input", sax::Token::TokenType::END_ELEMENT);
}

template < class EpsilonType, class SymbolType >
void AutomatonToXMLComposer::composeTransitionInputEpsilonSymbol(std::deque<sax::Token>& out, const std::variant<EpsilonType, SymbolType>& symbol) {
	out.emplace_back("input", sax::Token::TokenType::START_ELEMENT);
	if(symbol.template is<EpsilonType>()) {
		out.emplace_back("epsilon", sax::Token::TokenType::START_ELEMENT);
		out.emplace_back("epsilon", sax::Token::TokenType::END_ELEMENT);
	} else {
		alib::xmlApi<SymbolType>::compose(out, symbol.template get<SymbolType>());
	}
	out.emplace_back("input", sax::Token::TokenType::END_ELEMENT);
}

template < class EpsilonType, class SymbolType >
void AutomatonToXMLComposer::composeTransitionOutputEpsilonSymbol(std::deque<sax::Token>& out, const std::variant<EpsilonType, SymbolType>& symbol) {
	out.emplace_back("output", sax::Token::TokenType::START_ELEMENT);
	if(symbol.template is<EpsilonType>()) {
		out.emplace_back("epsilon", sax::Token::TokenType::START_ELEMENT);
		out.emplace_back("epsilon", sax::Token::TokenType::END_ELEMENT);
	} else {
		alib::xmlApi<SymbolType>::compose(out, symbol.template get<SymbolType>());
	}
	out.emplace_back("output", sax::Token::TokenType::END_ELEMENT);
}

} /* namespace automaton */

#endif /* AUTOMATON_TO_XML_COMPOSER_H_ */

