/*
 * ObjectsTree.h
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef OBJECTS_TREE_H_
#define OBJECTS_TREE_H_

#include <tree>
#include <string>

#include <object/Object.h>
#include "ContainerBase.h"

#include <XmlApi.hpp>

namespace container {

/**
 * Basic container from which are derived all other containers.
 * Contains reason why the container occured.
 */
class ObjectsTree : public std::tree < alib::Object >, public ContainerBase {
public:
	explicit ObjectsTree ( std::tree < alib::Object > );

	virtual ContainerBase * clone ( ) const;

	virtual ContainerBase * plunder ( ) &&;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const ObjectsTree & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName ( ) {
		static std::string xmlTagName = "Tree";

		return xmlTagName;
	}

	static ObjectsTree parse ( std::deque < sax::Token >::iterator & input );
	template < class T >
	static std::tree < T > parseRaw ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;
	template < class T >
	static void compose ( std::deque < sax::Token > & out, const std::tree < T > & input );
};

template < class T >
std::tree < T > ObjectsTree::parseRaw ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, ObjectsTree::getXmlTagName ( ) );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "Root" );

	std::tree < T > tree ( alib::xmlApi < T >::parse ( input ) );

	typename std::tree < T >::const_children_iterator under = tree.root ( );

	unsigned level = 0;

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) ) {
		if ( sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, "Children" ) ) {
			sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "Children" );

			if ( level > 0 )
				under = std::prev ( under->end ( ) );

			++level;
		}

		tree.push_back ( under, alib::xmlApi < T >::parse ( input ) );

		while ( sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::END_ELEMENT, "Children" ) ) {
			sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "Children" );

			under = under->parent ( );
			--level;
		}
	}

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "Root" );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, ObjectsTree::getXmlTagName ( ) );
	return tree;
}

template < class T >
void ObjectsTree::compose ( std::deque < sax::Token > & out, const std::tree < T > & input ) {
	out.emplace_back ( ObjectsTree::getXmlTagName ( ), sax::Token::TokenType::START_ELEMENT );
	out.emplace_back ( "Root", sax::Token::TokenType::START_ELEMENT );

	unsigned level = 0;

	for ( typename tree < T >::const_prefix_iterator iter = input.prefix_begin ( ); iter != input.prefix_end ( ); ) {
		while ( iter.getLevel ( ) > level ) {
			out.emplace_back ( "Children", sax::Token::TokenType::START_ELEMENT );
			++level;
		}

		alib::xmlApi < T >::compose ( out, * iter );
		++iter;

		while ( iter.getLevel ( ) < level ) {
			out.emplace_back ( "Children", sax::Token::TokenType::END_ELEMENT );
			--level;
		}
	}

	out.emplace_back ( "Root", sax::Token::TokenType::END_ELEMENT );
	out.emplace_back ( ObjectsTree::getXmlTagName ( ), sax::Token::TokenType::END_ELEMENT );
}

} /* namespace container */

namespace alib {

template < typename T >
struct xmlApi < std::tree < T > > {
	static std::tree < T > parse ( std::deque < sax::Token >::iterator & input );
	static bool first ( const std::deque < sax::Token >::const_iterator & input );
	static void compose ( std::deque < sax::Token > & output, const std::tree < T > & data );
};

template < typename T >
std::tree < T > xmlApi < std::tree < T > >::parse ( std::deque < sax::Token >::iterator & input ) {
	return container::ObjectsTree::parseRaw < T > ( input );
}

template < typename T >
bool xmlApi < std::tree < T > >::first ( const std::deque < sax::Token >::const_iterator & input ) {
	return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, container::ObjectsTree::getXmlTagName ( ) );
}

template < typename T >
void xmlApi < std::tree < T > >::compose ( std::deque < sax::Token > & output, const std::tree < T > & input ) {
	return container::ObjectsTree::compose < T > ( output, input );
}

} /* namespace alib */

#endif /* OBJECTS_TREE_H_ */
