/*
 * LinearString.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef LINEAR_STRING_H_
#define LINEAR_STRING_H_

#include <set>
#include <vector>
#include <sstream>
#include <algorithm>

#include <sax/FromXMLParserHelper.h>
#include <core/components.hpp>

#include "StringBase.h"
#include "StringFeatures.h"
#include "common/StringFromXMLParser.h"
#include "common/StringToXMLComposer.h"
#include "common/StringAuxiliary.h"
#include <exception/CommonException.h>

#include "Epsilon.h"
#include "../tree/ranked/PrefixRankedTree.h"
#include "../tree/ranked/PrefixRankedPattern.h"
#include "../tree/ranked/PrefixRankedNonlinearPattern.h"
#include "../tree/ranked/PrefixRankedBarTree.h"
#include "../tree/ranked/PrefixRankedBarPattern.h"
#include "../tree/ranked/PrefixRankedBarNonlinearPattern.h"
#include "../tree/unranked/PrefixBarTree.h"
#include "../tree/common/TreeAuxiliary.h"

namespace tree {

class PrefixRankedTree;
class PrefixRankedPattern;
class PrefixRankedNonlinearPattern;
class PrefixRankedBarTree;
class PrefixRankedBarPattern;
class PrefixRankedBarNonlinearPattern;
class PrefixBarTree;

}

namespace string {

class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of LinearStringElement.
 */
template < class SymbolType >
class LinearString : public StringBase, public std::Components < LinearString < SymbolType >, SymbolType, std::tuple < GeneralAlphabet >, std::tuple < > > {
	std::vector < SymbolType > m_Data;

public:
	explicit LinearString ( );
	explicit LinearString ( const tree::PrefixRankedTree & tree );
	explicit LinearString ( const tree::PrefixRankedPattern & tree );
	explicit LinearString ( const tree::PrefixRankedNonlinearPattern & tree );
	explicit LinearString ( const tree::PrefixRankedBarTree & tree );
	explicit LinearString ( const tree::PrefixRankedBarPattern & tree );
	explicit LinearString ( const tree::PrefixRankedBarNonlinearPattern & tree );
	explicit LinearString ( const tree::PrefixBarTree & tree );
	explicit LinearString ( std::set < SymbolType > alphabet, std::vector < SymbolType > data );
	explicit LinearString ( std::vector < SymbolType > data );
	explicit LinearString ( const std::string & string );
	explicit LinearString ( const char * string );
	explicit LinearString ( const Epsilon < > & epsilon );

	virtual StringBase * clone ( ) const;
	virtual StringBase * plunder ( ) &&;

	virtual const std::set < SymbolType > & getAlphabet ( ) const {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	void extendAlphabet ( const std::set < SymbolType > & symbols ) {
		this->template accessComponent < GeneralAlphabet > ( ).add( symbols );
	}

	/**
	 * @param element to append
	 */
	void appendSymbol ( SymbolType symbol );

	/**
	 * @return List of symbols forming string (const version).
	 */
	const std::vector < SymbolType > & getContent ( ) const;

	void setContent ( std::vector < SymbolType > data );

	/**
	 * @return true if string is an empty word (vector length is 0)
	 */
	bool isEmpty ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const LinearString & other ) const;

	virtual void operator >>( std::ostream & out ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "LinearString";

		return xmlTagName;
	}

	static LinearString parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;
};

template < class SymbolType >
LinearString < SymbolType >::LinearString(std::set<SymbolType> alphabet, std::vector<SymbolType> data) : std::Components < LinearString < SymbolType >, SymbolType, std::tuple < GeneralAlphabet >, std::tuple < > > ( std::make_tuple ( std::move ( alphabet ) ), std::tuple < > ( ) ) {
	setContent(std::move(data));
}

template < class SymbolType >
LinearString < SymbolType >::LinearString() : LinearString ( std::set < SymbolType > ( ), std::vector < SymbolType > ( ) ) {
}

template < class SymbolType >
LinearString < SymbolType >::LinearString(std::vector<SymbolType> data) : LinearString ( std::set < SymbolType> ( data.begin(), data.end() ), data ) {
}

template < class SymbolType >
LinearString < SymbolType >::LinearString(const std::string& string) : LinearString ( StringAuxiliary::toInternal ( string ) ) {
}

template < class SymbolType >
LinearString < SymbolType >::LinearString(const char* string) : LinearString((std::string) string) {
}

template < class SymbolType >
LinearString < SymbolType >::LinearString(const Epsilon < > & epsilon) : LinearString ( epsilon.getAlphabet( ), std::vector < SymbolType > ( ) ) {
}

template < class SymbolType >
LinearString < SymbolType >::LinearString ( const tree::PrefixRankedTree & tree ) : LinearString ( StringAuxiliary::wrapSymbols ( tree.getAlphabet ( ) ), StringAuxiliary::wrapSymbols ( tree.getContent ( ) ) ) {
}

template < class SymbolType >
LinearString < SymbolType >::LinearString ( const tree::PrefixRankedPattern & tree ) : LinearString ( StringAuxiliary::wrapSymbols ( tree.getAlphabet ( ) ), StringAuxiliary::wrapSymbols ( tree.getContent ( ) ) ) {
}

template < class SymbolType >
LinearString < SymbolType >::LinearString ( const tree::PrefixRankedNonlinearPattern & tree ) : LinearString ( StringAuxiliary::wrapSymbols ( tree.getAlphabet ( ) ), StringAuxiliary::wrapSymbols ( tree.getContent ( ) ) ) {
}

template < class SymbolType >
LinearString < SymbolType >::LinearString ( const tree::PrefixRankedBarTree & tree ) : LinearString ( StringAuxiliary::wrapSymbols ( tree.getAlphabet ( ) ), StringAuxiliary::wrapSymbols ( tree.getContent ( ) ) ) {
}

template < class SymbolType >
LinearString < SymbolType >::LinearString ( const tree::PrefixRankedBarPattern & tree ) : LinearString ( StringAuxiliary::wrapSymbols ( tree.getAlphabet ( ) ), StringAuxiliary::wrapSymbols ( tree.getContent ( ) ) ) {
}

template < class SymbolType >
LinearString < SymbolType >::LinearString ( const tree::PrefixRankedBarNonlinearPattern & tree ) : LinearString ( StringAuxiliary::wrapSymbols ( tree.getAlphabet ( ) ), StringAuxiliary::wrapSymbols ( tree.getContent ( ) ) ) {
}

template < class SymbolType >
LinearString < SymbolType >::LinearString ( const tree::PrefixBarTree & tree ) : LinearString ( tree.getAlphabet ( ), tree.getContent ( ) ) {
}

template < class SymbolType >
StringBase * LinearString < SymbolType >::clone ( ) const {
	return new LinearString ( * this );
}

template < class SymbolType >
StringBase * LinearString < SymbolType >::plunder ( ) && {
	return new LinearString ( std::move ( * this ) );
}

template < class SymbolType >
void LinearString < SymbolType >::appendSymbol ( SymbolType symbol ) {
	if ( getAlphabet().count ( symbol ) == 0 )
		throw exception::CommonException ( "Input symbol \"" + std::to_string ( symbol ) + "\" not in the alphabet." );

	m_Data.push_back ( std::move ( symbol ) );
}

template < class SymbolType >
const std::vector < SymbolType > & LinearString < SymbolType >::getContent ( ) const {
	return this->m_Data;
}

template < class SymbolType >
void LinearString < SymbolType >::setContent ( std::vector < SymbolType > data ) {
	std::set < SymbolType > minimalAlphabet ( data.begin ( ), data.end ( ) );
	std::set < SymbolType > unknownSymbols;
	std::set_difference ( minimalAlphabet.begin ( ), minimalAlphabet.end ( ), getAlphabet().begin ( ), getAlphabet().end ( ), std::inserter ( unknownSymbols, unknownSymbols.end ( ) ) );

	if ( unknownSymbols.size ( ) > 0 )
		throw exception::CommonException ( "Input symbols not in the alphabet." );

	this->m_Data = std::move ( data );
}

template < class SymbolType >
bool LinearString < SymbolType >::isEmpty ( ) const {
	return this->m_Data.size ( ) == 0;
}

template < class SymbolType >
int LinearString < SymbolType >::compare ( const LinearString & other ) const {
	auto first = std::tie ( m_Data, getAlphabet() );
	auto second = std::tie ( other.m_Data, other.getAlphabet() );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType >
void LinearString < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(LinearString ";

	for ( const SymbolType & symbol : this->m_Data )
		out << symbol;

	out << ")";
}

template < class SymbolType >
LinearString < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << "\"";

	for ( const SymbolType & symbol : this->m_Data )
		ss << symbol;

	ss << "\"";
	return std::move ( ss ).str ( );
}

template < class SymbolType >
LinearString < SymbolType > LinearString < SymbolType >::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, LinearString::getXmlTagName() );
	std::set < SymbolType > alphabet = StringFromXMLParser::parseAlphabet < SymbolType > ( input );
	std::vector < SymbolType > content = StringFromXMLParser::parseContent < SymbolType > ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, LinearString::getXmlTagName() );

	return LinearString < SymbolType > ( alphabet, content );
}

template < class SymbolType >
void LinearString < SymbolType >::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( LinearString::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	StringToXMLComposer::compose ( out, getAlphabet() );
	StringToXMLComposer::compose ( out, m_Data );
	out.emplace_back ( LinearString::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

} /* namespace string */

namespace std {

template < class SymbolType >
class ComponentConstraint< ::string::LinearString < SymbolType >, SymbolType, ::string::GeneralAlphabet > {
public:
	static bool used ( const ::string::LinearString < SymbolType > & string, const SymbolType & symbol ) {
		const std::vector<SymbolType>& content = string.getContent ( );
		return std::find(content.begin(), content.end(), symbol) != content.end();
	}

	static bool available ( const ::string::LinearString < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const ::string::LinearString < SymbolType > &, const SymbolType & ) {
	}
};

} /* namespace std */

#endif /* LINEAR_STRING_H_ */
