/*
 * PrefixRankedBarPattern.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef PREFIX_RANKED_BAR_PATTERN_H_
#define PREFIX_RANKED_BAR_PATTERN_H_

#include <set>
#include <vector>
#include <tree>
#include <core/components.hpp>

#include "../TreeException.h"
#include "../../alphabet/RankedSymbol.h"
#include "../RankedTreeBase.h"

namespace tree {

class RankedPattern;
class PrefixRankedBarTree;
class GeneralAlphabet;
class SubtreeWildcard;
class BarSymbols;
class VariablesBarSymbol;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of LinearStringElement.
 */
class PrefixRankedBarPattern : public RankedTreeBase, public std::Components < PrefixRankedBarPattern, alphabet::RankedSymbol, std::tuple < GeneralAlphabet, BarSymbols >, std::tuple < SubtreeWildcard, VariablesBarSymbol > > {
	std::vector < alphabet::RankedSymbol > m_Data;

	static std::vector < alphabet::RankedSymbol > toPrefixRankedBar ( const std::tree < alphabet::RankedSymbol > & node, const alphabet::RankedSymbol & subtreeWildcard, const alphabet::Symbol & barBase, const alphabet::RankedSymbol & variablesBar );

public:
	explicit PrefixRankedBarPattern ( std::set < alphabet::RankedSymbol > bar, alphabet::RankedSymbol variablesBar, alphabet::RankedSymbol subtreeWildcard, std::set < alphabet::RankedSymbol > alphabet, std::vector < alphabet::RankedSymbol > data );
	explicit PrefixRankedBarPattern ( std::set < alphabet::RankedSymbol > bar, alphabet::RankedSymbol variablesBar, alphabet::RankedSymbol subtreeWildcard, std::vector < alphabet::RankedSymbol > data );
	explicit PrefixRankedBarPattern ( alphabet::Symbol barBase, alphabet::RankedSymbol variablesBar, const RankedPattern & tree );
	explicit PrefixRankedBarPattern ( const PrefixRankedBarTree & tree );
	explicit PrefixRankedBarPattern ( const RankedPattern & tree );

	virtual RankedTreeBase * clone ( ) const;
	virtual RankedTreeBase * plunder ( ) &&;

	/**
	 * @return List of symbols forming tree (const version).
	 */
	const std::vector < alphabet::RankedSymbol > & getContent ( ) const;

	const std::set < alphabet::RankedSymbol > & getAlphabet ( ) const {
		return accessComponent < GeneralAlphabet > ( ).get ( );
	}

	void extendAlphabet ( const std::set < alphabet::RankedSymbol > & symbols ) {
		accessComponent < GeneralAlphabet > ( ).add ( symbols );
	}

	const std::set < alphabet::RankedSymbol > & getBars ( ) const {
		return accessComponent < BarSymbols > ( ).get ( );
	}

	const alphabet::RankedSymbol & getSubtreeWildcard ( ) const {
		return accessElement < SubtreeWildcard > ( ).get ( );
	}

	const alphabet::RankedSymbol & getVariablesBar ( ) const {
		return accessElement < VariablesBarSymbol > ( ).get ( );
	}

	void setContent ( std::vector < alphabet::RankedSymbol > data );

	void arityChecksum ( const std::vector < alphabet::RankedSymbol > & data );

	/**
	 * @return true if tree is an empty word (vector length is 0)
	 */
	bool isEmpty ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const PrefixRankedBarPattern & other ) const;

	virtual void operator >>( std::ostream & out ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "PrefixRankedBarPattern";

		return xmlTagName;
	}

	static PrefixRankedBarPattern parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;
};

} /* namespace tree */

namespace std {

template < >
class ComponentConstraint< ::tree::PrefixRankedBarPattern, alphabet::RankedSymbol, ::tree::GeneralAlphabet > {
public:
	static bool used ( const ::tree::PrefixRankedBarPattern & pattern, const alphabet::RankedSymbol & symbol ) {
		const std::vector < alphabet::RankedSymbol > & content = pattern.getContent ( );

		return std::find ( content.begin ( ), content.end ( ), symbol ) != content.end ( ) || pattern.accessElement < ::tree::VariablesBarSymbol > ( ).get ( ) == symbol || pattern.accessComponent < ::tree::BarSymbols > ( ).get ( ).count ( symbol ) || pattern.accessElement < ::tree::SubtreeWildcard > ( ).get ( ) == symbol;
	}

	static bool available ( const ::tree::PrefixRankedBarPattern &, const alphabet::RankedSymbol & ) {
		return true;
	}

	static void valid ( const ::tree::PrefixRankedBarPattern &, const alphabet::RankedSymbol & ) {
	}
};

template < >
class ComponentConstraint< ::tree::PrefixRankedBarPattern, alphabet::RankedSymbol, ::tree::BarSymbols > {
public:
	static bool used ( const ::tree::PrefixRankedBarPattern & pattern, const alphabet::RankedSymbol & symbol ) {
		const std::vector < alphabet::RankedSymbol > & content = pattern.getContent ( );

		return std::find ( content.begin ( ), content.end ( ), symbol ) != content.end ( ) || pattern.accessElement < ::tree::VariablesBarSymbol > ( ).get ( ) == symbol;
	}

	static bool available ( const ::tree::PrefixRankedBarPattern & pattern, const alphabet::RankedSymbol & symbol ) {
		return pattern.accessComponent < ::tree::GeneralAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const ::tree::PrefixRankedBarPattern &, const alphabet::RankedSymbol & ) {
	}
};

template < >
class ElementConstraint< ::tree::PrefixRankedBarPattern, alphabet::RankedSymbol, ::tree::SubtreeWildcard > {
public:
	static bool available ( const ::tree::PrefixRankedBarPattern & pattern, const alphabet::RankedSymbol & symbol ) {
		return pattern.accessComponent < ::tree::GeneralAlphabet > ( ).get ( ).count ( symbol );
}

	static void valid ( const ::tree::PrefixRankedBarPattern &, const alphabet::RankedSymbol & symbol) {
		if( symbol.getRank().getData() != 0 )
			throw ::tree::TreeException ( "SubtreeWildcard symbol has nonzero arity" );
	}
};

template < >
class ElementConstraint< ::tree::PrefixRankedBarPattern, alphabet::RankedSymbol, ::tree::VariablesBarSymbol > {
public:
	static bool available ( const ::tree::PrefixRankedBarPattern & pattern, const alphabet::RankedSymbol & symbol ) {
		return pattern.accessComponent < ::tree::BarSymbols > ( ).get ( ).count ( symbol );
	}

	static void valid ( const ::tree::PrefixRankedBarPattern &, const alphabet::RankedSymbol & symbol) {
		if( symbol.getRank().getData() != 0 )
			throw ::tree::TreeException ( "VariablesBarSymbol has nonzero arity" );
	}
};

} /* namespace std */

#endif /* PREFIX_RANKED_BAR_PATTERN_H_ */
