/*
 * PrefixRankedBarTree.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "PrefixRankedBarTree.h"

#include <sstream>
#include <algorithm>
#include <deque>

#include "RankedTree.h"

#include <sax/FromXMLParserHelper.h>
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include "../common/TreeAuxiliary.h"
#include "../Tree.h"
#include "../RankedTreeWrapper.h"
#include <object/Object.h>
#include <XmlApi.hpp>
#include <cast/CastApi.hpp>

#include "../../alphabet/BarSymbol.h"

namespace tree {

PrefixRankedBarTree::PrefixRankedBarTree ( std::set < alphabet::RankedSymbol > bars, std::set < alphabet::RankedSymbol > alphabet, std::vector < alphabet::RankedSymbol > data ) : std::Components < PrefixRankedBarTree, alphabet::RankedSymbol, std::tuple < GeneralAlphabet, BarSymbols >, std::tuple < > > ( std::make_tuple ( std::move ( alphabet ), std::move ( bars ) ), std::tuple < > ( ) ) {
	setContent ( std::move ( data ) );
}

PrefixRankedBarTree::PrefixRankedBarTree ( std::set < alphabet::RankedSymbol > bars, std::vector < alphabet::RankedSymbol > data ) : PrefixRankedBarTree ( bars, bars + std::set < alphabet::RankedSymbol > ( data.begin ( ), data.end ( ) ), data ) {
}

PrefixRankedBarTree::PrefixRankedBarTree ( alphabet::Symbol barBase, const RankedTree & tree ) : PrefixRankedBarTree ( TreeAuxiliary::computeBars ( tree.getAlphabet ( ), barBase ), tree.getAlphabet ( ) + TreeAuxiliary::computeBars ( tree.getAlphabet ( ), barBase ), toPrefixRankedBar ( tree.getContent ( ), barBase ) ) {
}

PrefixRankedBarTree::PrefixRankedBarTree ( const RankedTree & tree ) : PrefixRankedBarTree ( alphabet::BarSymbol::BAR_SYMBOL, tree ) {
}

std::vector < alphabet::RankedSymbol > PrefixRankedBarTree::toPrefixRankedBar ( const std::tree < alphabet::RankedSymbol > & tree, const alphabet::Symbol & barBase ) {
	std::vector < alphabet::RankedSymbol > res;

	for ( std::tree < alphabet::RankedSymbol >::const_structure_iterator iter = tree.structure_begin ( ); iter != tree.structure_end ( ); ++iter )
		if ( iter.getVirtual ( ) )
			res.push_back ( alphabet::RankedSymbol ( barBase, iter->getRank ( ) ) );
		else
			res.push_back ( * iter );

	return res;
}

RankedTreeBase * PrefixRankedBarTree::clone ( ) const {
	return new PrefixRankedBarTree ( * this );
}

RankedTreeBase * PrefixRankedBarTree::plunder ( ) && {
	return new PrefixRankedBarTree ( std::move ( * this ) );
}

const std::vector < alphabet::RankedSymbol > & PrefixRankedBarTree::getContent ( ) const {
	return this->m_Data;
}

void PrefixRankedBarTree::setContent ( std::vector < alphabet::RankedSymbol > data ) {
	arityChecksum ( data );

	std::set < alphabet::RankedSymbol > minimalAlphabet ( data.begin ( ), data.end ( ) );
	std::set < alphabet::RankedSymbol > unknownSymbols;
	std::set_difference ( minimalAlphabet.begin ( ), minimalAlphabet.end ( ), getAlphabet().begin ( ), getAlphabet().end ( ), std::inserter ( unknownSymbols, unknownSymbols.end ( ) ) );

	if ( unknownSymbols.size ( ) > 0 )
		throw TreeException ( "Input symbols not in the alphabet." );

	this->m_Data = std::move ( data );
}

void PrefixRankedBarTree::arityChecksum ( const std::vector < alphabet::RankedSymbol > & data ) {
	int arityChecksumTerminals = 1;
	int arityChecksumBars  = 1;
	int arityChecksumTypes = 0;

	for ( const alphabet::RankedSymbol & symbol : data ) {
		if ( getBars ( ).count ( symbol ) ) {
			arityChecksumBars += symbol.getRank ( ).getData ( );
			arityChecksumBars -= 1;
			arityChecksumTypes -= 1;
		} else {
			arityChecksumTerminals += symbol.getRank ( ).getData ( );
			arityChecksumTerminals -= 1;
			arityChecksumTypes += 1;
		}
	}

	if ( ( arityChecksumTerminals != 0 ) || ( arityChecksumBars != 0 ) || ( arityChecksumTypes != 0 ) ) throw TreeException ( "The string does not form a tree" );
}

bool PrefixRankedBarTree::isEmpty ( ) const {
	return this->m_Data.size ( ) == 0;
}

int PrefixRankedBarTree::compare ( const PrefixRankedBarTree & other ) const {
	auto first = std::tie ( m_Data, getAlphabet(), getBars() );
	auto second = std::tie ( other.m_Data, other.getAlphabet(), other.getBars() );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void PrefixRankedBarTree::operator >>( std::ostream & out ) const {
	out << "(PrefixRankedBarTree ";

	for ( const alphabet::RankedSymbol & symbol : this->m_Data )
		out << symbol;

	out << ")";
}

PrefixRankedBarTree::operator std::string ( ) const {
	std::stringstream ss;
	ss << "\"";

	for ( const alphabet::RankedSymbol & symbol : this->m_Data )
		ss << symbol;

	ss << "\"";
	return std::move ( ss ).str ( );
}

PrefixRankedBarTree PrefixRankedBarTree::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, PrefixRankedBarTree::getXmlTagName() );
	std::set < alphabet::RankedSymbol > bars = TreeFromXMLParser::parseBars ( input );
	std::set < alphabet::RankedSymbol > rankedAlphabet = TreeFromXMLParser::parseRankedAlphabet ( input );
	std::vector < alphabet::RankedSymbol > data = TreeFromXMLParser::parseRankedLinearContent ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, PrefixRankedBarTree::getXmlTagName() );

	return PrefixRankedBarTree ( std::move ( bars ), std::move ( rankedAlphabet ), std::move ( data ) );
}

void PrefixRankedBarTree::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( PrefixRankedBarTree::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	TreeToXMLComposer::composeBars ( out, getBars() );
	TreeToXMLComposer::composeAlphabet ( out, getAlphabet() );
	TreeToXMLComposer::composeContent ( out, m_Data );
	out.emplace_back ( PrefixRankedBarTree::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

} /* namespace tree */

namespace alib {

auto prefixRankedBarTreeParserRegister = xmlApi < ::tree::Tree >::ParserRegister < ::tree::PrefixRankedBarTree > ();
auto prefixRankedBarTreeParserRegister2 = xmlApi < ::tree::RankedTreeWrapper >::ParserRegister < ::tree::PrefixRankedBarTree > ();
auto prefixRankedBarTreeParserRegister3 = xmlApi < alib::Object >::ParserRegister < ::tree::PrefixRankedBarTree > ();

auto PrefixRankedBarTreeFromRankedTree = castApi::CastRegister < ::tree::PrefixRankedBarTree, ::tree::RankedTree > ( );
auto PrefixRankedBarTreeCastBinder = castApi::CastPoolStringBinder < ::tree::PrefixRankedBarTree > ( ::tree::PrefixRankedBarTree::getXmlTagName() );

} /* namespace alib */
