/*
 * PrefixRankedNonlinearPattern.h
 *
 *  Created on: Jul 20, 2016
 *      Author: Jan Travnicek
 */

#ifndef PREFIX_RANKED_NONLINEAR_PATTERN_H_
#define PREFIX_RANKED_NONLINEAR_PATTERN_H_

#include <set>
#include <vector>
#include <tree>
#include <core/components.hpp>

#include "../TreeException.h"
#include "../../alphabet/RankedSymbol.h"
#include "../RankedTreeBase.h"

namespace tree {

class RankedTree;
class RankedPattern;
class RankedNonlinearPattern;
class PrefixRankedTree;
class PrefixRankedPattern;

class GeneralAlphabet;
class SubtreeWildcard;
class NonlinearAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of LinearStringElement.
 */
class PrefixRankedNonlinearPattern : public RankedTreeBase, public std::Components < PrefixRankedNonlinearPattern, alphabet::RankedSymbol, std::tuple < GeneralAlphabet, NonlinearAlphabet >, std::tuple < SubtreeWildcard > > {
	std::vector < alphabet::RankedSymbol > m_Data;

	static std::vector < alphabet::RankedSymbol > toPrefixRanked ( const std::tree < alphabet::RankedSymbol > & tree );

public:
	explicit PrefixRankedNonlinearPattern ( alphabet::RankedSymbol subtreeWildcard, std::set < alphabet::RankedSymbol > nonlinearVariables, std::set < alphabet::RankedSymbol > alphabet, std::vector < alphabet::RankedSymbol > data );
	explicit PrefixRankedNonlinearPattern ( alphabet::RankedSymbol subtreeWildcard, std::set < alphabet::RankedSymbol > nonlinearVariables, std::vector < alphabet::RankedSymbol > data );
	explicit PrefixRankedNonlinearPattern ( alphabet::RankedSymbol subtreeWildcard, std::vector < alphabet::RankedSymbol > data );
	explicit PrefixRankedNonlinearPattern ( const PrefixRankedTree & tree );
	explicit PrefixRankedNonlinearPattern ( const PrefixRankedPattern & tree );
	explicit PrefixRankedNonlinearPattern ( const RankedTree & tree );
	explicit PrefixRankedNonlinearPattern ( const RankedPattern & tree );
	explicit PrefixRankedNonlinearPattern ( const RankedNonlinearPattern & tree );

	virtual RankedTreeBase * clone ( ) const;
	virtual RankedTreeBase * plunder ( ) &&;

	/**
	 * @return List of symbols forming tree (const version).
	 */
	const std::vector < alphabet::RankedSymbol > & getContent ( ) const;

	const std::set < alphabet::RankedSymbol > & getAlphabet ( ) const {
		return accessComponent < GeneralAlphabet > ( ).get ( );
	}

	void extendAlphabet ( const std::set < alphabet::RankedSymbol > & symbols ) {
		accessComponent < GeneralAlphabet > ( ).add ( symbols );
	}

	const alphabet::RankedSymbol & getSubtreeWildcard ( ) const {
		return accessElement < SubtreeWildcard > ( ).get ( );
	}

	const std::set < alphabet::RankedSymbol > & getNonlinearVariables ( ) const {
		return accessComponent < NonlinearAlphabet > ( ).get ( );
	}

	void setContent ( std::vector < alphabet::RankedSymbol > data );

	void arityChecksum ( const std::vector < alphabet::RankedSymbol > & data );

	/**
	 * @return true if tree is an empty word (vector length is 0)
	 */
	bool isEmpty ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const PrefixRankedNonlinearPattern & other ) const;

	virtual void operator >>( std::ostream & out ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName ( ) {
		static std::string xmlTagName = "PrefixRankedNonlinearPattern";

		return xmlTagName;
	}

	static PrefixRankedNonlinearPattern parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;
};

} /* namespace tree */

namespace std {

template < >
class ComponentConstraint< ::tree::PrefixRankedNonlinearPattern, alphabet::RankedSymbol, ::tree::GeneralAlphabet > {
public:
	static bool used ( const ::tree::PrefixRankedNonlinearPattern & pattern, const alphabet::RankedSymbol & symbol ) {
		const std::vector < alphabet::RankedSymbol > & content = pattern.getContent ( );

		return std::find ( content.begin ( ), content.end ( ), symbol ) != content.end ( ) || pattern.accessElement < ::tree::SubtreeWildcard > ( ).get ( ) == symbol;
	}

	static bool available ( const ::tree::PrefixRankedNonlinearPattern &, const alphabet::RankedSymbol & ) {
		return true;
	}

	static void valid ( const ::tree::PrefixRankedNonlinearPattern &, const alphabet::RankedSymbol & ) {
	}
};

template < >
class ComponentConstraint< ::tree::PrefixRankedNonlinearPattern, alphabet::RankedSymbol, ::tree::NonlinearAlphabet > {
public:
	static bool used ( const ::tree::PrefixRankedNonlinearPattern &, const alphabet::RankedSymbol & ) {
		return false;
	}

	static bool available ( const ::tree::PrefixRankedNonlinearPattern & pattern, const alphabet::RankedSymbol & symbol ) {
		return pattern.accessComponent < ::tree::GeneralAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const ::tree::PrefixRankedNonlinearPattern & pattern, const alphabet::RankedSymbol & symbol) {
		if ( symbol.getRank ( ).getData ( ) != 0 )
			throw ::tree::TreeException ( "SubtreeWildcard symbol has nonzero arity" );

		if ( pattern.accessElement < ::tree::SubtreeWildcard > ( ).get ( ) == symbol )
			throw ::tree::TreeException ( "Symbol " + std::to_string ( symbol ) + "cannot be set as nonlinear variable since it is already subtree wildcard" );
	}
};

template < >
class ElementConstraint< ::tree::PrefixRankedNonlinearPattern, alphabet::RankedSymbol, ::tree::SubtreeWildcard > {
public:
	static bool available ( const ::tree::PrefixRankedNonlinearPattern & pattern, const alphabet::RankedSymbol & symbol ) {
		return pattern.accessComponent < ::tree::GeneralAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const ::tree::PrefixRankedNonlinearPattern & pattern, const alphabet::RankedSymbol & symbol) {
		if ( symbol.getRank ( ).getData ( ) != 0 )
			throw ::tree::TreeException ( "SubtreeWildcard symbol has nonzero arity" );

		if ( pattern.accessComponent < ::tree::NonlinearAlphabet > ( ).get ( ).count ( symbol ) )
			throw ::tree::TreeException ( "Symbol " + std::to_string ( symbol ) + "cannot be set as subtree wildcard since it is already nonlinear variable" );
	}
};

} /* namespace std */

#endif /* PREFIX_RANKED_NONLINEAR_PATTERN_H_ */
