/*
 * PrefixRankedTree.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef PREFIX_RANKED_TREE_H_
#define PREFIX_RANKED_TREE_H_

#include <set>
#include <vector>
#include <tree>
#include <core/components.hpp>

#include "../TreeException.h"
#include "../../alphabet/RankedSymbol.h"
#include "../RankedTreeBase.h"

namespace tree {

class RankedTree;
class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of LinearStringElement.
 */
class PrefixRankedTree : public RankedTreeBase, public std::Components < PrefixRankedTree, alphabet::RankedSymbol, std::tuple < GeneralAlphabet >, std::tuple < > > {
	std::vector < alphabet::RankedSymbol > m_Data;

	static std::vector < alphabet::RankedSymbol > toPrefixRanked ( const std::tree < alphabet::RankedSymbol > & tree );

public:
	explicit PrefixRankedTree ( std::set < alphabet::RankedSymbol > alphabet, std::vector < alphabet::RankedSymbol > data );
	explicit PrefixRankedTree ( std::vector < alphabet::RankedSymbol > data );
	explicit PrefixRankedTree ( const RankedTree & tree );

	virtual RankedTreeBase * clone ( ) const;
	virtual RankedTreeBase * plunder ( ) &&;

	/**
	 * @return List of symbols forming tree (const version).
	 */
	const std::vector < alphabet::RankedSymbol > & getContent ( ) const;

	const std::set < alphabet::RankedSymbol > & getAlphabet ( ) const {
		return accessComponent < GeneralAlphabet > ( ).get ( );
	}

	void extendAlphabet ( const std::set < alphabet::RankedSymbol > & symbols ) {
		accessComponent < GeneralAlphabet > ( ).add ( symbols );
	}

	void setContent ( std::vector < alphabet::RankedSymbol > data );

	void arityChecksum ( const std::vector < alphabet::RankedSymbol > & data );

	/**
	 * @return true if tree is an empty word (vector length is 0)
	 */
	bool isEmpty ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const PrefixRankedTree & other ) const;

	virtual void operator >>( std::ostream & out ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "PrefixRankedTree";

		return xmlTagName;
	}

	static PrefixRankedTree parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;
};

} /* namespace tree */

namespace std {

template < >
class ComponentConstraint< ::tree::PrefixRankedTree, alphabet::RankedSymbol, ::tree::GeneralAlphabet > {
public:
	static bool used ( const ::tree::PrefixRankedTree & tree, const alphabet::RankedSymbol & symbol ) {
		const std::vector < alphabet::RankedSymbol > & content = tree.getContent ( );

		return std::find ( content.begin ( ), content.end ( ), symbol ) != content.end ( );
	}

	static bool available ( const ::tree::PrefixRankedTree &, const alphabet::RankedSymbol & ) {
		return true;
	}

	static void valid ( const ::tree::PrefixRankedTree &, const alphabet::RankedSymbol & ) {
	}
};

} /* namespace std */

#endif /* PREFIX_RANKED_TREE_H_ */
