/*
 * RankedNonlinearPattern.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "RankedNonlinearPattern.h"

#include <iostream>
#include <algorithm>
#include <sstream>

#include "../unranked/UnrankedNonlinearPattern.h"

#include <sax/FromXMLParserHelper.h>
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include "../common/TreeAuxiliary.h"
#include "../Tree.h"
#include "../RankedTreeWrapper.h"
#include <object/Object.h>
#include <XmlApi.hpp>
#include <cast/CastApi.hpp>

namespace tree {

RankedNonlinearPattern::RankedNonlinearPattern ( alphabet::RankedSymbol subtreeWildcard, std::set < alphabet::RankedSymbol > nonlinearVariables, std::set < alphabet::RankedSymbol > alphabet, std::tree < alphabet::RankedSymbol > pattern ) : std::Components < RankedNonlinearPattern, alphabet::RankedSymbol, std::tuple < GeneralAlphabet, NonlinearAlphabet >, std::tuple < SubtreeWildcard > > ( std::make_tuple ( std::move ( alphabet ), std::move ( nonlinearVariables ) ), std::make_tuple ( std::move ( subtreeWildcard ) ) ), content ( std::move ( pattern ) ) {
	checkAlphabet ( content );
	checkArities ( content );
}

RankedNonlinearPattern::RankedNonlinearPattern ( alphabet::RankedSymbol subtreeWildcard, std::set < alphabet::RankedSymbol > nonlinearVariables, std::tree < alphabet::RankedSymbol > pattern ) : RankedNonlinearPattern ( subtreeWildcard, nonlinearVariables, std::set < alphabet::RankedSymbol > ( pattern.prefix_begin ( ), pattern.prefix_end ( ) ) + nonlinearVariables + std::set < alphabet::RankedSymbol > { subtreeWildcard }, pattern ) {
}

RankedNonlinearPattern::RankedNonlinearPattern ( alphabet::RankedSymbol subtreeWildcard, std::tree < alphabet::RankedSymbol > pattern ) : RankedNonlinearPattern ( subtreeWildcard, std::set < alphabet::RankedSymbol > { }, pattern ) {
}

RankedNonlinearPattern::RankedNonlinearPattern ( const UnrankedNonlinearPattern & other ) : RankedNonlinearPattern ( alphabet::RankedSymbol ( other.getSubtreeWildcard ( ), 0 ), TreeAuxiliary::toRankedTree ( other.getContent ( ) ) ) {
}

RankedTreeBase * RankedNonlinearPattern::clone ( ) const {
	return new RankedNonlinearPattern ( * this );
}

RankedTreeBase * RankedNonlinearPattern::plunder ( ) && {
	return new RankedNonlinearPattern ( std::move ( * this ) );
}

const std::tree < alphabet::RankedSymbol > & RankedNonlinearPattern::getContent ( ) const {
	return content;
}

void RankedNonlinearPattern::checkAlphabet ( const std::tree < alphabet::RankedSymbol > & data ) const {
	std::set < alphabet::RankedSymbol > minimalAlphabet ( data.prefix_begin ( ), data.prefix_end ( ) );
	std::set < alphabet::RankedSymbol > unknownSymbols;
	std::set_difference ( minimalAlphabet.begin ( ), minimalAlphabet.end ( ), getAlphabet().begin ( ), getAlphabet().end ( ), std::inserter ( unknownSymbols, unknownSymbols.end ( ) ) );

	if ( unknownSymbols.size ( ) > 0 )
		throw exception::CommonException ( "Input symbols not in the alphabet." );
}

void RankedNonlinearPattern::checkArities ( const std::tree < alphabet::RankedSymbol > & data ) const {
	if ( data.getData ( ).getRank ( ).getData ( ) != data.getChildren ( ).size ( ) )
		throw exception::CommonException ( "Invalid rank." );

	for ( const std::tree < alphabet::RankedSymbol > & child : data )
		checkArities ( child );
}

void RankedNonlinearPattern::setTree ( std::tree < alphabet::RankedSymbol > pattern ) {
	checkAlphabet ( pattern );
	checkArities ( pattern );

	this->content = std::move ( pattern );
}

void RankedNonlinearPattern::operator >>( std::ostream & out ) const {
	out << "(RankedNonlinearPattern " << this->content << ")";
}

int RankedNonlinearPattern::compare ( const RankedNonlinearPattern & other ) const {
	auto first = std::tie ( content, getAlphabet(), getSubtreeWildcard(), getNonlinearVariables() );
	auto second = std::tie ( other.content, other.getAlphabet(), other.getSubtreeWildcard(), getNonlinearVariables() );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void RankedNonlinearPattern::nicePrint ( std::ostream & os ) const {
	content.nicePrint ( os );
}

RankedNonlinearPattern::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

RankedNonlinearPattern RankedNonlinearPattern::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, RankedNonlinearPattern::getXmlTagName() );
	alphabet::RankedSymbol subtreeWildcardSymbol = TreeFromXMLParser::parseSubtreeWildcardRankedSymbol ( input );
	std::set < alphabet::RankedSymbol > nonlinearVariables = TreeFromXMLParser::parseRankedNonlinearVariables ( input );
	std::set < alphabet::RankedSymbol > rankedAlphabet = TreeFromXMLParser::parseRankedAlphabet ( input );
	std::tree < alphabet::RankedSymbol > root = TreeFromXMLParser::parseRankedTreeContent ( input );
	RankedNonlinearPattern tree ( std::move ( subtreeWildcardSymbol ), std::move ( nonlinearVariables ), std::move ( rankedAlphabet ), std::move ( root ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, RankedNonlinearPattern::getXmlTagName() );
	return tree;
}

void RankedNonlinearPattern::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( RankedNonlinearPattern::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	TreeToXMLComposer::composeSubtreeWildcard ( out, getSubtreeWildcard() );
	TreeToXMLComposer::composeNonlinearVariables ( out, getNonlinearVariables() );
	TreeToXMLComposer::composeAlphabet ( out, getAlphabet() );
	TreeToXMLComposer::composeContent ( out, getContent ( ) );
	out.emplace_back ( RankedNonlinearPattern::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

} /* namespace tree */

namespace alib {

auto RankedNonlinearPatternParserRegister  = xmlApi < ::tree::Tree >::ParserRegister < ::tree::RankedNonlinearPattern > ( );
auto RankedNonlinearPatternParserRegister2 = xmlApi < ::tree::RankedTreeWrapper >::ParserRegister < ::tree::RankedNonlinearPattern > ( );
auto RankedNonlinearPatternParserRegister3 = xmlApi < alib::Object >::ParserRegister < ::tree::RankedNonlinearPattern > ( );

} /* namespace alib */
