/*
 * RankedPattern.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "RankedPattern.h"

#include <iostream>
#include <algorithm>
#include <sstream>

#include "../unranked/UnrankedPattern.h"

#include <sax/FromXMLParserHelper.h>
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include "../common/TreeAuxiliary.h"
#include "../Tree.h"
#include "../RankedTreeWrapper.h"
#include <object/Object.h>
#include <XmlApi.hpp>
#include <cast/CastApi.hpp>

namespace tree {

RankedPattern::RankedPattern ( alphabet::RankedSymbol subtreeWildcard, std::set < alphabet::RankedSymbol > alphabet, std::tree < alphabet::RankedSymbol > pattern ) : std::Components < RankedPattern, alphabet::RankedSymbol, std::tuple < GeneralAlphabet >, std::tuple < SubtreeWildcard > > ( std::make_tuple ( std::move ( alphabet ) ), std::make_tuple ( std::move ( subtreeWildcard ) ) ), content ( std::move ( pattern ) ) {
	checkAlphabet ( content );
	checkArities ( content );
}

RankedPattern::RankedPattern ( alphabet::RankedSymbol subtreeWildcard, std::tree < alphabet::RankedSymbol > pattern ) : RankedPattern ( subtreeWildcard, std::set < alphabet::RankedSymbol > ( pattern.prefix_begin ( ), pattern.prefix_end ( ) ) + std::set < alphabet::RankedSymbol > { subtreeWildcard }, pattern ) {
}

RankedPattern::RankedPattern ( const UnrankedPattern & other ) : RankedPattern ( alphabet::RankedSymbol ( other.getSubtreeWildcard ( ), 0 ), TreeAuxiliary::toRankedTree ( other.getContent ( ) ) ) {
}

RankedTreeBase * RankedPattern::clone ( ) const {
	return new RankedPattern ( * this );
}

RankedTreeBase * RankedPattern::plunder ( ) && {
	return new RankedPattern ( std::move ( * this ) );
}

const std::tree < alphabet::RankedSymbol > & RankedPattern::getContent ( ) const {
	return content;
}

void RankedPattern::checkAlphabet ( const std::tree < alphabet::RankedSymbol > & data ) const {
	std::set < alphabet::RankedSymbol > minimalAlphabet ( data.prefix_begin ( ), data.prefix_end ( ) );
	std::set < alphabet::RankedSymbol > unknownSymbols;
	std::set_difference ( minimalAlphabet.begin ( ), minimalAlphabet.end ( ), getAlphabet().begin ( ), getAlphabet().end ( ), std::inserter ( unknownSymbols, unknownSymbols.end ( ) ) );

	if ( unknownSymbols.size ( ) > 0 )
		throw exception::CommonException ( "Input symbols not in the alphabet." );
}

void RankedPattern::checkArities ( const std::tree < alphabet::RankedSymbol > & data ) const {
	if ( data.getData ( ).getRank ( ).getData ( ) != data.getChildren ( ).size ( ) )
		throw exception::CommonException ( "Invalid rank." );

	for ( const std::tree < alphabet::RankedSymbol > & child : data )
		checkArities ( child );
}

void RankedPattern::setTree ( std::tree < alphabet::RankedSymbol > pattern ) {
	checkAlphabet ( pattern );
	checkArities ( pattern );

	this->content = std::move ( pattern );
}

void RankedPattern::operator >>( std::ostream & out ) const {
	out << "(RankedPattern " << this->content << ")";
}

int RankedPattern::compare ( const RankedPattern & other ) const {
	auto first = std::tie ( content, getAlphabet(), getSubtreeWildcard() );
	auto second = std::tie ( other.content, other.getAlphabet(), other.getSubtreeWildcard() );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void RankedPattern::nicePrint ( std::ostream & os ) const {
	content.nicePrint ( os );
}

RankedPattern::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

RankedPattern RankedPattern::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, RankedPattern::getXmlTagName() );
	alphabet::RankedSymbol subtreeWildcardSymbol = TreeFromXMLParser::parseSubtreeWildcardRankedSymbol ( input );
	std::set < alphabet::RankedSymbol > rankedAlphabet = TreeFromXMLParser::parseRankedAlphabet ( input );
	std::tree < alphabet::RankedSymbol > root = TreeFromXMLParser::parseRankedTreeContent ( input );
	RankedPattern tree ( std::move ( subtreeWildcardSymbol ), std::move ( rankedAlphabet ), std::move ( root ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, RankedPattern::getXmlTagName() );
	return tree;
}

void RankedPattern::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( RankedPattern::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	TreeToXMLComposer::composeSubtreeWildcard ( out, getSubtreeWildcard() );
	TreeToXMLComposer::composeAlphabet ( out, getAlphabet() );
	TreeToXMLComposer::composeContent ( out, getContent ( ) );
	out.emplace_back ( RankedPattern::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

} /* namespace tree */

namespace alib {

auto rankedPatternParserRegister = xmlApi < ::tree::Tree >::ParserRegister < ::tree::RankedPattern > ();
auto rankedPatternParserRegister2 = xmlApi < ::tree::RankedTreeWrapper >::ParserRegister < ::tree::RankedPattern > ();
auto rankedPatternParserRegister3 = xmlApi < alib::Object >::ParserRegister < ::tree::RankedPattern > ();

auto RankedPatternFromUnrankedPattern = castApi::CastRegister < ::tree::RankedPattern, ::tree::UnrankedPattern > ( );
auto RankedPatternCastBinder = castApi::CastPoolStringBinder < ::tree::RankedPattern > ( ::tree::RankedPattern::getXmlTagName() );

} /* namespace alib */
