/*
 * RankedTree.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Stepan Plachy
 */

#include "RankedTree.h"

#include <iostream>
#include <algorithm>
#include <sstream>

#include "../unranked/UnrankedTree.h"

#include <sax/FromXMLParserHelper.h>
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include "../common/TreeAuxiliary.h"
#include "../Tree.h"
#include "../RankedTreeWrapper.h"
#include <object/Object.h>
#include <XmlApi.hpp>
#include <cast/CastApi.hpp>

namespace tree {

RankedTree::RankedTree ( std::set < alphabet::RankedSymbol > alphabet, std::tree < alphabet::RankedSymbol > tree ) : std::Components < RankedTree, alphabet::RankedSymbol, std::tuple < GeneralAlphabet >, std::tuple < > > ( std::make_tuple ( std::move ( alphabet ) ), std::tuple < > ( ) ), content ( std::move ( tree ) ) {
	checkAlphabet ( content );
	checkArities ( content );
}

RankedTree::RankedTree ( std::tree < alphabet::RankedSymbol > tree ) : RankedTree ( std::set < alphabet::RankedSymbol > ( tree.prefix_begin ( ), tree.prefix_end ( ) ), tree ) {
}

RankedTree::RankedTree ( const UnrankedTree & other ) : RankedTree ( TreeAuxiliary::toRankedTree ( other.getContent ( ) ) ) {
}

RankedTreeBase * RankedTree::clone ( ) const {
	return new RankedTree ( * this );
}

RankedTreeBase * RankedTree::plunder ( ) && {
	return new RankedTree ( std::move ( * this ) );
}

const std::tree < alphabet::RankedSymbol > & RankedTree::getContent ( ) const {
	return content;
}

void RankedTree::checkAlphabet ( const std::tree < alphabet::RankedSymbol > & data ) const {
	std::set < alphabet::RankedSymbol > minimalAlphabet ( data.prefix_begin ( ), data.prefix_end ( ) );
	std::set < alphabet::RankedSymbol > unknownSymbols;
	std::set_difference ( minimalAlphabet.begin ( ), minimalAlphabet.end ( ), getAlphabet().begin ( ), getAlphabet().end ( ), std::inserter ( unknownSymbols, unknownSymbols.end ( ) ) );

	if ( unknownSymbols.size ( ) > 0 )
		throw exception::CommonException ( "Input symbols not in the alphabet." );
}

void RankedTree::checkArities ( const std::tree < alphabet::RankedSymbol > & data ) const {
	if ( data.getData ( ).getRank ( ).getData ( ) != data.getChildren ( ).size ( ) )
		throw exception::CommonException ( "Invalid rank." );

	for ( const std::tree < alphabet::RankedSymbol > & child : data )
		checkArities ( child );
}

void RankedTree::setTree ( std::tree < alphabet::RankedSymbol > tree ) {
	checkAlphabet ( tree );
	checkArities ( tree );

	this->content = std::move ( tree );
}

void RankedTree::operator >>( std::ostream & out ) const {
	out << "(RankedTree " << this->content << ")";
}

int RankedTree::compare ( const RankedTree & other ) const {
	auto first = std::tie ( content, getAlphabet() );
	auto second = std::tie ( other.content, other.getAlphabet() );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void RankedTree::nicePrint ( std::ostream & os ) const {
	content.nicePrint ( os );
}

RankedTree::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

RankedTree RankedTree::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, RankedTree::getXmlTagName() );
	std::set < alphabet::RankedSymbol > rankedAlphabet = TreeFromXMLParser::parseRankedAlphabet ( input );
	std::tree < alphabet::RankedSymbol > root = TreeFromXMLParser::parseRankedTreeContent ( input );
	RankedTree tree ( std::move ( rankedAlphabet ), std::move ( root ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, RankedTree::getXmlTagName() );
	return tree;
}

void RankedTree::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( RankedTree::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	TreeToXMLComposer::composeAlphabet ( out, getAlphabet ( ) );
	TreeToXMLComposer::composeContent ( out, getContent ( ) );
	out.emplace_back ( RankedTree::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

} /* namespace tree */

namespace alib {

auto rankedTreeParserRegister = xmlApi < ::tree::Tree >::ParserRegister < ::tree::RankedTree > ();
auto rankedTreeParserRegister2 = xmlApi < ::tree::RankedTreeWrapper >::ParserRegister < ::tree::RankedTree > ();
auto rankedTreeParserRegister3 = xmlApi < alib::Object >::ParserRegister < ::tree::RankedTree > ();

auto RankedTreeFromUnrankedTree = castApi::CastRegister < ::tree::RankedTree, ::tree::UnrankedTree > ( );
auto RankedTreeCastBinder = castApi::CastPoolStringBinder < ::tree::RankedTree > ( ::tree::RankedTree::getXmlTagName() );

} /* namespace alib */
