/*
 * RankedTree.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Stepan Plachy
 */

#ifndef RANKED_TREE_H_
#define RANKED_TREE_H_

#include <string>
#include <set>
#include <tree>
#include <core/components.hpp>

#include "../TreeException.h"
#include "../RankedTreeBase.h"

namespace tree {

class UnrankedTree;
class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
class RankedTree : public RankedTreeBase, public std::Components < RankedTree, alphabet::RankedSymbol, std::tuple < GeneralAlphabet >, std::tuple < > > {
	std::tree < alphabet::RankedSymbol > content;

	void checkAlphabet ( const std::tree < alphabet::RankedSymbol > & pattern ) const;

	void checkArities ( const std::tree < alphabet::RankedSymbol > & data ) const;

public:
	/**
	 * @copydoc std::tree < alphabet::RankedSymbol >::clone() const
	 */
	virtual RankedTreeBase * clone ( ) const;

	/**
	 * @copydoc std::tree < alphabet::RankedSymbol >::plunder() const
	 */
	virtual RankedTreeBase * plunder ( ) &&;

	explicit RankedTree ( std::set < alphabet::RankedSymbol > alphabet, std::tree < alphabet::RankedSymbol > tree );
	explicit RankedTree ( std::tree < alphabet::RankedSymbol > tree );
	explicit RankedTree ( const UnrankedTree & other );

	/**
	 * @return Root node of the regular expression tree
	 */
	const std::tree < alphabet::RankedSymbol > & getContent ( ) const;

	const std::set < alphabet::RankedSymbol > & getAlphabet ( ) const {
		return accessComponent < GeneralAlphabet > ( ).get ( );
	}

	void extendAlphabet ( const std::set < alphabet::RankedSymbol > & symbols ) {
		accessComponent < GeneralAlphabet > ( ).add ( symbols );
	}

	/**
	 * Sets the root node of the regular expression tree
	 * @param tree root node to set
	 */
	void setTree ( std::tree < alphabet::RankedSymbol > tree );

	/**
	 * Prints XML representation of the tree to the output stream.
	 * @param out output stream to which print the tree
	 * @param tree tree to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const RankedTree & other ) const;

	void nicePrint ( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "RankedTree";

		return xmlTagName;
	}

	static RankedTree parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;
};

} /* namespace tree */

namespace std {

template < >
class ComponentConstraint< ::tree::RankedTree, alphabet::RankedSymbol, ::tree::GeneralAlphabet > {
public:
	static bool used ( const ::tree::RankedTree & tree, const alphabet::RankedSymbol & symbol ) {
		const std::tree < alphabet::RankedSymbol > & content = tree.getContent ( );
		return std::find(content.prefix_begin(), content.prefix_end(), symbol) != content.prefix_end();
	}

	static bool available ( const ::tree::RankedTree &, const alphabet::RankedSymbol & ) {
		return true;
	}

	static void valid ( const ::tree::RankedTree &, const alphabet::RankedSymbol & ) {
	}
};

} /* namespace std */

#endif /* RANKED_TREE_H_ */
