/*
 * PrefixBarTree.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "PrefixBarTree.h"
#include "../TreeException.h"

#include <sstream>
#include <algorithm>
#include <deque>

#include "UnrankedTree.h"

#include <sax/FromXMLParserHelper.h>
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include "../Tree.h"
#include "../UnrankedTreeWrapper.h"
#include <object/Object.h>
#include <XmlApi.hpp>
#include <cast/CastApi.hpp>

#include "../../alphabet/BarSymbol.h"

namespace tree {

PrefixBarTree::PrefixBarTree ( alphabet::Symbol bar, std::set < alphabet::Symbol > alphabet, std::vector < alphabet::Symbol > data ) : std::Components < PrefixBarTree, alphabet::Symbol, std::tuple < GeneralAlphabet >, std::tuple < BarSymbol > > ( std::make_tuple ( std::move ( alphabet ) ), std::make_tuple ( std::move ( bar ) ) ) {
	setContent ( std::move ( data ) );
}

PrefixBarTree::PrefixBarTree ( alphabet::Symbol bar, std::vector < alphabet::Symbol > data ) : PrefixBarTree ( bar, std::set < alphabet::Symbol > ( data.begin ( ), data.end ( ) ) + std::set < alphabet::Symbol > { bar }, data ) {
}

PrefixBarTree::PrefixBarTree ( alphabet::Symbol bar, const UnrankedTree & tree ) : PrefixBarTree ( bar, tree.getAlphabet ( ) + std::set < alphabet::Symbol > { bar }, toPrefixBar ( tree.getContent ( ), bar ) ) {
}

PrefixBarTree::PrefixBarTree ( const UnrankedTree & tree ) : PrefixBarTree ( alphabet::BarSymbol::BAR_SYMBOL, tree ) {
}

std::vector < alphabet::Symbol > PrefixBarTree::toPrefixBar ( const std::tree < alphabet::Symbol > & tree, const alphabet::Symbol & bar ) {
	std::vector < alphabet::Symbol > res;

	for ( std::tree < alphabet::Symbol >::const_structure_iterator iter = tree.structure_begin ( ); iter != tree.structure_end ( ); ++iter )
		if ( iter.getVirtual ( ) )
			res.push_back ( bar );
		else
			res.push_back ( * iter );

	return res;
}

UnrankedTreeBase * PrefixBarTree::clone ( ) const {
	return new PrefixBarTree ( * this );
}

UnrankedTreeBase * PrefixBarTree::plunder ( ) && {
	return new PrefixBarTree ( std::move ( * this ) );
}

const std::vector < alphabet::Symbol > & PrefixBarTree::getContent ( ) const {
	return this->m_Data;
}

void PrefixBarTree::setContent ( std::vector < alphabet::Symbol > data ) {
	arityChecksum ( data );

	std::set < alphabet::Symbol > minimalAlphabet ( data.begin ( ), data.end ( ) );
	std::set < alphabet::Symbol > unknownSymbols;
	std::set_difference ( minimalAlphabet.begin ( ), minimalAlphabet.end ( ), getAlphabet ( ).begin ( ), getAlphabet ( ).end ( ), std::inserter ( unknownSymbols, unknownSymbols.end ( ) ) );

	if ( unknownSymbols.size ( ) > 0 )
		throw TreeException ( "Input symbols not in the alphabet." );

	this->m_Data = std::move ( data );
}

void PrefixBarTree::arityChecksum ( const std::vector < alphabet::Symbol > & data ) {
	int arityChecksumTypes = 0;

	for ( const alphabet::Symbol & symbol : data ) {
		if ( symbol == getBar ( ) )
			arityChecksumTypes -= 1;
		else
			arityChecksumTypes += 1;
	}

	if ( arityChecksumTypes != 0 ) throw TreeException ( "The string does not form a tree" );
}

bool PrefixBarTree::isEmpty ( ) const {
	return this->m_Data.size ( ) == 0;
}

int PrefixBarTree::compare ( const PrefixBarTree & other ) const {
	auto first = std::tie ( m_Data, getAlphabet ( ), getBar ( ) );
	auto second = std::tie ( other.m_Data, other.getAlphabet ( ), other.getBar ( ) );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void PrefixBarTree::operator >>( std::ostream & out ) const {
	out << "(PrefixBarTree ";

	for ( const alphabet::Symbol & symbol : this->m_Data )
		out << symbol;

	out << ")";
}

PrefixBarTree::operator std::string ( ) const {
	std::stringstream ss;
	ss << "\"";

	for ( const alphabet::Symbol & symbol : this->m_Data )
		ss << symbol;

	ss << "\"";
	return std::move ( ss ).str ( );
}

PrefixBarTree PrefixBarTree::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, PrefixBarTree::getXmlTagName() );
	alphabet::Symbol bar = TreeFromXMLParser::parseBar ( input );
	std::set < alphabet::Symbol > unrankedAlphabet = TreeFromXMLParser::parseUnrankedAlphabet ( input );
	std::vector < alphabet::Symbol > data = TreeFromXMLParser::parseUnrankedLinearContent ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, PrefixBarTree::getXmlTagName() );

	return PrefixBarTree ( std::move ( bar ), std::move ( unrankedAlphabet ), std::move ( data ) );
}

void PrefixBarTree::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( PrefixBarTree::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	TreeToXMLComposer::composeBar ( out, getBar ( ) );
	TreeToXMLComposer::composeAlphabet ( out, getAlphabet ( ) );
	TreeToXMLComposer::composeContent ( out, m_Data );
	out.emplace_back ( PrefixBarTree::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

} /* namespace tree */

namespace alib {

auto prefixBarTreeParserRegister = xmlApi < ::tree::Tree >::ParserRegister < ::tree::PrefixBarTree > ( );
auto prefixBarTreeParserRegister2 = xmlApi < ::tree::UnrankedTreeWrapper >::ParserRegister < ::tree::PrefixBarTree > ( );
auto prefixBarTreeParserRegister3 = xmlApi < alib::Object >::ParserRegister < ::tree::PrefixBarTree > ( );

auto PrefixBarTreeFromUnrankedTree = castApi::CastRegister < ::tree::PrefixBarTree, ::tree::UnrankedTree > ( );
auto PrefixBarTreeCastBinder = castApi::CastPoolStringBinder < ::tree::PrefixBarTree > ( ::tree::PrefixBarTree::getXmlTagName() );

} /* namespace alib */
