/*
 * UnrankedNonlinearPattern.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "UnrankedNonlinearPattern.h"

#include <iostream>
#include <algorithm>
#include <sstream>

#include "../ranked/RankedNonlinearPattern.h"

#include <sax/FromXMLParserHelper.h>
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include "../common/TreeAuxiliary.h"
#include "../Tree.h"
#include "../UnrankedTreeWrapper.h"
#include <object/Object.h>
#include <XmlApi.hpp>
#include <cast/CastApi.hpp>

namespace tree {

UnrankedNonlinearPattern::UnrankedNonlinearPattern ( alphabet::Symbol subtreeWildcard, std::set < alphabet::Symbol > nonlinearVariables, std::set < alphabet::Symbol > alphabet, std::tree < alphabet::Symbol > pattern ) : std::Components < UnrankedNonlinearPattern, alphabet::Symbol, std::tuple < GeneralAlphabet, NonlinearAlphabet >, std::tuple < SubtreeWildcard > > ( std::make_tuple ( std::move ( alphabet ), std::move ( nonlinearVariables ) ), std::make_tuple ( std::move ( subtreeWildcard ) ) ), content ( std::move ( pattern ) ) {
	checkAlphabet ( content );
}

UnrankedNonlinearPattern::UnrankedNonlinearPattern ( alphabet::Symbol subtreeWildcard, std::set < alphabet::Symbol > nonlinearVariables, std::tree < alphabet::Symbol > pattern ) : UnrankedNonlinearPattern ( subtreeWildcard, nonlinearVariables, std::set < alphabet::Symbol > ( pattern.prefix_begin ( ), pattern.prefix_end ( ) ) + nonlinearVariables + std::set < alphabet::Symbol > { subtreeWildcard }, pattern ) {
}

UnrankedNonlinearPattern::UnrankedNonlinearPattern ( alphabet::Symbol subtreeWildcard, std::tree < alphabet::Symbol > pattern ) : UnrankedNonlinearPattern ( subtreeWildcard, std::set < alphabet::Symbol > { }, pattern ) {
}

UnrankedNonlinearPattern::UnrankedNonlinearPattern ( const RankedNonlinearPattern & other ) : UnrankedNonlinearPattern ( other.getSubtreeWildcard ( ).getSymbol ( ), TreeAuxiliary::unrankSymbols ( other.getAlphabet ( ) ), TreeAuxiliary::toUnrankedTree ( other.getContent ( ) ) ) {
}

UnrankedTreeBase * UnrankedNonlinearPattern::clone ( ) const {
	return new UnrankedNonlinearPattern ( * this );
}

UnrankedTreeBase * UnrankedNonlinearPattern::plunder ( ) && {
	return new UnrankedNonlinearPattern ( std::move ( * this ) );
}

const std::tree < alphabet::Symbol > & UnrankedNonlinearPattern::getContent ( ) const {
	return content;
}

void UnrankedNonlinearPattern::checkAlphabet ( const std::tree < alphabet::Symbol > & data ) const {
	std::set < alphabet::Symbol > minimalAlphabet ( data.prefix_begin ( ), data.prefix_end ( ) );
	std::set < alphabet::Symbol > unknownSymbols;
	std::set_difference ( minimalAlphabet.begin ( ), minimalAlphabet.end ( ), getAlphabet().begin ( ), getAlphabet().end ( ), std::inserter ( unknownSymbols, unknownSymbols.end ( ) ) );

	if ( unknownSymbols.size ( ) > 0 )
		throw exception::CommonException ( "Input symbols not in the alphabet." );
}

void UnrankedNonlinearPattern::setTree ( std::tree < alphabet::Symbol > pattern ) {
	checkAlphabet ( pattern );

	this->content = std::move ( pattern );
}

void UnrankedNonlinearPattern::operator >>( std::ostream & out ) const {
	out << "(UnrankedNonlinearPattern " << this->content << ")";
}

int UnrankedNonlinearPattern::compare ( const UnrankedNonlinearPattern & other ) const {
	auto first = std::tie ( content, getAlphabet(), getSubtreeWildcard(), getNonlinearVariables() );
	auto second = std::tie ( other.content, other.getAlphabet(), other.getSubtreeWildcard(), getNonlinearVariables() );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void UnrankedNonlinearPattern::nicePrint ( std::ostream & os ) const {
	content.nicePrint ( os );
}

UnrankedNonlinearPattern::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

UnrankedNonlinearPattern UnrankedNonlinearPattern::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, UnrankedNonlinearPattern::getXmlTagName() );
	alphabet::Symbol subtreeWildcardSymbol = TreeFromXMLParser::parseSubtreeWildcardSymbol ( input );
	std::set < alphabet::Symbol > nonlinearVariables = TreeFromXMLParser::parseUnrankedNonlinearVariables ( input );
	std::set < alphabet::Symbol > rankedAlphabet = TreeFromXMLParser::parseUnrankedAlphabet ( input );
	std::tree < alphabet::Symbol > root = TreeFromXMLParser::parseUnrankedTreeContent ( input );
	UnrankedNonlinearPattern tree ( std::move ( subtreeWildcardSymbol ), std::move ( nonlinearVariables ), std::move ( rankedAlphabet ), std::move ( root ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, UnrankedNonlinearPattern::getXmlTagName() );
	return tree;
}

void UnrankedNonlinearPattern::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( UnrankedNonlinearPattern::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	TreeToXMLComposer::composeSubtreeWildcard ( out, getSubtreeWildcard ( ) );
	TreeToXMLComposer::composeNonlinearVariables ( out, getNonlinearVariables ( ) );
	TreeToXMLComposer::composeAlphabet ( out, getAlphabet ( ) );
	TreeToXMLComposer::composeContent ( out, getContent ( ) );
	out.emplace_back ( UnrankedNonlinearPattern::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

} /* namespace tree */

namespace alib {

auto UnrankedNonlinearPatternParserRegister = xmlApi < ::tree::Tree >::ParserRegister < ::tree::UnrankedNonlinearPattern > ( );
auto UnrankedNonlinearPatternParserRegister2 = xmlApi < ::tree::UnrankedTreeWrapper >::ParserRegister < ::tree::UnrankedNonlinearPattern > ( );
auto UnrankedNonlinearPatternParserRegister3 = xmlApi < alib::Object >::ParserRegister < ::tree::UnrankedNonlinearPattern > ( );

} /* namespace alib */
