/*
 * UnrankedNonlinearPattern.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef UNRANKED_NONLINEAR_PATTERN_H_
#define UNRANKED_NONLINEAR_PATTERN_H_

#include <string>
#include <set>
#include <tree>
#include <core/components.hpp>

#include "../TreeException.h"
#include "../UnrankedTreeBase.h"

namespace tree {

class RankedNonlinearPattern;
class GeneralAlphabet;
class SubtreeWildcard;
class NonlinearAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a pattern of RegExpElement.
 */
class UnrankedNonlinearPattern : public UnrankedTreeBase, public std::Components < UnrankedNonlinearPattern, alphabet::Symbol, std::tuple < GeneralAlphabet, NonlinearAlphabet >, std::tuple < SubtreeWildcard > > {
	std::tree < alphabet::Symbol > content;

	void checkAlphabet ( const std::tree < alphabet::Symbol > & pattern ) const;

public:
	/**
	 * @copydoc std::tree < alphabet::Symbol >::clone() const
	 */
	virtual UnrankedTreeBase * clone ( ) const;

	/**
	 * @copydoc std::tree < alphabet::Symbol >::plunder() const
	 */
	virtual UnrankedTreeBase * plunder ( ) &&;

	explicit UnrankedNonlinearPattern ( alphabet::Symbol subtreeWildcard, std::set < alphabet::Symbol > nonlinearVariables, std::set < alphabet::Symbol > alphabet, std::tree < alphabet::Symbol > pattern );
	explicit UnrankedNonlinearPattern ( alphabet::Symbol subtreeWildcard, std::set < alphabet::Symbol > nonlinearVariables, std::tree < alphabet::Symbol > pattern );
	explicit UnrankedNonlinearPattern ( alphabet::Symbol subtreeWildcard, std::tree < alphabet::Symbol > pattern );
	explicit UnrankedNonlinearPattern ( const RankedNonlinearPattern & other );

	/**
	 * @return Content of the pattern
	 */
	const std::tree < alphabet::Symbol > & getContent ( ) const;

	const std::set < alphabet::Symbol > & getAlphabet ( ) const {
		return accessComponent < GeneralAlphabet > ( ).get ( );
	}

	void extendAlphabet ( const std::set < alphabet::Symbol > & symbols ) {
		accessComponent < GeneralAlphabet > ( ).add ( symbols );
	}

	const alphabet::Symbol & getSubtreeWildcard ( ) const {
		return accessElement < SubtreeWildcard > ( ).get ( );
	}

	const std::set < alphabet::Symbol > & getNonlinearVariables ( ) const {
		return accessComponent < NonlinearAlphabet > ( ).get ( );
	}

	/**
	 * Sets the root node of the regular expression pattern
	 * @param pattern root node to set
	 */
	void setTree ( std::tree < alphabet::Symbol > pattern );

	/**
	 * Removes symbol from the alphabet of symbol available in the regular expression
	 * @param symbol removed symbol from the alphabet
	 */
	bool removeSymbolFromAlphabet ( const alphabet::Symbol & symbol );

	/**
	 * Prints XML representation of the pattern to the output stream.
	 * @param out output stream to which print the pattern
	 * @param pattern pattern to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const UnrankedNonlinearPattern & other ) const;

	void nicePrint ( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "UnrankedNonlinearPattern";

		return xmlTagName;
	}

	static UnrankedNonlinearPattern parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;
};

} /* namespace tree */

namespace std {

template < >
class ComponentConstraint< ::tree::UnrankedNonlinearPattern, alphabet::Symbol, ::tree::GeneralAlphabet > {
public:
	static bool used ( const ::tree::UnrankedNonlinearPattern & pattern, const alphabet::Symbol & symbol ) {
		const std::tree<alphabet::Symbol>& content = pattern.getContent ( );
		return std::find(content.prefix_begin(), content.prefix_end(), symbol) != content.prefix_end() || pattern.accessElement < ::tree::SubtreeWildcard > ( ).get ( ) == symbol || pattern.accessComponent < ::tree::NonlinearAlphabet > ( ).get ( ).count ( symbol );
	}

	static bool available ( const ::tree::UnrankedNonlinearPattern &, const alphabet::Symbol & ) {
		return true;
	}

	static void valid ( const ::tree::UnrankedNonlinearPattern &, const alphabet::Symbol & ) {
	}
};

template < >
class ComponentConstraint< ::tree::UnrankedNonlinearPattern, alphabet::Symbol, ::tree::NonlinearAlphabet > {
public:
	static bool used ( const ::tree::UnrankedNonlinearPattern &, const alphabet::Symbol & ) {
		return false;
	}

	static bool available ( const ::tree::UnrankedNonlinearPattern & pattern, const alphabet::Symbol & symbol ) {
		return pattern.accessComponent < ::tree::GeneralAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const ::tree::UnrankedNonlinearPattern & pattern, const alphabet::Symbol & symbol) {
		if ( pattern.accessElement < ::tree::SubtreeWildcard > ( ).get ( ) == symbol )
			throw ::tree::TreeException ( "Symbol " + std::to_string ( symbol ) + "cannot be set as nonlinear variable since it is already subtree wildcard" );
	}
};

template < >
class ElementConstraint< ::tree::UnrankedNonlinearPattern, alphabet::Symbol, ::tree::SubtreeWildcard > {
public:
	static bool available ( const ::tree::UnrankedNonlinearPattern & pattern, const alphabet::Symbol & symbol ) {
		return pattern.accessComponent < ::tree::GeneralAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const ::tree::UnrankedNonlinearPattern & pattern, const alphabet::Symbol & symbol) {
		if ( pattern.accessComponent < ::tree::NonlinearAlphabet > ( ).get ( ).count ( symbol ) )
			throw ::tree::TreeException ( "Symbol " + std::to_string ( symbol ) + "cannot be set as subtree wildcard since it is already nonlinear variable" );
	}
};

} /* namespace std */

#endif /* UNRANKED_NONLINEAR_PATTERN_H_ */
