/*
 * UnrankedPattern.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "UnrankedPattern.h"

#include <iostream>
#include <algorithm>
#include <sstream>

#include "../ranked/RankedPattern.h"

#include <sax/FromXMLParserHelper.h>
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include "../common/TreeAuxiliary.h"
#include "../Tree.h"
#include "../UnrankedTreeWrapper.h"
#include <object/Object.h>
#include <XmlApi.hpp>
#include <cast/CastApi.hpp>

namespace tree {

UnrankedPattern::UnrankedPattern ( alphabet::Symbol subtreeWildcard, std::set < alphabet::Symbol > alphabet, std::tree < alphabet::Symbol > pattern ) : std::Components < UnrankedPattern, alphabet::Symbol, std::tuple < GeneralAlphabet >, std::tuple < SubtreeWildcard > > ( std::make_tuple ( std::move ( alphabet ) ), std::make_tuple ( std::move ( subtreeWildcard ) ) ), content ( std::move ( pattern ) ) {
	checkAlphabet ( content );
}

UnrankedPattern::UnrankedPattern ( alphabet::Symbol subtreeWildcard, std::tree < alphabet::Symbol > pattern ) : UnrankedPattern ( subtreeWildcard, std::set < alphabet::Symbol > ( pattern.prefix_begin ( ), pattern.prefix_end ( ) ) + std::set < alphabet::Symbol > { subtreeWildcard }, pattern ) {
}

UnrankedPattern::UnrankedPattern ( const RankedPattern & other ) : UnrankedPattern ( other.getSubtreeWildcard ( ).getSymbol ( ), TreeAuxiliary::unrankSymbols ( other.getAlphabet ( ) ), TreeAuxiliary::toUnrankedTree ( other.getContent ( ) ) ) {
}

UnrankedTreeBase * UnrankedPattern::clone ( ) const {
	return new UnrankedPattern ( * this );
}

UnrankedTreeBase * UnrankedPattern::plunder ( ) && {
	return new UnrankedPattern ( std::move ( * this ) );
}

const std::tree < alphabet::Symbol > & UnrankedPattern::getContent ( ) const {
	return content;
}

void UnrankedPattern::checkAlphabet ( const std::tree < alphabet::Symbol > & data ) const {
	std::set < alphabet::Symbol > minimalAlphabet ( data.prefix_begin ( ), data.prefix_end ( ) );
	std::set < alphabet::Symbol > unknownSymbols;
	std::set_difference ( minimalAlphabet.begin ( ), minimalAlphabet.end ( ), getAlphabet().begin ( ), getAlphabet().end ( ), std::inserter ( unknownSymbols, unknownSymbols.end ( ) ) );

	if ( unknownSymbols.size ( ) > 0 )
		throw exception::CommonException ( "Input symbols not in the alphabet." );
}

void UnrankedPattern::setTree ( std::tree < alphabet::Symbol > pattern ) {
	checkAlphabet ( pattern );

	this->content = std::move ( pattern );
}

void UnrankedPattern::operator >>( std::ostream & out ) const {
	out << "(UnrankedPattern " << this->content << ")";
}

int UnrankedPattern::compare ( const UnrankedPattern & other ) const {
	auto first = std::tie ( content, getAlphabet(), getSubtreeWildcard() );
	auto second = std::tie ( other.content, other.getAlphabet(), other.getSubtreeWildcard() );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void UnrankedPattern::nicePrint ( std::ostream & os ) const {
	content.nicePrint ( os );
}

UnrankedPattern::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

UnrankedPattern UnrankedPattern::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, UnrankedPattern::getXmlTagName() );
	alphabet::Symbol subtreeWildcardSymbol = TreeFromXMLParser::parseSubtreeWildcardSymbol ( input );
	std::set < alphabet::Symbol > rankedAlphabet = TreeFromXMLParser::parseUnrankedAlphabet ( input );
	std::tree < alphabet::Symbol > root = TreeFromXMLParser::parseUnrankedTreeContent ( input );
	UnrankedPattern tree ( std::move ( subtreeWildcardSymbol ), std::move ( rankedAlphabet ), std::move ( root ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, UnrankedPattern::getXmlTagName() );
	return tree;
}

void UnrankedPattern::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( UnrankedPattern::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	TreeToXMLComposer::composeSubtreeWildcard ( out, getSubtreeWildcard ( ) );
	TreeToXMLComposer::composeAlphabet ( out, getAlphabet ( ) );
	TreeToXMLComposer::composeContent ( out, getContent ( ) );
	out.emplace_back ( UnrankedPattern::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

} /* namespace tree */

namespace alib {

auto unrankedPatternParserRegister = xmlApi < ::tree::Tree >::ParserRegister < ::tree::UnrankedPattern > ( );
auto unrankedPatternParserRegister2 = xmlApi < ::tree::UnrankedTreeWrapper >::ParserRegister < ::tree::UnrankedPattern > ( );
auto unrankedPatternParserRegister3 = xmlApi < alib::Object >::ParserRegister < ::tree::UnrankedPattern > ( );

auto UnrankedPatternFromRankedPattern = castApi::CastRegister < ::tree::UnrankedPattern, ::tree::RankedPattern > ( );
auto UnrankedPatternCastBinder = castApi::CastPoolStringBinder < ::tree::UnrankedPattern > ( ::tree::UnrankedPattern::getXmlTagName() );

} /* namespace alib */
