/*
 * UnrankedPattern.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef UNRANKED_PATTERN_H_
#define UNRANKED_PATTERN_H_

#include <string>
#include <set>
#include <tree>
#include <core/components.hpp>

#include "../TreeException.h"
#include "../UnrankedTreeBase.h"

#include "../../alphabet/RankedSymbol.h"

namespace tree {

class RankedPattern;
class GeneralAlphabet;
class SubtreeWildcard;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a pattern of RegExpElement.
 */
class UnrankedPattern : public UnrankedTreeBase, public std::Components < UnrankedPattern, alphabet::Symbol, std::tuple < GeneralAlphabet >, std::tuple < SubtreeWildcard > > {
	std::tree < alphabet::Symbol > content;

	void checkAlphabet ( const std::tree < alphabet::Symbol > & pattern ) const;

public:
	/**
	 * @copydoc UnrankedNode::clone() const
	 */
	virtual UnrankedTreeBase * clone ( ) const;

	/**
	 * @copydoc UnrankedNode::plunder() const
	 */
	virtual UnrankedTreeBase * plunder ( ) &&;

	explicit UnrankedPattern ( alphabet::Symbol subtreeWildcard, std::set < alphabet::Symbol > alphabet, std::tree < alphabet::Symbol > pattern );
	explicit UnrankedPattern ( alphabet::Symbol subtreeWildcard, std::tree < alphabet::Symbol > pattern );
	explicit UnrankedPattern ( const RankedPattern & other );

	/**
	 * @return Content of the pattern
	 */
	const std::tree < alphabet::Symbol > & getContent ( ) const;

	virtual const std::set < alphabet::Symbol > & getAlphabet ( ) const {
		return accessComponent < GeneralAlphabet > ( ).get ( );
	}

	void extendAlphabet ( const std::set < alphabet::Symbol > & symbols ) {
		accessComponent < GeneralAlphabet > ( ).add ( symbols );
	}

	const alphabet::Symbol & getSubtreeWildcard ( ) const {
		return accessElement < SubtreeWildcard > ( ).get ( );
	}

	/**
	 * Sets the root node of the regular expression pattern
	 * @param pattern root node to set
	 */
	void setTree ( std::tree < alphabet::Symbol > pattern );

	/**
	 * Prints XML representation of the pattern to the output stream.
	 * @param out output stream to which print the pattern
	 * @param pattern pattern to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const UnrankedPattern & other ) const;

	void nicePrint ( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "UnrankedPattern";

		return xmlTagName;
	}

	static UnrankedPattern parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;
};

} /* namespace tree */

namespace std {

template < >
class ComponentConstraint< ::tree::UnrankedPattern, alphabet::Symbol, ::tree::GeneralAlphabet > {
public:
	static bool used ( const ::tree::UnrankedPattern & pattern, const alphabet::Symbol & symbol ) {
		const std::tree<alphabet::Symbol>& content = pattern.getContent ( );
		return std::find(content.prefix_begin(), content.prefix_end(), symbol) != content.prefix_end() || pattern.accessElement < ::tree::SubtreeWildcard > ( ).get ( ) == symbol;
	}

	static bool available ( const ::tree::UnrankedPattern &, const alphabet::Symbol & ) {
		return true;
	}

	static void valid ( const ::tree::UnrankedPattern &, const alphabet::Symbol & ) {
	}
};

template < >
class ElementConstraint< ::tree::UnrankedPattern, alphabet::Symbol, ::tree::SubtreeWildcard > {
public:
	static bool available ( const ::tree::UnrankedPattern & pattern, const alphabet::Symbol & symbol ) {
		return pattern.accessComponent < ::tree::GeneralAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const ::tree::UnrankedPattern &, const alphabet::Symbol & ) {
	}
};

} /* namespace std */

#endif /* UNRANKED_PATTERN_H_ */
