/*
 * UnrankedTree.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Stepan Plachy
 */

#include "UnrankedTree.h"

#include <iostream>
#include <algorithm>
#include <sstream>

#include "../ranked/RankedTree.h"

#include <sax/FromXMLParserHelper.h>
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include "../common/TreeAuxiliary.h"
#include "../Tree.h"
#include "../UnrankedTreeWrapper.h"
#include <object/Object.h>
#include <XmlApi.hpp>
#include <cast/CastApi.hpp>

namespace tree {

UnrankedTree::UnrankedTree ( std::set < alphabet::Symbol > alphabet, std::tree < alphabet::Symbol > tree ) : std::Components < UnrankedTree, alphabet::Symbol, std::tuple < GeneralAlphabet >, std::tuple < > > ( std::make_tuple ( std::move ( alphabet ) ), std::tuple < > ( ) ), content ( std::move ( tree ) ) {
	checkAlphabet ( content );
}

UnrankedTree::UnrankedTree ( std::tree < alphabet::Symbol > pattern ) : UnrankedTree ( std::set < alphabet::Symbol > ( pattern.prefix_begin ( ), pattern.prefix_end ( ) ), pattern ) {
}

UnrankedTree::UnrankedTree ( const RankedTree & other ) : UnrankedTree ( TreeAuxiliary::unrankSymbols ( other.getAlphabet ( ) ), TreeAuxiliary::toUnrankedTree ( other.getContent ( ) ) ) {
}

UnrankedTreeBase * UnrankedTree::clone ( ) const {
	return new UnrankedTree ( * this );
}

UnrankedTreeBase * UnrankedTree::plunder ( ) && {
	return new UnrankedTree ( std::move ( * this ) );
}

const std::tree < alphabet::Symbol > & UnrankedTree::getContent ( ) const {
	return content;
}

void UnrankedTree::checkAlphabet ( const std::tree < alphabet::Symbol > & data ) const {
	std::set < alphabet::Symbol > minimalAlphabet ( data.prefix_begin ( ), data.prefix_end ( ) );
	std::set < alphabet::Symbol > unknownSymbols;
	std::set_difference ( minimalAlphabet.begin ( ), minimalAlphabet.end ( ), getAlphabet().begin ( ), getAlphabet().end ( ), std::inserter ( unknownSymbols, unknownSymbols.end ( ) ) );

	if ( unknownSymbols.size ( ) > 0 )
		throw exception::CommonException ( "Input symbols not in the alphabet." );
}

void UnrankedTree::setTree ( std::tree < alphabet::Symbol > data ) {
	checkAlphabet ( data );

	this->content = std::move ( data );
}

void UnrankedTree::operator >>( std::ostream & out ) const {
	out << "(UnrankedTree " << this->content << ")";
}

int UnrankedTree::compare ( const UnrankedTree & other ) const {
	auto first = std::tie ( content, getAlphabet() );
	auto second = std::tie ( other.content, other.getAlphabet() );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void UnrankedTree::nicePrint ( std::ostream & os ) const {
	content.nicePrint ( os );
}

UnrankedTree::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

UnrankedTree UnrankedTree::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, UnrankedTree::getXmlTagName() );
	std::set < alphabet::Symbol > rankedAlphabet = TreeFromXMLParser::parseUnrankedAlphabet ( input );
	std::tree < alphabet::Symbol > root = TreeFromXMLParser::parseUnrankedTreeContent ( input );
	UnrankedTree tree ( std::move ( rankedAlphabet ), std::move ( root ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, UnrankedTree::getXmlTagName() );
	return tree;
}

void UnrankedTree::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( UnrankedTree::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	TreeToXMLComposer::composeAlphabet ( out, getAlphabet ( ) );
	TreeToXMLComposer::composeContent ( out, getContent ( ) );
	out.emplace_back ( UnrankedTree::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

} /* namespace tree */

namespace alib {

auto unrankedTreeParserRegister = xmlApi < ::tree::Tree >::ParserRegister < ::tree::UnrankedTree > ( );
auto unrankedTreeParserRegister2 = xmlApi < ::tree::UnrankedTreeWrapper >::ParserRegister < ::tree::UnrankedTree > ( );
auto unrankedTreeParserRegister3 = xmlApi < alib::Object >::ParserRegister < ::tree::UnrankedTree > ( );

auto UnrankedTreeFromRankedTree = castApi::CastRegister < ::tree::UnrankedTree, ::tree::RankedTree > ( );
auto UnrankedTreeCastBinder = castApi::CastPoolStringBinder < ::tree::UnrankedTree > ( ::tree::UnrankedTree::getXmlTagName() );

} /* namespace alib */
