/*
 * UnrankedTree.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Stepan Plachy
 */

#ifndef UNRANKED_TREE_H_
#define UNRANKED_TREE_H_

#include <string>
#include <set>
#include <tree>
#include <core/components.hpp>

#include "../TreeException.h"
#include "../UnrankedTreeBase.h"

#include "../../alphabet/RankedSymbol.h"

namespace tree {

class RankedTree;
class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
class UnrankedTree : public UnrankedTreeBase, public std::Components < UnrankedTree, alphabet::Symbol, std::tuple < GeneralAlphabet >, std::tuple < > > {
	std::tree < alphabet::Symbol > content;

	static std::set < alphabet::Symbol > unrankSymbols ( const std::set < alphabet::RankedSymbol > & alphabet );

	void checkAlphabet ( const std::tree < alphabet::Symbol > & pattern ) const;

public:
	/**
	 * @copydoc UnrankedNode::clone() const
	 */
	virtual UnrankedTreeBase * clone ( ) const;

	/**
	 * @copydoc UnrankedNode::plunder() const
	 */
	virtual UnrankedTreeBase * plunder ( ) &&;

	explicit UnrankedTree ( std::set < alphabet::Symbol > alphabet, std::tree < alphabet::Symbol > tree );
	explicit UnrankedTree ( std::tree < alphabet::Symbol > tree );
	explicit UnrankedTree ( const RankedTree & other );

	const std::set < alphabet::Symbol > & getAlphabet ( ) const {
		return accessComponent < GeneralAlphabet > ( ).get ( );
	}

	void extendAlphabet ( const std::set < alphabet::Symbol > & symbols ) {
		accessComponent < GeneralAlphabet > ( ).add ( symbols );
	}

	/**
	 * @return content of the tree
	 */
	const std::tree < alphabet::Symbol > & getContent ( ) const;

	/**
	 * Sets the root node of the regular expression tree
	 * @param tree root node to set
	 */
	void setTree ( std::tree < alphabet::Symbol > data );

	/**
	 * Prints XML representation of the tree to the output stream.
	 * @param out output stream to which print the tree
	 * @param tree tree to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const UnrankedTree & other ) const;

	void nicePrint ( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "UnrankedTree";

		return xmlTagName;
	}

	static UnrankedTree parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;
};

} /* namespace tree */

namespace std {

template < >
class ComponentConstraint< ::tree::UnrankedTree, alphabet::Symbol, ::tree::GeneralAlphabet > {
public:
	static bool used ( const ::tree::UnrankedTree & tree, const alphabet::Symbol & symbol ) {
		const std::tree<alphabet::Symbol>& content = tree.getContent ( );
		return std::find(content.prefix_begin(), content.prefix_end(), symbol) != content.prefix_end();
	}

	static bool available ( const ::tree::UnrankedTree &, const alphabet::Symbol & ) {
		return true;
	}

	static void valid ( const ::tree::UnrankedTree &, const alphabet::Symbol & ) {
	}
};

} /* namespace std */

#endif /* UNRANKED_TREE_H_ */
