#include "AutomatonTest.h"

#include "sax/SaxParseInterface.h"
#include "sax/SaxComposeInterface.h"

#include "automaton/FSM/DFA.h"
#include "automaton/FSM/ExtendedNFA.h"
#include "automaton/PDA/SinglePopDPDA.h"
#include "automaton/PDA/DPDA.h"
#include "automaton/PDA/NPDA.h"
#include "automaton/PDA/RealTimeHeightDeterministicNPDA.h"
#include "automaton/TA/NFTA.h"

#include "automaton/AutomatonException.h"

#include "factory/XmlDataFactory.hpp"

#include "alphabet/LabeledSymbol.h"
#include "alphabet/BottomOfTheStackSymbol.h"

#include "regexp/RegExp.h"
#include "regexp/unbounded/UnboundedRegExp.h"
#include "regexp/unbounded/UnboundedRegExpElements.h"

#define CPPUNIT_IMPLY(x, y) CPPUNIT_ASSERT(!(x) || (y))

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION( AutomatonTest, "automaton" );
CPPUNIT_TEST_SUITE_REGISTRATION( AutomatonTest );

void AutomatonTest::setUp() {
}

void AutomatonTest::tearDown() {
}

void AutomatonTest::testDFAParser() {
	automaton::DFA<> automaton(label::labelFrom(1));

	automaton.addState(label::labelFrom(1));
	automaton.addState(label::labelFrom(2));
	automaton.addState(label::labelFrom(3));
	automaton.addInputSymbol(alphabet::symbolFrom("a"));
	automaton.addInputSymbol(alphabet::symbolFrom("b"));

	automaton.addTransition(label::labelFrom(1), alphabet::symbolFrom("a"), label::labelFrom(2));
	automaton.addTransition(label::labelFrom(2), alphabet::symbolFrom("b"), label::labelFrom(1));

	automaton.addFinalState(label::labelFrom(3));

	CPPUNIT_ASSERT( automaton == automaton );
	{
		std::deque<sax::Token> tokens = alib::XmlDataFactory::toTokens(automaton);
		std::string tmp;
		sax::SaxComposeInterface::printMemory(tmp, tokens);

		std::deque<sax::Token> tokens2;
		sax::SaxParseInterface::parseMemory(tmp, tokens2);
		automaton::DFA<> automaton2 = alib::XmlDataFactory::fromTokens<automaton::DFA<>>(std::move(tokens2));

		CPPUNIT_ASSERT( automaton == automaton2 );
	}
}

void AutomatonTest::SinglePopDPDATransitions() {
	automaton::SinglePopDPDA < > automaton(label::labelFrom(1), alphabet::symbolFrom("S"));

	automaton.addState(label::labelFrom(1));
	automaton.addState(label::labelFrom(2));
	automaton.addState(label::labelFrom(3));

	automaton.addInputSymbol(alphabet::symbolFrom("a"));
	automaton.addInputSymbol(alphabet::symbolFrom("b"));

	automaton.addPushdownStoreSymbol(alphabet::symbolFrom("S"));
	automaton.addPushdownStoreSymbol(alphabet::symbolFrom("X"));
	automaton.addPushdownStoreSymbol(alphabet::symbolFrom("Y"));

	automaton.addTransition(label::labelFrom(3), /* eps, */ alphabet::symbolFrom("X"), label::labelFrom(1), { alphabet::symbolFrom("X") } );

	CPPUNIT_ASSERT(!automaton.addTransition(label::labelFrom(3), /* eps, */ alphabet::symbolFrom("X"), label::labelFrom(1), { alphabet::symbolFrom("X") } ) );

	automaton.addTransition(label::labelFrom(3), alphabet::symbolFrom("a"), alphabet::symbolFrom("Y"), label::labelFrom(1), { alphabet::symbolFrom("X") } );

	CPPUNIT_ASSERT_THROW(
		automaton.addTransition(label::labelFrom(3), alphabet::symbolFrom("a"), alphabet::symbolFrom("X"), label::labelFrom(1), { alphabet::symbolFrom("X") } ),
		automaton::AutomatonException
	);

	automaton.addTransition(label::labelFrom(1), alphabet::symbolFrom("a"), alphabet::symbolFrom("X"), label::labelFrom(2), { alphabet::symbolFrom("Y") } );
	automaton.addTransition(label::labelFrom(1), alphabet::symbolFrom("a"), alphabet::symbolFrom("X"), label::labelFrom(2), { alphabet::symbolFrom("Y") } );
	automaton.addTransition(label::labelFrom(1), /* eps, */ alphabet::symbolFrom("Y"), label::labelFrom(2), { alphabet::symbolFrom("Y") } );

	automaton.addTransition(label::labelFrom(2), alphabet::symbolFrom("b"), alphabet::symbolFrom("X"), label::labelFrom(1), { alphabet::symbolFrom("X") } );
	CPPUNIT_ASSERT_THROW(
		automaton.addTransition(label::labelFrom(2), /* eps, */ alphabet::symbolFrom("X"), label::labelFrom(1), { alphabet::symbolFrom("X") } ),
		automaton::AutomatonException
	);

	automaton.addFinalState(label::labelFrom(3));
}

void AutomatonTest::DPDATransitions() {
	automaton::DPDA < > automaton(label::labelFrom(1), alphabet::symbolFrom("S"));

	automaton.addState(label::labelFrom(1));
	automaton.addState(label::labelFrom(2));
	automaton.addState(label::labelFrom(3));
	automaton.addState(label::labelFrom(4));

	automaton.addInputSymbol(alphabet::symbolFrom("a"));
	automaton.addInputSymbol(alphabet::symbolFrom("b"));

	automaton.addPushdownStoreSymbol(alphabet::symbolFrom("S"));
	automaton.addPushdownStoreSymbol(alphabet::symbolFrom("X"));
	automaton.addPushdownStoreSymbol(alphabet::symbolFrom("Y"));

	automaton.addTransition(label::labelFrom(3), /* eps, */ { alphabet::symbolFrom("X") }, label::labelFrom(1), { alphabet::symbolFrom("X") } );

	CPPUNIT_ASSERT(!automaton.addTransition(label::labelFrom(3), /* eps, */ { alphabet::symbolFrom("X") }, label::labelFrom(1), { alphabet::symbolFrom("X") } ) );

	automaton.addTransition(label::labelFrom(3), alphabet::symbolFrom("a"), { alphabet::symbolFrom("Y") }, label::labelFrom(1), { alphabet::symbolFrom("X") } );

	CPPUNIT_ASSERT_THROW(
		automaton.addTransition(label::labelFrom(3), alphabet::symbolFrom("a"), { alphabet::symbolFrom("X") }, label::labelFrom(1), { alphabet::symbolFrom("X") } ),
		automaton::AutomatonException
	);

	automaton.addTransition(label::labelFrom(1), alphabet::symbolFrom("a"), { alphabet::symbolFrom("X") }, label::labelFrom(2), { alphabet::symbolFrom("Y") } );
	automaton.addTransition(label::labelFrom(1), alphabet::symbolFrom("a"), { alphabet::symbolFrom("X") }, label::labelFrom(2), { alphabet::symbolFrom("Y") } );
	automaton.addTransition(label::labelFrom(1), /* eps, */ { alphabet::symbolFrom("Y") }, label::labelFrom(2), { alphabet::symbolFrom("Y") } );

	automaton.addTransition(label::labelFrom(2), alphabet::symbolFrom("b"), { alphabet::symbolFrom("X") }, label::labelFrom(1), { alphabet::symbolFrom("X") } );
	CPPUNIT_ASSERT_THROW(
		automaton.addTransition(label::labelFrom(2), /* eps, */ { alphabet::symbolFrom("X") }, label::labelFrom(1), { alphabet::symbolFrom("X") } ),
		automaton::AutomatonException
	);

	automaton.addTransition(label::labelFrom(4), alphabet::symbolFrom("a"), { alphabet::symbolFrom("X"), alphabet::symbolFrom("Y") }, label::labelFrom(2), { alphabet::symbolFrom("Y") } );
	automaton.addTransition(label::labelFrom(4), alphabet::symbolFrom("a"), { alphabet::symbolFrom("X"), alphabet::symbolFrom("S") }, label::labelFrom(2), { alphabet::symbolFrom("Y") } );
	automaton.addTransition(label::labelFrom(4), /* eps, */ { alphabet::symbolFrom("X"), alphabet::symbolFrom("X") }, label::labelFrom(2), { alphabet::symbolFrom("Y") } );
	CPPUNIT_ASSERT(!automaton.addTransition(label::labelFrom(4), /* eps, */ { alphabet::symbolFrom("X"), alphabet::symbolFrom("X") }, label::labelFrom(2), { alphabet::symbolFrom("Y") } ));
	CPPUNIT_ASSERT_THROW(
		automaton.addTransition(label::labelFrom(4), /* eps, */ { alphabet::symbolFrom("X"), alphabet::symbolFrom("Y") }, label::labelFrom(2), { alphabet::symbolFrom("Y") } ),
		automaton::AutomatonException
	);
	automaton.addTransition(label::labelFrom(4), alphabet::symbolFrom("b"), { alphabet::symbolFrom("X"), alphabet::symbolFrom("Y") }, label::labelFrom(2), { alphabet::symbolFrom("Y") } );
	CPPUNIT_ASSERT_THROW(
		automaton.addTransition(label::labelFrom(4), alphabet::symbolFrom("a"), { alphabet::symbolFrom("X") }, label::labelFrom(2), { alphabet::symbolFrom("Y") } ),
		automaton::AutomatonException
	);

	automaton.addFinalState(label::labelFrom(3));
}

void AutomatonTest::testExtendedNFAAlphabet() {
	automaton::ExtendedNFA automaton (label::labelFrom(0));
	label::Label s0 = label::labelFrom(0);
	label::Label s1 = label::labelFrom(1);
	automaton.addState(s0);
	automaton.addState(s1);

	automaton.setInputAlphabet({alphabet::symbolFrom('a'), alphabet::symbolFrom('b') });

	regexp::UnboundedRegExpConcatenation goodConcat1, goodConcat2;
	regexp::UnboundedRegExpConcatenation badConcat;

	goodConcat1.appendElement(regexp::UnboundedRegExpSymbol(alphabet::symbolFrom('a')));
	goodConcat1.appendElement(regexp::UnboundedRegExpSymbol(alphabet::symbolFrom('b')));

	goodConcat2.appendElement(regexp::UnboundedRegExpSymbol(alphabet::symbolFrom('b')));
	goodConcat2.appendElement(regexp::UnboundedRegExpSymbol(alphabet::symbolFrom('b')));

	badConcat.appendElement(regexp::UnboundedRegExpSymbol(alphabet::symbolFrom('b')));
	badConcat.appendElement(regexp::UnboundedRegExpSymbol(alphabet::symbolFrom('d')));

	CPPUNIT_ASSERT_THROW(automaton.addTransition(s0, regexp::RegExp{regexp::UnboundedRegExp(badConcat)}, s1), automaton::AutomatonException);
	CPPUNIT_ASSERT_NO_THROW(automaton.addTransition(s0, regexp::RegExp{regexp::UnboundedRegExp(goodConcat1)}, s1));
	CPPUNIT_ASSERT_NO_THROW(automaton.addTransition(s0, regexp::RegExp{regexp::UnboundedRegExp(goodConcat2)}, s1));
	CPPUNIT_ASSERT_NO_THROW(automaton.addTransition(s0, regexp::RegExp{regexp::UnboundedRegExp(regexp::UnboundedRegExpEpsilon())}, s1));
}

void AutomatonTest::testNPDATransitions() {
	automaton::NPDA < > automaton(label::labelFrom(0), alphabet::symbolFrom("T"));

	automaton.setStates({label::labelFrom(0), label::labelFrom(1), label::labelFrom(2), label::labelFrom(3), label::labelFrom(4)});
	automaton.setInputAlphabet({alphabet::symbolFrom("a2"), alphabet::symbolFrom("a1"), alphabet::symbolFrom("a0")});
	automaton.setPushdownStoreAlphabet({alphabet::symbolFrom("T"), alphabet::symbolFrom("R")});

	automaton.addTransition(label::labelFrom(0), alphabet::symbolFrom("a2"), {alphabet::symbolFrom("T")}, label::labelFrom(0), {alphabet::symbolFrom("T"), alphabet::symbolFrom("T")});
	automaton.addTransition(label::labelFrom(0), alphabet::symbolFrom("a1"), {alphabet::symbolFrom("T")}, label::labelFrom(0), {alphabet::symbolFrom("T")});
	automaton.addTransition(label::labelFrom(0), alphabet::symbolFrom("a0"), {alphabet::symbolFrom("T")}, label::labelFrom(0), {});

	automaton.addTransition(label::labelFrom(0), alphabet::symbolFrom("a2"), {alphabet::symbolFrom("T")}, label::labelFrom(1), {alphabet::symbolFrom("R"), alphabet::symbolFrom("T")});

	automaton.addTransition(label::labelFrom(1), alphabet::symbolFrom("a2"), {alphabet::symbolFrom("T")}, label::labelFrom(1), {alphabet::symbolFrom("T"), alphabet::symbolFrom("T")});
	automaton.addTransition(label::labelFrom(1), alphabet::symbolFrom("a2"), {alphabet::symbolFrom("R")}, label::labelFrom(1), {alphabet::symbolFrom("T"), alphabet::symbolFrom("R")});
	automaton.addTransition(label::labelFrom(1), alphabet::symbolFrom("a1"), {alphabet::symbolFrom("T")}, label::labelFrom(1), {alphabet::symbolFrom("T")});
	automaton.addTransition(label::labelFrom(1), alphabet::symbolFrom("a1"), {alphabet::symbolFrom("R")}, label::labelFrom(1), {alphabet::symbolFrom("R")});
	automaton.addTransition(label::labelFrom(1), alphabet::symbolFrom("a0"), {alphabet::symbolFrom("T")}, label::labelFrom(1), {});
	automaton.addTransition(label::labelFrom(1), alphabet::symbolFrom("a0"), {alphabet::symbolFrom("R")}, label::labelFrom(2), {});

	automaton.addTransition(label::labelFrom(2), alphabet::symbolFrom("a1"), {alphabet::symbolFrom("T")}, label::labelFrom(3), {alphabet::symbolFrom("R")});

	automaton.addTransition(label::labelFrom(3), alphabet::symbolFrom("a2"), {alphabet::symbolFrom("T")}, label::labelFrom(3), {alphabet::symbolFrom("T"), alphabet::symbolFrom("T")});
	automaton.addTransition(label::labelFrom(3), alphabet::symbolFrom("a2"), {alphabet::symbolFrom("R")}, label::labelFrom(3), {alphabet::symbolFrom("T"), alphabet::symbolFrom("R")});
	automaton.addTransition(label::labelFrom(3), alphabet::symbolFrom("a1"), {alphabet::symbolFrom("T")}, label::labelFrom(3), {alphabet::symbolFrom("T")});
	automaton.addTransition(label::labelFrom(3), alphabet::symbolFrom("a1"), {alphabet::symbolFrom("R")}, label::labelFrom(3), {alphabet::symbolFrom("R")});
	automaton.addTransition(label::labelFrom(3), alphabet::symbolFrom("a0"), {alphabet::symbolFrom("T")}, label::labelFrom(3), {});
	automaton.addTransition(label::labelFrom(3), alphabet::symbolFrom("a0"), {alphabet::symbolFrom("R")}, label::labelFrom(4), {});

	automaton.setFinalStates({label::labelFrom(4)});
	alib::XmlDataFactory::toStdout(automaton);

	CPPUNIT_ASSERT(!(automaton < automaton));
	CPPUNIT_ASSERT(automaton == automaton);
}

void AutomatonTest::testRHPDATransitions() {
	automaton::RealTimeHeightDeterministicNPDA < > automaton{alphabet::BottomOfTheStackSymbol::BOTTOM_OF_THE_STACK_SYMBOL};
	automaton.setStates({label::labelFrom(1), label::labelFrom(2), label::labelFrom(3)});
	automaton.setInitialStates({label::labelFrom(1)});

	automaton.setInputAlphabet({alphabet::symbolFrom("a"), alphabet::symbolFrom("b")});
	automaton.setPushdownStoreAlphabet({alphabet::BottomOfTheStackSymbol::BOTTOM_OF_THE_STACK_SYMBOL, alphabet::symbolFrom("X"), alphabet::symbolFrom("Y")});

	automaton.addCallTransition(label::labelFrom(1), alphabet::symbolFrom("a"), label::labelFrom(2), alphabet::symbolFrom("X"));
	automaton.addCallTransition(label::labelFrom(2), label::labelFrom(3), alphabet::symbolFrom("X"));
	automaton.addReturnTransition(label::labelFrom(3), alphabet::symbolFrom("Y"), label::labelFrom(1));

	automaton.setFinalStates({label::labelFrom(3)});
	alib::XmlDataFactory::toStdout(automaton);

	CPPUNIT_ASSERT(!(automaton < automaton));
	CPPUNIT_ASSERT(automaton == automaton);
}

void AutomatonTest::testNFTAParser() {
	alphabet::RankedSymbol a ('a', 2);
	alphabet::RankedSymbol b ('b', 1);
	alphabet::RankedSymbol c ('c', 0);

	label::Label q2 = label::labelFrom ("q2");
	label::Label q1 = label::labelFrom ("q1");
	label::Label q0 = label::labelFrom ("q0");

	automaton::NFTA automaton;

	automaton.addState(q2);
	automaton.addState(q1);
	automaton.addState(q0);
	automaton.addInputSymbol(a);
	automaton.addInputSymbol(b);
	automaton.addInputSymbol(c);

	std::vector<label::Label> states1 = {q1, q0};
	automaton.addTransition(a, states1, q2);
	automaton.addTransition(a, states1, q1);
	std::vector<label::Label> states2 = {q0};
	automaton.addTransition(b, states2, q1);
	automaton.addTransition(b, states2, q0);
	std::vector<label::Label> states3 = {};
	automaton.addTransition(c, states3, q0);

	automaton.addFinalState(q2);

	CPPUNIT_ASSERT( automaton == automaton );
	{
		std::deque<sax::Token> tokens = alib::XmlDataFactory::toTokens(automaton);
		std::string tmp;
		sax::SaxComposeInterface::printMemory(tmp, tokens);
		std::cout << tmp << std::endl;

		std::deque<sax::Token> tokens2;
		sax::SaxParseInterface::parseMemory(tmp, tokens2);
		automaton::NFTA automaton2 = alib::XmlDataFactory::fromTokens<automaton::NFTA>(std::move(tokens2));

		CPPUNIT_ASSERT( automaton == automaton2 );
	}
}

void AutomatonTest::testNFTATransitions() {
	alphabet::RankedSymbol a ('a', 2);
	alphabet::RankedSymbol b ('b', 1);
	alphabet::RankedSymbol c ('c', 0);
	alphabet::RankedSymbol d ('d', 0);

	label::Label q2 = label::labelFrom ("q2");
	label::Label q1 = label::labelFrom ("q1");
	label::Label q0 = label::labelFrom ("q0");

	automaton::NFTA automaton;

	automaton.addState(q2);
	automaton.addState(q1);
	automaton.addState(q0);
	automaton.addInputSymbol(a);
	automaton.addInputSymbol(b);
	automaton.addInputSymbol(c);

	std::vector<label::Label> states1 = {q1, q0};
	automaton.addTransition(a, states1, q2);
	automaton.addTransition(a, states1, q1);
	std::vector<label::Label> states2 = {q0};
	automaton.addTransition(b, states2, q1);
	automaton.addTransition(b, states2, q0);
	std::vector<label::Label> states3 = {};
	automaton.addTransition(c, states3, q0);

	automaton.addFinalState(q2);

	CPPUNIT_ASSERT_THROW(automaton.removeInputSymbol(a), exception::CommonException);
	CPPUNIT_ASSERT_NO_THROW(automaton.removeInputSymbol(d));
	CPPUNIT_ASSERT_NO_THROW(automaton.removeTransition(b, states1, q2));
	CPPUNIT_ASSERT_THROW(automaton.addTransition(b, states1, q1), exception::CommonException);
	CPPUNIT_ASSERT_THROW(automaton.addTransition(a, states2, q1), exception::CommonException);
	CPPUNIT_ASSERT_THROW(automaton.addTransition(d, states3, q0), exception::CommonException);
}
