#include "TreeTest.h"

#include "sax/SaxParseInterface.h"
#include "sax/SaxComposeInterface.h"

#include "tree/ranked/RankedTree.h"
#include "tree/unranked/UnrankedTree.h"
#include "tree/ranked/PrefixRankedTree.h"
#include "tree/ranked/PrefixRankedBarTree.h"
#include "tree/ranked/RankedPattern.h"

#include "tree/TreeException.h"

#include "factory/XmlDataFactory.hpp"

#include "alphabet/RankedSymbol.h"
#include "alphabet/SubtreeWildcardSymbol.h"
#include "alphabet/BarSymbol.h"

#define CPPUNIT_IMPLY(x, y) CPPUNIT_ASSERT(!(x) || (y))

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION( TreeTest, "tree" );
CPPUNIT_TEST_SUITE_REGISTRATION( TreeTest );

void TreeTest::setUp() {
}

void TreeTest::tearDown() {
}

std::tree < alphabet::RankedSymbol > prefixToNode(const std::string & s, int & pos) {
	char c = s[pos];
	int r = s[pos+1] - '0';
	pos += 2;
	std::vector<std::tree < alphabet::RankedSymbol >> children;
	for (int i = 0; i < r; i++) {
		children.push_back(prefixToNode(s, pos));
	}
	return std::tree < alphabet::RankedSymbol >(alphabet::RankedSymbol(c, r), std::move(children));
}

void TreeTest::testRankedTreeParser() {
/*
	L=(a(2), b(1), c(0))

	\-1.a(2)
	  |
	  |-2.b(1)
	  | |
	  | \-3.c(0)
	  |
	  \-4.c(0)
*/
	const alphabet::RankedSymbol a ('a', 2);
	const alphabet::RankedSymbol b ('b', 1);
	const alphabet::RankedSymbol c ('c', 0);

	const std::set<alphabet::RankedSymbol> alphabet {a, b, c};

	std::tree < alphabet::RankedSymbol > node3(c, {});
	std::tree < alphabet::RankedSymbol > node4(c, {});
	std::tree < alphabet::RankedSymbol > node2(b, {std::move(node3)});
	std::tree < alphabet::RankedSymbol > node1(a, {std::move(node2), std::move(node4)});

	tree::RankedTree tree(alphabet, std::move(node1));

	CPPUNIT_ASSERT( tree == tree );
	tree.getContent().nicePrint(std::cout);
	{
		std::deque<sax::Token> tokens = alib::XmlDataFactory::toTokens(tree);
		std::string tmp;
		sax::SaxComposeInterface::printMemory(tmp, tokens);

		std::deque<sax::Token> tokens2;
		sax::SaxParseInterface::parseMemory(tmp, tokens2);
		tree::RankedTree tree2 = alib::XmlDataFactory::fromTokens<tree::RankedTree>(std::move(tokens2));

		CPPUNIT_ASSERT( tree == tree2 );
		std::cout << std::endl;
		tree2.getContent().nicePrint(std::cout);
	}

	std::string s = "a2a2a2b1a2c0c0c0c0c0";
	int itmp = 0;
	std::tree < alphabet::RankedSymbol > node = prefixToNode(s, itmp);
	std::set<alphabet::RankedSymbol> al;
	for (unsigned i = 0; i < s.length(); i += 2) al.insert(alphabet::RankedSymbol(s[i], (int) (s[i+1] - '0')));
	tree::RankedTree t(al, std::move(node));
	alib::XmlDataFactory::toStdout(t);

	//TODO test it somehow
}

void TreeTest::testRankedTreeCompare() {
/*
	L=(a(2), b(1), c(0))

	\-1.a(2)
	  |
	  |-2.b(1)
	  | |
	  | \-3.c(0)
	  |
	  \-4.c(0)

	\-5.a(2)
	  |
	  |-6.c(0)
	  |
	  \-7.b(1)
	    |
	    \-8.c(0)
*/
	const alphabet::RankedSymbol a ('a', 2);
	const alphabet::RankedSymbol b ('b', 1);
	const alphabet::RankedSymbol c ('c', 0);

	const std::set<alphabet::RankedSymbol> alphabet {a, b, c};

	std::tree < alphabet::RankedSymbol > node3(c, {});
	std::tree < alphabet::RankedSymbol > node4(c, {});
	std::tree < alphabet::RankedSymbol > node2(b, {std::move(node3)});
	std::tree < alphabet::RankedSymbol > node1(a, {std::move(node2), std::move(node4)});

	std::tree < alphabet::RankedSymbol > node6(c, {});
	std::tree < alphabet::RankedSymbol > node8(c, {});
	std::tree < alphabet::RankedSymbol > node7(b, {std::move(node8)});
	std::tree < alphabet::RankedSymbol > node5(a, {std::move(node6), std::move(node7)});

	tree::RankedTree tree1(alphabet, std::move(node1));
	tree::RankedTree tree2(alphabet, std::move(node5));

	CPPUNIT_ASSERT( tree1 != tree2 );
	CPPUNIT_ASSERT( tree1 < tree2 );
	CPPUNIT_ASSERT( tree2 > tree1 );
	CPPUNIT_ASSERT( !(tree1 > tree2) );
}

void TreeTest::testRankedTreeSymbolValidityCheck() {
	const alphabet::RankedSymbol a ('a', 2);
	const alphabet::RankedSymbol b ('b', 1);
	const alphabet::RankedSymbol c ('c', 0);
	const alphabet::RankedSymbol d ('d', 0);
	const alphabet::RankedSymbol e ('e', 0);

	const std::set<alphabet::RankedSymbol> alphabet {a, b, c, e};

	std::tree < alphabet::RankedSymbol > node3(d, {});
	std::tree < alphabet::RankedSymbol > node4(c, {});
	std::tree < alphabet::RankedSymbol > node2(b, {std::move(node3)});
	std::tree < alphabet::RankedSymbol > node1(a, {std::move(node2), std::move(node4)});

	CPPUNIT_ASSERT_THROW(tree::RankedTree(alphabet, std::move(node1)), exception::CommonException);

	std::tree < alphabet::RankedSymbol > node6(e, {});
	std::tree < alphabet::RankedSymbol > node8(c, {});
	std::tree < alphabet::RankedSymbol > node7(b, {std::move(node8)});
	std::tree < alphabet::RankedSymbol > node5(a, {std::move(node6), std::move(node7)});

	tree::RankedTree * tree = NULL;
	CPPUNIT_ASSERT_NO_THROW(tree = new tree::RankedTree(alphabet, std::move(node5)));

	delete tree;
}

void TreeTest::testUnrankedTreeParser() {
/*
	L=(a, b, c)

	\-1.a
	  |
	  |-2.b
	  | |
	  | \-3.c
	  |
	  \-4.c
*/
	const alphabet::Symbol a = alphabet::symbolFrom('a');
	const alphabet::Symbol b = alphabet::symbolFrom('b');
	const alphabet::Symbol c = alphabet::symbolFrom('c');

	const std::set<alphabet::Symbol> alphabet {a, b, c};

	std::tree < alphabet::Symbol > node3(c, {});
	std::tree < alphabet::Symbol > node4(c, {});
	std::tree < alphabet::Symbol > node2(b, {std::move(node3)});
	std::tree < alphabet::Symbol > node1(a, {std::move(node2), std::move(node4)});

	tree::UnrankedTree tree(alphabet, std::move(node1));

	CPPUNIT_ASSERT( tree == tree );
	tree.getContent().nicePrint(std::cout);
	{
		std::deque<sax::Token> tokens = alib::XmlDataFactory::toTokens(tree);
		std::string tmp;
		sax::SaxComposeInterface::printMemory(tmp, tokens);
		std::cout << std::endl << tmp << std::endl << std::endl;

		std::deque<sax::Token> tokens2;
		sax::SaxParseInterface::parseMemory(tmp, tokens2);
		tree::UnrankedTree tree2 = alib::XmlDataFactory::fromTokens<tree::UnrankedTree>(std::move(tokens2));

		CPPUNIT_ASSERT( tree == tree2 );
		std::cout << std::endl;
		tree2.getContent().nicePrint(std::cout);
	}
}

void TreeTest::testUnrankedTreeCompare() {
/*
	L=(a, b, c)

	\-1.a
	  |
	  |-2.b
	  | |
	  | \-3.c
	  |
	  \-4.c

	\-5.a
	  |
	  |-6.c
	  |
	  \-7.b
	    |
	    \-8.c
*/
	const alphabet::Symbol a = alphabet::symbolFrom('a');
	const alphabet::Symbol b = alphabet::symbolFrom('b');
	const alphabet::Symbol c = alphabet::symbolFrom('c');

	const std::set<alphabet::Symbol> alphabet {a, b, c};

	std::tree < alphabet::Symbol > node3(c, {});
	std::tree < alphabet::Symbol > node4(c, {});
	std::tree < alphabet::Symbol > node2(b, {std::move(node3)});
	std::tree < alphabet::Symbol > node1(a, {std::move(node2), std::move(node4)});

	std::tree < alphabet::Symbol > node6(c, {});
	std::tree < alphabet::Symbol > node8(c, {});
	std::tree < alphabet::Symbol > node7(b, {std::move(node8)});
	std::tree < alphabet::Symbol > node5(a, {std::move(node6), std::move(node7)});

	tree::UnrankedTree tree1(alphabet, std::move(node1));
	tree::UnrankedTree tree2(alphabet, std::move(node5));

	CPPUNIT_ASSERT( tree1 != tree2 );
	CPPUNIT_ASSERT( tree1 < tree2 );
	CPPUNIT_ASSERT( tree2 > tree1 );
	CPPUNIT_ASSERT( !(tree1 > tree2) );
}

void TreeTest::testUnrankedTreeSymbolValidityCheck() {
	const alphabet::Symbol a = alphabet::symbolFrom('a');
	const alphabet::Symbol b = alphabet::symbolFrom('b');
	const alphabet::Symbol c = alphabet::symbolFrom('c');
	const alphabet::Symbol d = alphabet::symbolFrom('d');
	const alphabet::Symbol e = alphabet::symbolFrom('e');

	const std::set<alphabet::Symbol> alphabet {a, b, c, e};

	std::tree < alphabet::Symbol > node3(d, {});
	std::tree < alphabet::Symbol > node4(c, {});
	std::tree < alphabet::Symbol > node2(b, {std::move(node3)});
	std::tree < alphabet::Symbol > node1(a, {std::move(node2), std::move(node4)});

	CPPUNIT_ASSERT_THROW(tree::UnrankedTree(alphabet, std::move(node1)), exception::CommonException);

	std::tree < alphabet::Symbol > node6(e, {});
	std::tree < alphabet::Symbol > node8(c, {});
	std::tree < alphabet::Symbol > node7(b, {std::move(node8)});
	std::tree < alphabet::Symbol > node5(a, {std::move(node6), std::move(node7)});

	tree::UnrankedTree * tree = NULL;
	CPPUNIT_ASSERT_NO_THROW(tree = new tree::UnrankedTree(alphabet, std::move(node5)));

	delete tree;
}

void TreeTest::testPrefixRankedBarParser() {
	const alphabet::Symbol bar{alphabet::BarSymbol{}};

	const alphabet::RankedSymbol a ('a', 2);
	const alphabet::RankedSymbol b ('b', 1);
	const alphabet::RankedSymbol c ('c', 0);

	const std::set<alphabet::RankedSymbol> alphabet {a, b, c};

	std::tree < alphabet::RankedSymbol > node3(c, {});
	std::tree < alphabet::RankedSymbol > node4(c, {});
	std::tree < alphabet::RankedSymbol > node2(b, {std::move(node3)});
	std::tree < alphabet::RankedSymbol > node1(a, {std::move(node2), std::move(node4)});

	tree::RankedTree tree(alphabet, std::move(node1));
	tree::PrefixRankedBarTree tree2(bar, tree);

	CPPUNIT_ASSERT( tree2 == tree2 );
	{
		std::deque<sax::Token> tokens = alib::XmlDataFactory::toTokens(tree2);
		std::string tmp;
		sax::SaxComposeInterface::printMemory(tmp, tokens);

		std::cout << tmp << std::endl << std::endl;

		std::deque<sax::Token> tokens2;
		sax::SaxParseInterface::parseMemory(tmp, tokens2);
		tree::PrefixRankedBarTree tree3 = alib::XmlDataFactory::fromTokens<tree::PrefixRankedBarTree>(std::move(tokens2));

		CPPUNIT_ASSERT( tree2 == tree3 );
		std::cout << std::endl;
	}
}

